/**********************************************************************
 *<
	FILE: light.h

	DESCRIPTION:  Defines lights

	CREATED BY: Dan Silva

	HISTORY: created 2 January 1995
			 added directional and spot lights, Feb 95

 *>	Copyright (c) 1995, All Rights Reserved.
 **********************************************************************/

#ifndef __LIGHT__H__ 

#define __LIGHT__H__

#include "shadgen.h"

// reference #s
#define PBLOCK_REF 0
#define PROJMAP_REF 1

#define DECAY_NONE  0
#define DECAY_INV   1
#define DECAY_INVSQ 2

#define SOFTSPOT_CLASS_ID 0x448d1abc, 0x377b264a

class GeneralLight;

class ProjMapDADMgr: public DADMgr {
	GeneralLight *gl;
	public:
		ProjMapDADMgr(GeneralLight *l=NULL) { gl = l;}
		void Init(GeneralLight *d) { gl = d; }
		// called on the draggee to see what if anything can be dragged from this x,y
		SClass_ID GetDragType(HWND hwnd, POINT p);
		// called on potential dropee to see if can drop type at this x,y
		BOOL OkToDrop(ReferenceTarget *dropThis, HWND hfrom, HWND hto, POINT p, SClass_ID type, BOOL isNew);
		int SlotOwner() { return OWNER_SCENE; }  
		ReferenceTarget *GetInstance(HWND hwnd, POINT p, SClass_ID type);
		void Drop(ReferenceTarget *dropThis, HWND hwnd, POINT p, SClass_ID type);
		BOOL AutoTooltip() { return TRUE; }
	};

class GeneralLightCreateCallBack;
class STSpotLightClassDesc;




class ShadBufImp: public ShadowGenerator { 
	public:
//	ShadBufType *theShad;
	float *buffer;
	int shadsize;
	float fsize,fsm;
	float sz2; // shadsize*.5;
	float xscale; // = l->zfac*l->aspw;
	float yscale; // = -l->zfac*l->asph;
	float zfac;

	// used only in sampling
	BOOL bias_absolute;
   	BOOL parallel;
	float bias;	// maybe should be in light
	float sampleSize;

	// used only while building buffer
	RenderGlobalContext *RGC;

	//SRendParams* RP;
	Matrix3 lightToWorld;
	Matrix3 worldToLight;
	Matrix3 camToLight;
	LightObject *light;
	ObjLightDesc *ltDesc;

	// methods
	ShadBufImp( LightObject *l,  ObjLightDesc *ld, ULONG flags) { 
		buffer = NULL; 
		light = l;
		ltDesc = ld;
		parallel = flags&SHAD_PARALLEL?1:0;
		shadsize = 0;
		} 

	ShadBufImp() { buffer = NULL; } 
	~ShadBufImp() { FreeBuffer();  }

//	int Update(const ObjLightDesc& light,const RendContext& rendCntxt, Matrix3 &lightTM,int size, float aspect, float param,
//		float biasp, float sampSize, ULONG flags, float clipDist);

	int Update(
		TimeValue t,
		const RendContext& rendCntxt,   // Mostly for progress bar.
		RenderGlobalContext *rgc,       // Need to get at instance list.
		Matrix3& lightToWorld, // light to world space: not necessarly same as that of light
		float aspect,      // aspect
		float param,   	   // persp:field-of-view (radians) -- parallel : width in world coords
		float clipDist  
		);

	void FreeBuffer() {
		if (buffer) {
			delete [] buffer;
			buffer = NULL;
			}
		}

	int UpdateViewDepParams(const Matrix3& worldToCam) { 
		camToLight = Inverse(lightToWorld * worldToCam);
		return 1;
		}

	void DeleteThis() {delete this; }

	// buffer sampling
	inline float InShadow(int x, int y, float z);
//	float InterpShadow( float x, float y, float z,float xslope, float yslope);
	float EstimateZ(float x, float y);
	BOOL QuickSample(int x, int y, float z);
	float FiltSample(int x, int y, float z,int level);
	float LineSample(int x1, int y1, float z1,int x2, int y2, float z2);

	float Sample(ShadeContext &sc, float x, float y, float z, float xslope, float yslope);
	
	float Sample(ShadeContext &sc, Point3 &norm, Color& color);
	 
#ifdef SHOWSHADMAP
	void ShowShadbuf();
#endif
	};




//class GeneralLight: public GenLight {			   
class GeneralLight: public LightObject {			   
	friend class GeneralLightCreateCallBack;
	friend class LightPostLoad;
	friend class STSpotLightClassDesc;
	friend class ProjMapDADMgr;
	friend BOOL CALLBACK GeneralLightParamDialogProc( HWND hDlg, UINT message,
		 WPARAM wParam, LPARAM lParam );
	friend void resetLightParams();

	// Class vars
	static Mesh staticMesh[2];
	static short meshBuilt;
	static HWND hGeneralLight;
	static HWND hSpotLight;
	static HWND hOmni;
	static HWND hShadow;
	static IObjParam* iObjParams;
	static int dlgRed;
	static int dlgGreen;
	static int dlgBlue;
	static int dlgH;
	static int dlgS;
	static int dlgV;
	static float dlgIntensity;
	static float dlgContrast;
	static float dlgDiffsoft;
	static float dlgHotsize;
	static float dlgFallsize;
	static float dlgTDist;
	static short dlgShowCone;
	static short dlgCrossover;
	static short dlgUseAtten;
	static float dlgAttenStart;
	static float dlgAttenEnd;
	static float dlgAtten1Start;
	static float dlgAtten1End;
	static short dlgShowAtten;
	static short dlgShape;
	static float dlgAspect;
	static float dlgMapBias;
	static float dlgMapRange;
	static float dlgRayBias;
	static int dlgMapSize;

//	static float dlgRadius;
//	static int dlgNumberLights;

	static short globShadowType; 
	static short globAbsMapBias;
	static float globMapRange;
	static float globMapBias;
	static int   globMapSize;
	static float globRayBias;

	static ISpinnerControl *redSpin;
	static ISpinnerControl *greenSpin;
	static ISpinnerControl *blueSpin;
	static ISpinnerControl *hSpin;
	static ISpinnerControl *sSpin;
	static ISpinnerControl *vSpin;
	static ISpinnerControl *intensitySpin;
	static ISpinnerControl *contrastSpin;
	static ISpinnerControl *diffsoftSpin;
	static ISpinnerControl *hotsizeSpin;
	static ISpinnerControl *fallsizeSpin;
	static ISpinnerControl *tDistSpin;
	static ISpinnerControl *attenStartSpin;
	static ISpinnerControl *attenEndSpin;
	static ISpinnerControl *attenStart1Spin;
	static ISpinnerControl *attenEnd1Spin;
	static ISpinnerControl *aspectSpin;
	static ISpinnerControl *mapBiasSpin;
	static ISpinnerControl *mapSizeSpin;
	static ISpinnerControl *mapRangeSpin;
	static ISpinnerControl *rayBiasSpin;

	static ISpinnerControl *radiusSpin;
	static ISpinnerControl *numberLightsSpin;


	static IColorSwatch *colorSwatch;
	static ICustButton *projMapName;
	static ProjMapDADMgr projDADMgr;
	
	// Object parameters
	IParamBlock *pblock;
	Texmap *projMap;   // a reference
	short updateHSVSpin;
	short type;
	short enable;
	short useLight;		// on/off toggle
	short useAtten;
	short attenDisplay;
	short useAttenNear;
	short attenNearDisplay;
	short coneDisplay;
	short shape;
	short shadow;
	short useGlobalShadowParams;
	short shadowType;  // 0: ShadowMap   1: RayTrace
	short absMapBias;
	short overshoot;
	short projector;
	short softenDiffuse;
	short affectDiffuse;
	short affectSpecular;
	short decayType;
	int extDispFlags;
	float targDist; 
	Mesh spotMesh;
	Mesh *mesh;
	NameTab exclList;
	void BuildStaticMeshes();
	void BuildSpotMesh(float coneSize);
	void UpdateUI(TimeValue t);

	//  inherited virtual methods for Reference-management
	RefResult NotifyRefChanged( Interval changeInt, RefTargetHandle hTarget, PartID& partID, RefMessage message );

	void GetMat(TimeValue t, INode* inode, ViewExp *vpt, Matrix3& mat);

	
public:
	GeneralLight(int type);
	~GeneralLight();

	//  inherited virtual methods:

	// From GenLight
	LightObject *NewLight(int type) { return new GeneralLight(type); }

	// From BaseObject
	int HitTest(TimeValue t, INode* inode, int type, int crossing, int flags, IPoint2 *p, ViewExp *vpt);
	void Snap(TimeValue t, INode* inode, SnapInfo *snap, IPoint2 *p, ViewExp *vpt);
	void SetExtendedDisplay(int flags);
	int Display(TimeValue t, INode* inode, ViewExp *vpt, int flags);
	CreateMouseCallBack* GetCreateMouseCallBack();
	void BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev);
	void EndEditParams( IObjParam *ip, ULONG flags,Animatable *next);
	TCHAR *GetObjectName();

	void GetWorldBoundBox(TimeValue t, INode *mat, ViewExp* vpt, Box3& box );
	void GetLocalBoundBox(TimeValue t, INode *mat, ViewExp* vpt,  Box3& box );
	void GetDeformBBox(TimeValue t, Box3& box, Matrix3 *tm, BOOL useSel );

	// From Object
	ObjectState Eval(TimeValue time);
	void InitNodeName(TSTR& s);
	Interval ObjectValidity(TimeValue t);
	BOOL UsesWireColor()	{ return 1; }
	int DoOwnSelectHilite() { return 1; }		

	// From Light
	RefResult EvalLightState(TimeValue t, Interval& valid, LightState* cs);
	int Type() { return type; }  // OMNI_LIGHT, TSPOT_LIGHT, DIR_LIGHT, FSPOT_LIGHT, TDIR_LIGHT
#ifdef SIMPLEDIR
	BOOL IsSpot()	{ return TRUE; }
#else
	BOOL IsSpot()	{ return TRUE; }
#endif
	BOOL IsDir()	{ return FALSE; return type == DIR_LIGHT|| type == TDIR_LIGHT; }
	void SetUseLight(int onOff);
	BOOL GetUseLight(void)		{ return useLight; }
	void SetSpotShape(int s);
	int GetSpotShape(void)	{ return shape; }
	void SetHotspot(TimeValue time, float f);
	float GetHotspot(TimeValue t, Interval& valid = Interval(0,0));
	void SetFallsize(TimeValue time, float f);
	float GetFallsize(TimeValue t, Interval& valid = Interval(0,0));
	void SetAtten(TimeValue time, int which, float f);
	float GetAtten(TimeValue t, int which, Interval& valid = Interval(0,0));
	void SetTDist(TimeValue time, float f);
	float GetTDist(TimeValue t, Interval& valid = Interval(0,0));
	int GetShadowMethod() {return shadow?(shadowType==0?LIGHTSHADOW_MAPPED:LIGHTSHADOW_RAYTRACED):LIGHTSHADOW_NONE;}
	void ReloadShadCtrls(HWND hDlg, TimeValue t);	

	ObjLightDesc *CreateLightDesc(INode *n);
	
	// GenLight routines
	void SetRGBColor(TimeValue t, Point3& rgb);
	Point3 GetRGBColor(TimeValue t, Interval &valid = Interval(0,0));
	void SetHSVColor(TimeValue t, Point3& hsv);
	Point3 GetHSVColor(TimeValue t, Interval &valid = Interval(0,0));
	void SetUpdateHSVSpin(int b)	{ updateHSVSpin = b; }
	void SetIntensity(TimeValue time, float f);
	float GetIntensity(TimeValue t, Interval& valid = Interval(0,0));
	void SetContrast(TimeValue time, float f);
	float GetContrast(TimeValue t, Interval& valid = Interval(0,0));
	void SetDiffuseSoft(TimeValue time, float f);
	float GetDiffuseSoft(TimeValue t, Interval& valid = Interval(0,0));
	void SetAspect(TimeValue t, float f);
	float GetAspect(TimeValue t, Interval& valid = Interval(0,0));
	void SetConeDisplay(int s, int notify=TRUE);
	BOOL GetConeDisplay(void)		{ return coneDisplay; }
	void DrawSphereArcs(TimeValue t, GraphicsWindow *gw, float r, Point3 *q);
    int GetCirXPoints(TimeValue t, float angle, float dist, Point3 *q);
	void DrawCircleX(TimeValue t, GraphicsWindow *gw, float angle, float dist, Point3 *q);
	int GetRectXPoints(TimeValue t, float angle, float dist, Point3 *q);
	void DrawWarpRect(TimeValue t, GraphicsWindow *gw, float angle, float dist, Point3 *q);
	void GetConePoints(TimeValue t, float aspect, float angle, float dist, Point3 *q);
	void DrawCone(TimeValue t, GraphicsWindow *gw, float dist);
	int DrawConeAndLine(TimeValue t, INode* inode, GraphicsWindow *gw, int drawing );
	void SetUseAtten(int s);
	BOOL GetUseAtten(void)			{ return useAtten; }
	void SetUseAttenNear(int s);	   	
	BOOL GetUseAttenNear(void)	   	{ return useAttenNear; }
	void SetAttenDisplay(int s);
	BOOL GetAttenDisplay(void)		{ return attenDisplay; }
	void SetAttenNearDisplay(int s);
	BOOL GetAttenNearDisplay(void)		{ return attenNearDisplay; }
	void GetAttenPoints(TimeValue t, float rad, Point3 *q);
	int DrawAtten(TimeValue t, INode *inode, GraphicsWindow *gw);
	void Enable(int enab) { enable = enab; }

	void FixHotFallConstraint();
	int GetUseGlobal() { return useGlobalShadowParams; }
	void SetUseGlobal(int a) {useGlobalShadowParams =a; }
	int GetShadow() { return shadow; }
	void SetShadow(int a) { shadow = a; }
	int GetShadowType() { return useGlobalShadowParams?globShadowType:shadowType; }
	void SetShadowType(int a) { 
		if (useGlobalShadowParams) 	globShadowType = a; 
		else shadowType = a;
		}
	void SetMapBias(TimeValue t, float f);
	float GetMapBias(TimeValue t, Interval& valid = Interval(0,0));
	void SetMapRange(TimeValue t, float f);
	float GetMapRange(TimeValue t, Interval& valid = Interval(0,0));
	void SetMapSize(TimeValue t, int f);
	int GetMapSize(TimeValue t, Interval& valid = Interval(0,0));


/*
	void SetRadius(TimeValue t, float f);
	float GetRadius(TimeValue t, Interval& valid = Interval(0,0));

	void SetNumberLights(TimeValue t, int f);
	int GetNumberLights(TimeValue t, Interval& valid = Interval(0,0));
*/

	void SetRayBias(TimeValue t, float f);
	float GetRayBias(TimeValue t, Interval& valid = Interval(0,0));
	int GetAbsMapBias() { return useGlobalShadowParams ? globAbsMapBias : absMapBias; }
	void SetAbsMapBias(int a) {
		if (useGlobalShadowParams) 	globAbsMapBias = a; 
		else absMapBias = a; 
		}
	int GetOvershoot() { return overshoot; }
	void SetOvershoot(int a);
	int GetProjector() { return projector; }
	void SetProjector(int a) { projector = a; }
	void DoBitmapFit(HWND hwnd, TimeValue t);
	NameTab* GetExclList() {return &exclList; }
	BOOL Include() {return exclList.TestFlag(NT_INCLUDE); }
	Texmap* GetProjMap() { return projMap;  }
	void SetProjMap(Texmap* pmap);
	void UpdateTargDistance(TimeValue t, INode* inode);
	void AssignProjectorMap(Texmap *m, BOOL newmat);
	void BrowseProjectorMap(HWND hWnd);	

	NameTab& GetExclusionList() {return exclList;}
	void SetExclusionList(NameTab &list) {exclList = list;}

	BOOL SetHotSpotControl(Control *c);
	BOOL SetFalloffControl(Control *c);
	BOOL SetColorControl(Control *c);
	Control* GetHotSpotControl();
	Control* GetFalloffControl();
	Control* GetColorControl();
	void SetAffectDiffuse(BOOL onOff) { affectDiffuse = onOff; }
	BOOL GetAffectDiffuse() {return affectDiffuse; }
	void SetAffectSpecular(BOOL onOff) { affectSpecular = onOff; }
	BOOL GetAffectSpecular() {return affectSpecular; }
	void SetSoftenDiffuse(BOOL onOff) { softenDiffuse = onOff; }
	BOOL GetSoftenDiffuse() {return softenDiffuse; }
	void SetDecayType(BOOL onOff) { decayType = onOff; }
	BOOL GetDecayType() {return decayType; }

	// internal routines
	void BoxCircle(TimeValue t, float r, float d, Box3& box, Point3 *q, int extraPt=0, Matrix3 *tm=NULL);
	void BoxLight(TimeValue t, INode *inode, Box3& box, Matrix3 *tm=NULL);
	ISpinnerControl* AttenSpinner(int atype);
	void ChangeAtten(int atype);
	void UpdateColBrackets(TimeValue t);
	void SetSpinnerBracket(ISpinnerControl *spin, int pbid);

	// Animatable methods
	void DeleteThis() { delete this; }

	Class_ID ClassID() { return  Class_ID(SOFTSPOT_CLASS_ID); }  
	void GetClassName(TSTR& s) { s = TSTR(GetString(IDS_DB_LIGHT_CLASS)); }

	void EnumAuxFiles(NameEnumCallback& nameEnum, DWORD flags); 
	
	int NumSubs() { return 1; }  
	Animatable* SubAnim(int i) { return pblock; }
	TSTR SubAnimName(int i) { return TSTR(GetString(IDS_RB_PARAMETERS));}

	// From ref
	RefTargetHandle Clone(RemapDir& remap = NoRemap());
	int NumRefs() {return 2;}
	RefTargetHandle GetReference(int i) {
		 return (i==0)?(RefTargetHandle)pblock: (RefTargetHandle)projMap;
		 }
	void SetReference(int i, RefTargetHandle rtarg) {
		if (i==0)pblock=(IParamBlock*)rtarg; 
		else {
			projMap = (Texmap *)rtarg;
			if (projMapName) {
				projMapName->SetText(projMap?projMap->GetFullName().data():GetString(IDS_DB_NONE));
				}
			}
		}

	// IO
	IOResult Save(ISave *isave);
	IOResult Load(ILoad *iload);

	LRESULT CALLBACK TrackViewWinProc( HWND hwnd,  UINT message, 
            WPARAM wParam,   LPARAM lParam ){return(0);}
};



#endif
