/**********************************************************************
 *<
	FILE: rendshad.cpp

	DESCRIPTION:  Default Shadow Map  

	CREATED BY: Dan Silva

	HISTORY:

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/


#include "prim.h"
#include "imtl.h"
#include "abuf.h"
#include "render.h"  
#include "shadgen.h"
#include "rend.h"
#include "shadbuf.h"
#include "light.h"
#include "gbuf.h"


#include "iparamm.h"


#define FLABS(x) ((x)>0.0f?(x):-(x))

static float min_shad_bias = .001f; // Was this a 3DS.SET var?
int abTamper=0;


//--------------------------------------------------------------------------------------
// ShadBufImp methods
//--------------------------------------------------------------------------------------

int ShadBufImp::Update(
		TimeValue t,
		const RendContext& rendCntxt,   // Mostly for progress bar.
		RenderGlobalContext *rgc,       // Need to get at instance list.
		Matrix3& ltToWorld, // light to world space: not necessarly same as that of light (for omni's)
		float aspect,      // aspect
		float param,   	   // persp:field-of-view (radians) -- parallel : width in world coords
		float clipDist  
		)
	{
	float *z;
//	RendContextImp& rc = (RendContextImp &)rendCntxt;
	RGC = rgc;

	if (RGC==NULL) 
		return 0;
	
	lightToWorld  = ltToWorld;
	worldToLight = Inverse(lightToWorld);

//FIX	sampleSize = theShad->GetMapRange(t);

//FIX	int size = theShad->GetMapSize(t);
		int size = 512;
//FIX	bias = theShad->GetMapBias(t);
//FIX 	bias_absolute = theShad->GetAbsMapBias(t);

	if (buffer==NULL||shadsize!=size) {
		shadsize = size;
		fsize = float(size);
		fsm = fsize -.001f;
		delete [] buffer;
		int nshad = shadsize*shadsize;
		buffer = new float[nshad];
		if (buffer==NULL) return 0;
		sz2 = (float)shadsize/2.0f;
		}

	if (parallel) {
		xscale =  sz2/param;
		yscale = -xscale*aspect;
		}
	else {
		zfac = -sz2/float(tan(0.5*(double)param));
		xscale = zfac;
		yscale = -zfac*aspect;
		}

	int nshad = shadsize*shadsize;
	z = buffer;

	for (int ix=0; ix<nshad; ix++)	*z++ =  BIGFLOAT;  

	ShadBufRenderer *sbr = NewDefaultShadBufRenderer();

	int nRendered = sbr->Render(rc, RGC, buffer, parallel, shadsize, param, aspect, clipDist, ltDesc, worldToLight);

	if (nRendered==0) {
		delete [] buffer;
		sbr->DeleteThis();
		buffer = NULL;
		return 1;
		}

	for (ix=0,z=buffer; ix<nshad; ix++,z++)	{
		if (*z == BIGFLOAT) *z = -1.0f;
		}

	if (!bias_absolute) {
		bias = bias*(sbr->Furthest()-sbr->Closest())/100.0f;
		if (bias<min_shad_bias) bias = min_shad_bias;
		}
	sbr->DeleteThis();
	return 1;
	}



// z-buffer values correspond to pixel centers, so find point on
// plane that corresponds. (ix and iy are guaranteed non-negative)
//#define IXOFFS(ix) ((float(int(ix))+0.5f-(xsamp))*xslope)   
//#define IYOFFS(iy) ((float(int(iy))+0.5f-(ysamp))*yslope)   

//-------------------------------------------------------------------------
// DS: 8/11/98
// The purpose of the offset is to avoid self shadowing of surfaces seen
// by the light at angle.  Instead of comparing the various points in the
// shadow buffer within the sampling rectangle with the same z, the slope 
// of the plane is taken into account to compute different z's at each 
// sample.   But this is really only neccessary when points on the plane get
// closer to the light, where if the z isn't moved the object will shadow itself.
// When moving away that isn't necessary, and avoiding it gets rid of problems 
// with incorrect shadowing of face by adjacent faces.  
// Also I have added an extra slope increment to always tend to slant away from
// the light a little bit.  This cleans up things a bit more.


//-------------------------------------------------------------------------

#define EXTSLOPE 0.5f

static inline float offsfunc(float d,float slope, float limslope) {
	// return d*slope;  // old method equivalent
	// if (d*slope) is positive, plane slants toward light -- exaggerate this.
	// if (d*slope) is negative, plane slants away -- change to slant toward a bit. 
	if (d<0.0f) 	
		return (slope>0.0f)? -limslope*d: (slope-limslope)*d;
	else 
		return (slope<0.0f)?  limslope*d: (slope+limslope)*d;
	}

#define XOFFS(dx) offsfunc(dx,xslope,slext)	   
#define YOFFS(dy) offsfunc(dy,yslope,slext)	   

#define ZOFFS(xx,yy) (XOFFS(xx) + YOFFS(yy))
//-------------------------------------------------------------------------

#define addin(x,y,area)  { totarea+=area; v += (area)*InShadow(x,y,zp); /*printf("x=%d, y=%d, a=%.3f v=%.4f inshad:%.4f\n",x,y,area,v,InShadow(x,y,zp));*/ }
#define SHADOW(xx,yy,zz) (((samp=buffer[(yy)*size+xx])>0.0f)&&(-zz>samp)) // ZSIGN, BIAS
			  
// return attenuation: 0.0 means in shadow, 1.0 not shadowed.
inline float ShadBufImp::InShadow(int x, int y, float z) {
	float samp;

	// DS 1/16/97: this can't happen, because the sample
	// is now clipped at a higher level in Sample.
	// CCJ 12/18/97 - enabled the code again, since it _does_ happen.
	// DS 5/15/98: I fixed clipping problem at higher level- so I'm removing the check again.
	//	if (x<0||y<0||x>shadsize-1||y>shadsize-1) 
	//		return (1.0f); 

	if ((samp=buffer[(y)*shadsize+x])<=0.0f) 
		return(1.0f);
	return(-z>(samp/*+bias*/)?0.0f:1.0f);  // ZSIGN,  BIAS
	}

// Samples a line in the shadow buffer using a DDA line algorithm.
// Z is interpolated in floating point... might be faster to do integer?
float ShadBufImp::LineSample(
		int x1, int y1, float z1,int x2, int y2, float z2)
	{
	
	if (buffer==NULL) return 1.0f; 

	// Cheap clipping -- this is OK since the line segment
	// shouldn't be very far out of range.
	int size1 = shadsize-1;
	if (x1<0) x1 = 0; if (x1>size1) x1 = size1;
	if (x2<0) x2 = 0; if (x2>size1) x2 = size1;
	if (y1<0) y1 = 0; if (y1>size1) y1 = size1;
	if (y2<0) y2 = 0; if (y2>size1) y2 = size1;	
	int	dx = x2-x1;
	int dy = y2-y1;
	int ax = ABS(dx) << 1;
	int ay = ABS(dy) << 1;
	int sx = dx < 0 ? -1 : 1;
	int sy = dy < 0 ? -shadsize : shadsize;
	int d;	
	float z = float(-z1), dz, samp;	
	float *ptr = &buffer[(y1)*shadsize+x1];
	int vis=0, steps=1, end;
	
	// Single point
	if (!ax && !ay) {
		if ((samp=*ptr)<=0.0f) return 1.0f;
		else return z<=samp ? 1.0f: 0.0f;
		}

	if (ax>ay) {
		d   = ay - (ax>>1);
		end = ABS(dx);
		dz  = float(z1-z2)/float(end);
		while(1) {			
			if ((samp=*ptr)>0.0f) {
				if (z<=samp) vis++;
			} else {
				vis++;
				}
			if (steps>=end) break;
			if (d>=0) {				
				ptr += sy;
				d -= ax;
				}			
			steps++;
			ptr += sx;
			d += ay;
			z += dz;
			}
	} else {
		d   = ax - (ay>>1);
		end = ABS(dy);
		dz  = float(z1-z2)/float(end);		
		while(1) {			
			if ((samp=*ptr)>0.0f) {
				if (z<=samp) vis++;
			} else {
				vis++;
				}
			if (steps>=end) break;
			if (d>=0) {
				ptr += sx;
				d -= ay;
				}			
			steps++;
			ptr += sy;
			d += ax;
			z += dz;
			}
		}

	return float(vis)/float(steps);
	}

BOOL ShadBufImp::QuickSample(int x, int y, float z) {
	if (buffer==NULL) return TRUE;
	float samp;
	int size = shadsize;
	
	// RB 4-6-96: As far as volume lights are concerned (since they don't support overshoot)
	// being outside the map is the same as being in a shadow. 	
	if (x<0||y<0||x>size-1||y>size-1) return FALSE;
	//if (x<0||y<0||x>size-1||y>size-1) return TRUE;

	if ((samp=buffer[(y)*size+x])<=0.0f) 
		return TRUE;
	return(-z>samp?FALSE:TRUE); // ZSIGN
	}

float ShadBufImp::FiltSample(int x, int y, float z,int level)
	{
	if (buffer==NULL) return 1.0f;
	float samp = 0.0f, s;
	float tot = 0.0f;
	int size  = shadsize;
	
#define	DOSAMPLE(a) \
	if (x>=0&&y>=0&&x<size&&y<size) {  tot +=a; \
		if ((s=buffer[(y)*size+x])<=0.0f||(-z<=s)) \
			samp += a;  \
		}
	
	if (level==1) {
		DOSAMPLE(0.2f)
		x-=1;
		DOSAMPLE(0.2f)
		x+=2;
		DOSAMPLE(0.2f)
		x-=1; y-=1;
		DOSAMPLE(0.2f)
		y+=2;
		DOSAMPLE(0.2f)
	} else {
		x-=1; y-=1;
		DOSAMPLE(0.11f)
		x++;
		DOSAMPLE(0.11f)
		x++;
		DOSAMPLE(0.11f)
		y++;
		DOSAMPLE(0.11f)
		x--;
		DOSAMPLE(0.12f)
		x--;
		DOSAMPLE(0.11f)
		y++;
		DOSAMPLE(0.11f)
		x++;
		DOSAMPLE(0.11f)
		x++;
		DOSAMPLE(0.11f)
		}
	if (tot==0.0f) return 1.0f;	
	return samp/tot;
	}


float ShadBufImp::Sample(ShadeContext &sc, Point3 &norm, Color& color) { 
	Point3 p = sc.P();
	Point3 n = VectorTransform(camToLight,norm);
	Point3 plt = camToLight*p; 
	float mx,my,x,y;
	if (parallel) {
		if (n.z==0.0f) return 1.0f;
		mx =  xscale*plt.x;
		my =  yscale*plt.y;
		x = sz2 + mx;
		y = sz2 + my;
		return Sample(sc, x, y, plt.z, -n.x/(n.z*xscale), -n.y/(n.z*yscale));
		}
	else {
		mx =  xscale*plt.x/plt.z;
	    my =  yscale*plt.y/plt.z;
		x = sz2 + mx;
	    y = sz2 + my;
		float k = plt.z*plt.z/(zfac*DotProd(n,plt));
		return Sample(sc, x, y, plt.z, -n.x*k, n.y*k);
		}
	}


float ShadBufImp::Sample(ShadeContext &sc, float x, float y, float z, float xslope, float yslope) {
	int xa,xb,ya,yb,i,j;
	float x2,y2,lfrac,rfrac,tfrac,bfrac,xfrac,yfrac,v,z0,zp,zy,totarea;
	float xoffsTab[52];
	if (buffer==NULL) return 1.0f;
// 	InterpShadow was malfunctioning, and it is not necessary, since nobody in their
//   right mind would use samplesize<1.0f, so I removed this code. ( DS: 8-12-98)
//	if (sampleSize<1.0f) {
//		return(InterpShadow(x, y, z, xslope, yslope));
//		}
	float xsamp = x;  // for use in ZOFFSET macro
	float ysamp = y; 
	v = 0.0f;
	totarea=0.0f;

	// DS 2/6/96-- Reduce the size of the sample at grazing angles
	#define XFACT 0.2f
	float xSampSize = sampleSize*1.0f/(1.0f+XFACT*FLABS(xslope));
	float ySampSize = sampleSize*1.0f/(1.0f+XFACT*FLABS(yslope));

	x -= (xSampSize/2.0f);
	y -= (ySampSize/2.0f);
	x2 = x + xSampSize;
	y2 = y + ySampSize;

	// DS 1/16/97: clip sample to map rectangle
	if (x<0.0f) x = 0.0f;
	if (y<0.0f) y = 0.0f;
	if (x2>fsm) 
		x2 = fsm;
	if (y2>fsm) 
		y2 = fsm;
	if (x>=x2||y>=y2) return 0.0f;
	
	xa = (int)x;
	xb = (int)x2;
 	ya = (int)y;
	yb = (int)y2;
	lfrac = 1.0f-(x-(float)floor(x));	
	rfrac = x2-(float)floor(x2);
	tfrac = 1.0f-(y-(float)floor(y));	
	bfrac = y2-(float)floor(y2);
	
	z0 = z + bias;
	float dx0 =  float(xa)+0.5f-xsamp;
	float dy0 =  float(ya)+0.5f-ysamp;
	
	float slext = parallel? EXTSLOPE/xscale: EXTSLOPE*z/xscale;
	if (slext<0.0) slext = -slext;
		
	if (xa==xb) {
		xfrac = lfrac+rfrac-1.0f;
		if (ya==yb) { 
			return(InShadow(xa, ya, z0 + ZOFFS(dx0,dy0)));
			}
		else {  
			/* --- column of pixels */
			float dy = dy0;
			float zx = z0 + XOFFS(dx0);
			zp = zx + YOFFS(dy);
			addin(xa,ya,xfrac*tfrac);
			dy += 1.0f;
			for (j=ya+1; j<yb; j++) {
				zp = zx + YOFFS(dy);
				addin(xa,j,xfrac);
				dy += 1.0f;
				}
			zp = zx + YOFFS(dy);
			addin(xa,yb,xfrac*bfrac);
			}
		}
	else {
		if (ya==yb) { 
			/* --- row of pixels */
			float dx = dx0;
			float zy = z0 + YOFFS(dy0);
			zp = zy + XOFFS(dx);
			yfrac = tfrac+bfrac-1.0f;
			addin(xa,ya,yfrac*lfrac);
			dx += 1.0f;
			for (i=xa+1; i<xb; i++) {
				zp = zy + XOFFS(dx);
				addin(i,ya,yfrac);
				dx += 1.0f;
				}
			zp = zy + XOFFS(dx);
			addin(xb,ya,yfrac*rfrac);
			}
		else {  
			/* general case */
			/* --- top row */
			{
			// Make a table of x offsets, to avoid recomputing on each row
			float dx = dx0;
			for (int i = xa; i<=xb; i++) {
				xoffsTab[i-xa] = XOFFS(dx);  
				dx += 1.0f;
				}
			}
			float dy = dy0;
			zy = z0 + YOFFS(dy);
			zp = zy + xoffsTab[0];
			addin(xa,ya,tfrac*lfrac);
			for (i=xa+1; i<xb; i++) {
				zp = zy + xoffsTab[i-xa];
				addin(i,ya,tfrac);
				}
			zp = zy + xoffsTab[xb-xa];
			addin(xb,ya,tfrac*rfrac);
			dy += 1.0f;

			/* --- middle (whole pixels ) */
			for (j=ya+1; j<yb; j++) {
				zy = z0 + YOFFS(dy);
				zp = zy + xoffsTab[0];
				addin(xa,j,lfrac);
				for (i=xa+1; i<xb; i++) {
		   			zp = zy + xoffsTab[i-xa];  
					addin(i,j,1.0f);
					}
	   			zp = zy + xoffsTab[xb-xa];
				addin(xb,j,rfrac);
				dy += 1.0f;
				}

			/* --- bottom row */
			zy = z0 + YOFFS(dy);
			zp = zy + xoffsTab[0];
			addin(xa,yb,bfrac*lfrac);
			for (i=xa+1; i<xb; i++) {
				zp = zy + xoffsTab[i-xa];
				addin(i,yb,bfrac);
				}
			zp = zy + xoffsTab[xb-xa];
			addin(xb,yb,bfrac*rfrac);
			}
		}
	float a = v/totarea;
	return (a>.9999f)?1.0f:a;
	}

#define AB_CHECK_SUM 0 

// Check sum of abuf tables.
int CheckSumABTab() {
	if (!bitmasksDecrypted&&!abTamper) {
		int sum = 0;
		for (int i=0; i<18160; i++)  
			sum += bitmasks[i];
		if (sum!=AB_CHECK_SUM) 
			abTamper = 1;
		}
	return 0;
	}


