/**********************************************************************
 *<
	FILE: shadbuf.cpp

	DESCRIPTION:  Shadow buffer renderer

	CREATED BY: Dan Silva

	HISTORY:

 *>	Copyright (c) 1994, All Rights Reserved.
 **********************************************************************/


#include "prim.h"
#include "imtl.h"
#include "abuf.h"
#include "iparamm.h"
#include "render.h"  
#include "shadgen.h"
#include "shadbuf.h"
#include "light.h"

#ifdef INRENDER
#include "rend.h"
#else 
struct DPoint {double x,y; };
//struct Point4 {float x,y,z,w; };
typedef unsigned short ushort;
typedef float Plane[4];
typedef double DPlane[4];
struct IColor { int r,g,b;};
struct SColor { WORD r,g,b;};


// SPlane
#define  PLANE_X 		1 
#define  PLANE_Y 		2 
#define  PLANE_POS_Z 	3 
#define  PLANE_NEG_Z 	4 
#define  PLANE_XZ 		5 
#define  PLANE_YZ 		6 
#define  PLANE_GEN 		7
#define  PLANE_PX   	8
#define  PLANE_PY   	9
#define  PLANE_PZ   	10

struct SPlane {
	int flags;
	Point3 p;  // direction
	float d;
	SPlane() { flags = PLANE_GEN; }
	void Set(int fl, float x, float y, float z=0.0f, float dd=0.0f) {	flags = fl;	p = Point3(x,y,z);	d = dd;	}
	};

#define BIGFLOAT (float(1.0e30))

static int	tnormal(Point2 a, Point2 b, Point2 c)
	{
	float vwz;
	/* Cross product */

	vwz=(b.x-a.x)*(c.y-a.y)-(b.y-a.y)*(c.x-a.x);

	if(vwz>0.0)
		return(1);	/* Counter-clockwise */
	else
		if(vwz<0.0)
	return(-1);	/* Clockwise */
	/* Edge-on! */

	return(0);
	}

static int isFacing(Point3& p0, Point3& p1, Point3& p2 , int projType) {
	if (projType==PROJ_PERSPECTIVE) {
		FLOAT t0,t1,t2,d;
		t0 = p1.y*p2.z - p1.z*p2.y;
		t1 = p2.y*p0.z - p2.z*p0.y;
		t2 = p0.y*p1.z - p0.z*p1.y;
		d = -(p0.x*t0 + p1.x*t1 + p2.x*t2);
		return((d<0.0)?0:1);
		}
	else {
		float d = (p1.x-p0.x)*(p2.y-p0.y) - (p2.x-p0.x)*(p1.y-p0.y);
		return((d<0.0)?0:1);
		}
	}

static float DistToPlane(Point3& p, SPlane& s) {
	switch(s.flags) {
		case PLANE_POS_Z: return s.d + p.z;
		case PLANE_NEG_Z: return s.d - p.z;
		
		case PLANE_X: return p.x*s.p.x - p.z;
		case PLANE_Y: return p.y*s.p.y - p.z;
		
		case PLANE_XZ: return p.x*s.p.x + p.z*s.p.z;
		case PLANE_YZ: return p.y*s.p.y + p.z*s.p.z;

		case PLANE_PX: return p.x*s.p.x + s.d;
		case PLANE_PY: return p.y*s.p.y + s.d;
		case PLANE_PZ: return p.z*s.p.z + s.d;
		default:
			return p.x*s.p.x + p.y*s.p.y + p.z*s.p.z + s.d ;
		}
	}


static WORD CompClipFlags(Point3 p, SPlane *pl, int nPlanes) {
	float d[10];
	WORD flags = 0;
	for (int i=0; i<nPlanes; i++)  {
		// is point behind plane?
		if ((d[i] = DistToPlane(p,pl[i])) < 0.0f) 
			flags |= (1 << i);
		}
	return flags;
	}

static void NormPoint3(Point3& p) {
	float l = Length(p);
	if (l==0.0f) l= .0000001f;
	p/=l;
	}

static void CompPlaneEqn(Plane plane, const Point3& p0, const Point3& p1, const Point3& p2) {
	Point3 e1 = p1-p0;
	Point3 e2 = p2-p0;
	Point3 p = CrossProd(e1,e2);
	NormPoint3(p);
	plane[0] = p.x;
	plane[1] = p.y;
	plane[2] = p.z;
	plane[3] = -DotProd(p0,p);
	}


#endif

#define MAX3(a,b,c) ((a)>(b)?((a)>(c)?(a):(c)):((b)>(c)?(b):(c)))
#define MIN3(a,b,c) ((a)<(b)?((a)<(c)?(a):(c)):((b)<(c)?(b):(c)))
#define FLABS(x) ((x)>0.0f?(x):-(x))


class ShadBufRendImp:public ShadBufRenderer {
	public:
	// these values must be initialized	before calling Render
	BOOL parallel;
	int shadsize;
	float param;   	   // persp:field-of-view (radians) -- parallel : width in world coords
	float aspect;
	float clipDist;
	ObjLightDesc *ltDesc;
	Matrix3 worldToLight;
	RenderGlobalContext *RGC;
	Mtl *curMtl;
	ULONG matreq;
	float sz2,xscale,yscale;
	float *buffer;
	float furthest,closest; // returned values

	int RendShadows(RenderInstance *inst, ClipFlags *cf, Matrix3& objToLight, BOOL isClip=FALSE);
	Point2 Proj2D(Point3 p);
	void ScanFace(Point2 *pts, int ystart, int yend, Plane &theplane, int flip);
	void ReadyShadWireTriangle(Point3 *pv, Point2*spt, Point2 *wspt, int *vis);
	void DeleteThis() { delete this; }
#ifdef SHOWSHADMAP
	void ShowShadbuf();
#endif


	int Render(
		RendContext &rc, 
		RenderGlobalContext *RGC, 
		float *buf,
		BOOL parallel,
		int shadsize,
		float param,   	   // persp:field-of-view (radians) -- parallel : width in world coords
		float aspect,
		float clipDist,
		ObjLightDesc *ltDesc,
		Matrix3 worldToLight
		);

	float Furthest(){ return furthest; }
	float Closest() { return closest; }

	};

ShadBufRenderer *NewDefaultShadBufRenderer() { return new ShadBufRendImp(); }

class ClipFlags {
	public:
	int culled;
	BitArray clipped;    	// array of bits[numFaces] that tells which faces were clipped.
	int numFaces;
	ClipFlags() { culled = 0; numFaces = 0; }
	void SetClip(int i, BOOL on = TRUE) {
		if (clipped.GetSize()==0) 
			clipped.SetSize(numFaces); 
		clipped.Set(i,on);
		}
	void ClearFlags() {
		clipped.SetSize(0);
		culled = 0;
		}
	};


class ShClipInstance: public RenderInstance {
	int nFaces;
	int nVerts;
	int nTVerts;
	int nTVerts2;
	RenderInstance** origInst;  	// array[numFaces] of original instances
	DWORD* 	origFaceNum;  	// array[numFaces] of original face#
	public:
		ShClipInstance(const RenderGlobalContext *rc=NULL);
		~ShClipInstance();
	   	RenderInstance *GetOrigInst(int nf) { return origInst[nf]; }
		BOOL IsClip() { return TRUE; }  // is this a clip object?
		BOOL IsEmpty() { return (nFaces==0)?1:0; }
		void AddFaces(Point3* verts, int nv, Face *face, int nf,
			 RenderInstance *origInst,  int origFace);
		//TriBary GetTriBary(int nf) { TriBary(); }
		int NumVerts() { return nVerts; }
		int NumTVerts() { return nTVerts; }
		int NumTVerts2() { return nTVerts2; }
		void Finish() {
			mesh->setNumFaces(nFaces, TRUE);
			mesh->setNumVerts(nVerts, TRUE);
			}

		RenderInstance *Next() { return NULL; }	// next in list
	    Interval MeshValidity() { return FOREVER; }
		int NumLights() { return 0; }
		LightDesc *Light(int n) { return NULL; }
		BOOL CastsShadowsFrom(const ObjLightDesc& lt) { return 1; }
		INode *GetINode() { return NULL; }
		Object *GetEvalObject() { return NULL; }
		ULONG MtlRequirements(int mtlNum) { return 0; }
		Point3 getFaceVNormal(int nf, int nv);
		
		Point3 GetFaceNormal(int faceNum) { return Point3(0,0,0); }
		Point3 GetFaceVertNormal(int faceNum, int vertNum){ return Point3(0,0,0); }  // camera coords
	    void GetFaceVertNormals(int faceNum, Point3 n[3]){}   // camera coords
		Point3 GetCamVert(int vertnum){ return Point3(0,0,0); }
		void GetObjVerts(int fnum, Point3 obp[3]) {} // vertices of face in object coords
		void GetCamVerts(int fnum, Point3 cp[3]){}  // vertices of face in camera(view) coords
	};

#define VERT_GROW 500
#define FACE_GROW 5100

typedef RenderInstance *HRInstance;

void ShClipInstance::AddFaces(
	Point3* vert,		// array of verts to add
	int nv, 			// number of verts 
	Face *face, 		// array of faces
	int nf,				// number of faces
    RenderInstance *orig, 	// original Instance
	int origFace 		// face number in original mesh
	) {
	// if gotuv, then assume number of tfaces added is same as
	// number of faces, and ditto for UVVerts and Verts.	
	int i;
	if (nVerts+nv > mesh->numVerts) {
		// Grow number of verts;
		int nvnew = mesh->numVerts + nv + VERT_GROW;
		int oldnverts = mesh->numVerts;
		mesh->setNumVerts(nvnew, TRUE);
		}
	if (nFaces+nf > mesh->numFaces) {
		// Grow number of faces;
		int nnew = mesh->numFaces + nf+FACE_GROW;
		int oldnf = mesh->numFaces;
		mesh->setNumFaces(nnew, TRUE);
		HRInstance *newOI = new HRInstance[nnew];
		DWORD *newOFN = new DWORD[nnew];
//		TriBary *newTrib = new TriBary[nnew];
		if (origInst) {
			for (i=0; i<oldnf; i++) {
				newOI[i] = origInst[i];
				newOFN[i] = origFaceNum[i];
				}
			delete [] origInst;
			delete [] origFaceNum;
			}
		origInst = newOI;
		origFaceNum = newOFN;
		}

	for ( i = 0; i< nv; i++ ) {
		mesh->verts[nVerts+i] = vert[i];		
		}

	for ( i = 0; i< nf; i++ ) {
		mesh->faces[nFaces+i] = face[i];		
		origInst[nFaces+i] = orig;
		origFaceNum[nFaces+i] = origFace;
		}

	nVerts += nv;
	nFaces += nf;
	}


ShClipInstance::ShClipInstance(const RenderGlobalContext *rc) {
	flags = INST_CLIP;
	mesh = new Mesh;
	nFaces = nVerts = nTVerts = nTVerts2 = 0;
	origInst = NULL;
	origFaceNum = NULL;
	if (rc) 
		objToWorld = rc->camToWorld;
	objToCam.IdentityMatrix();
	camToObj.IdentityMatrix();
	}

ShClipInstance::~ShClipInstance() {
	delete [] origInst;
	delete [] origFaceNum;
	delete mesh;
	}

class ShFaceClip {
	public:
	RenderInstance *inst;
	ShClipInstance *clip;
	ShFaceClip(RenderInstance *i, ShClipInstance *ci) { inst = i; clip = ci; }
	void ClipMe(Face &f, int nf, int ct, Point3* h,  Point2* b, BOOL *vis);
	};

void Clip(RenderInstance *inst, ClipFlags *cf, SPlane *planes, int nPlanes, Matrix3& objToView, ShFaceClip& fclip, RenderGlobalContext *rgc, int projType );

void ShFaceClip::ClipMe(Face &f, int nf, int ct, Point3* h,  Point2* b, BOOL *vis) {
	int nv = clip->NumVerts();
	Face newf[8];
	for (int i=0; i<ct-2; i++) {
		newf[i] = f;			
		newf[i].v[0] = nv;
		newf[i].v[1] = nv+i+1;
		newf[i].v[2] = nv+i+2;
		newf[i].flags &= ~VIS_MASK;
		if ((i==0) && vis[0]) 		newf[i].flags |= EDGE_A;
		if (vis[i+1]) 				newf[i].flags |= EDGE_B;
		if ((i==ct-3)&&vis[ct-1]) 	newf[i].flags |= EDGE_C;
		}
	clip->AddFaces(h, ct, newf, ct-2, inst, nf); 
	}


int ShadBufRendImp::Render(
		RendContext &rc, 
		RenderGlobalContext *rgc, 
		float *buf,
		BOOL parl,
		int shadsz,
		float parm,   	   // persp:field-of-view (radians) -- parallel : width in world coords
		float asp,
		float clipD,
		ObjLightDesc *ltd,
		Matrix3 worldToLt
		) {
	RenderInstance *inst;
	SPlane pl[8];
	int facesDone = 0;
	int nRendered = 0;
	int np = 0;	
	// Note the Z buffer is positive=farther, even though the coord system isn't
	closest = BIGFLOAT;
	furthest= -1.0f;

	buffer = buf;
	RGC = rgc;
	parallel = parl;
	shadsize = shadsz;
	param = parm;
	aspect = asp;
	clipDist =clipD;
	ltDesc = ltd;
	worldToLight = worldToLt;

	sz2 = (float)shadsize/2.0f;
	ShClipInstance* clipInst = new ShClipInstance(RGC);
	if (parallel) {
		xscale =  sz2/param;
		yscale = -xscale*aspect;
		}
	else {
		float zfac = -sz2/float(tan(0.5*(double)param));
		xscale = zfac;
		yscale = -zfac*aspect;
		}
	if (parallel) {
		xscale =  sz2/param;
		yscale = -xscale*aspect;
		pl[np++].Set(PLANE_NEG_Z, 0.0f, 0.0f, 0.0f, .0f);	
		pl[np++].Set(PLANE_PX,  xscale, 0.0f, 0.0f, sz2);
		pl[np++].Set(PLANE_PX, -xscale, 0.0f, 0.0f, sz2);
		pl[np++].Set(PLANE_PY, 0.0f,  yscale, 0.0f, sz2);	
		pl[np++].Set(PLANE_PY, 0.0f, -yscale, 0.0f, sz2);
		if (clipDist!=DONT_CLIP)
			pl[np++].Set(PLANE_POS_Z, 0.0f,0.0f,0.0f, clipDist);	
		}	
	else {
		float zfac = -sz2/float(tan(0.5*(double)param));
		xscale = zfac;
		yscale = -zfac*aspect;
		float ycl = yscale/sz2;
		float xcl = xscale/sz2;
		pl[np++].Set(PLANE_NEG_Z, 0.0f,0.0f,0.0f, .001f);	
		pl[np++].Set(PLANE_Y, 0.0f,-ycl, 0.0f);	
		pl[np++].Set(PLANE_Y, 0.0f, ycl, 0.0f);	
		pl[np++].Set(PLANE_X, -xcl, 0.0f, 0.0f);	
	    pl[np++].Set(PLANE_X,  xcl, 0.0f, 0.0f);	
		if (clipDist!=DONT_CLIP)
			pl[np++].Set(PLANE_POS_Z, 0.0f,0.0f,0.0f, clipDist);	
		}

	// count faces for progress bar
	int totFaces = 0;
	int nInst = RGC->NumRenderInstances();
	for (int i=0; i<nInst; i++) {
		inst = RGC->GetRenderInstance(i);
		if ((!inst->Hidden()) && inst->GetINode()->CastShadows() && inst->CastsShadowsFrom(*ltDesc)) {
			totFaces += inst->mesh->numFaces;
			}		
		}

	ClipFlags clipF;

	rc.Progress(facesDone,totFaces); 
	for (i=0; i<nInst; i++) {
		inst = RGC->GetRenderInstance(i);
		if ((!inst->Hidden()) && inst->GetINode()->CastShadows() && inst->CastsShadowsFrom(*ltDesc)) {
			Matrix3 objToLight = inst->objToWorld*worldToLight;
			ShFaceClip faceClip(inst,clipInst);
			Clip(inst, &clipF, pl, np, objToLight, faceClip, RGC, parallel);
			if (!(clipF.culled)) {
				RendShadows(inst, &clipF, objToLight);
				nRendered++;
				}
			facesDone += inst->mesh->numFaces;
			if (!rc.Progress(facesDone,totFaces)) {
				delete clipInst;
				return 0;
				}
			clipF.ClearFlags();
			}
		}
		
	if (!clipInst->IsEmpty()) {
		clipInst->Finish();
		Matrix3 idtm;
		idtm.IdentityMatrix();
		clipF.ClearFlags();
		RendShadows(clipInst, &clipF, idtm, TRUE);
		nRendered++;
		}

	delete clipInst;
#ifdef SHOWSHADMAP
	ShowShadbuf();
#endif
	return nRendered; 
	}



static int GetFlip(RenderInstance *inst, BOOL backface, BOOL force2Side, int matreq) {
	BOOL negParity = inst->TestFlag(INST_TM_NEGPARITY);
	if (backface||negParity) {
		if (backface) {
			BOOL dblSided = ((matreq&MTLREQ_2SIDE)||force2Side)?1:0; 
			if (negParity||dblSided) return 1;
			}
		}
	return 0;
	}

static int nextpt[3] = {1,2,0};
static int prevpt[3] = {2,0,1};

int ShadBufRendImp::RendShadows(RenderInstance *inst, ClipFlags *cf, Matrix3& objToLight, BOOL isClip) {
	int i,ystart,yend,flip;
	int face;
	Face *f,*flist;
	Point3 pv[3];
	Point2 spt[3], wpt[3];
	Plane theplane;
	Point3 *pos,p;
	Point2 *scrPos;
	RenderInstance *origInst;
	int npoints=0;

	Mesh *mesh = inst->mesh;
	pos = new Point3[mesh->numVerts];
	scrPos = new Point2[mesh->numVerts];

	// transform to light coords:
	for ( i= 0; i<mesh->numVerts; i++) {
		p = pos[i] = objToLight*mesh->verts[i];
		scrPos[i] = Proj2D(p);
		}

	flist = mesh->faces;

	ShClipInstance *clip = NULL;
	if (isClip) 
		clip = (ShClipInstance*)inst; 
	for (face=0; face<mesh->numFaces; ++face)	{
		if (cf->clipped[face])
			continue;

		f = &flist[face];

		origInst = isClip?clip->GetOrigInst(face):inst;
		curMtl = origInst->mtl;
		matreq = origInst->MtlRequirements(f->getMatID());

		pv[0] = pos[f->v[0]];
		pv[1] = pos[f->v[1]];
		pv[2] = pos[f->v[2]];

  		spt[0] = scrPos[f->v[0]];
		spt[1] = scrPos[f->v[1]];
		spt[2] = scrPos[f->v[2]];
		 
		flip = 0;
	
		BOOL backface = (tnormal(spt[0],spt[1],spt[2])>=0)?1:0;
		flip = GetFlip(inst, backface, RGC->force2Side, matreq);

		CompPlaneEqn(theplane, pv[0], pv[1], pv[2]);
		theplane[0] /= xscale; // scale to work with screen x 
		theplane[1] /= yscale; // scale to work with screen y 

		float fminy = MIN3(spt[0].y,spt[1].y,spt[2].y);
		float fmaxy = MAX3(spt[0].y,spt[1].y,spt[2].y);

		ystart = int(fminy-.5f);
		yend = int (fmaxy+.5);
		if (ystart<0) ystart = 0;
		if (yend>shadsize) yend = shadsize;

		if  (RGC->wireMode||(matreq&MTLREQ_WIRE)) {
			Point2 q[3];
			int j;
			int edgvis[3];
			edgvis[0] = (f->flags&EDGE_A)?1:0;
			edgvis[1] = (f->flags&EDGE_B)?1:0;
			edgvis[2] = (f->flags&EDGE_C)?1:0;
			ReadyShadWireTriangle(pv,spt,wpt,edgvis);
			for (i=0; i<3; i++) {
				if (edgvis[i]) {
					j = nextpt[i];
					q[0] = spt[i];	q[1] = spt[j];	q[2] = wpt[j];
					ScanFace(q,ystart,yend,theplane,flip);
					q[0] = spt[i];	q[1] = wpt[j];	q[2] = wpt[i];
					ScanFace(q,ystart,yend,theplane, flip);
					}
				}
			}
		else 
			ScanFace(spt,ystart,yend,theplane,flip);
		}
	delete [] pos;
	delete [] scrPos;	
	return(1);
	}


void ShadBufRendImp::ScanFace(Point2 *pts, int ystart, int yend, Plane &theplane, int flip) {
	int l1,r1,l2,r2,ix,y;
	int scanmax,scanmin;
	float xl,xr,xs,denom,z;
	Point2 vl1,vl2,vr1,vr2;
	float *pz,fy,*zb;

	for (y=ystart; y<yend; y++)  {
		pz = &buffer[y*shadsize];
		fy = (float)y+0.5f;

		if (pts[0].y>fy) {
			if (pts[1].y>fy) {
				if (pts[2].y>fy) continue;
				l1 = r1 = 2;
				if (flip) { l2 = 1; r2 = 0; }
				else { l2 = 0; r2 = 1; }
				}
			else  {
				if (pts[2].y>fy) { 
					l1 = r1 = 1;	
					if (flip) {	l2 = 0; r2 = 2; }
					else { l2 = 2; r2 = 0; }
					}
				else 	{ 	
					if (flip) {	l1 = 1; r1 = 2; }
					else { l1 = 2; r1 = 1; }
					l2 = r2 = 0;	
					}
				}
			}
		else {
			if (pts[1].y<fy) {   /* -a -b +c */
				if (pts[2].y<fy) return; /* done with this triangle */
				if (flip) {	l1 = 0; r1 = 1; }
				else { l1 = 1; r1 = 0; }
				l2 = r2 = 2;
				}
			else {
				if (pts[2].y<fy)  {  
					if (flip) {	l1 = 2; r1 = 0; }
					else { l1 = 0; r1 = 2; }
					l2 =  r2 = 1;	
					}
				else 	{ 
					l1 = r1 = 0; 
					if (flip) {	l2 = 2; r2 = 1; }
					else { l2 = 1; r2 = 2; }
					}
				}
			}
	
		vl1 = pts[l1];  vr1 = pts[r1]; 
		vl2 = pts[l2];  vr2 = pts[r2];				
	
		float dy;
		if ((dy=vl2.y-vl1.y)!=0.0f)
			xl = vl1.x + ((fy-vl1.y)/dy)*(vl2.x-vl1.x);
		else xl = vl1.x;
		if ((dy=vr2.y-vr1.y)!=0.0f)
			xr = vr1.x + ((fy-vr1.y)/dy)*(vr2.x-vr1.x);
		else xr = vr1.x;
	
		if (xl>=xr) continue;
	
		scanmin = (int)(xl+0.5);
		scanmax = (int)(xr-0.5);
	
		if (scanmin<0) scanmin = 0;
		if (scanmax>shadsize-1) scanmax = shadsize-1;
		if (scanmax<scanmin) continue;
	
		zb = &pz[scanmin];
		xs = (float)scanmin+0.5f;
	
		if (parallel) {
			z = +(theplane[0]*(xs-sz2) + theplane[1]*(fy-sz2)+theplane[3])/theplane[2];
			float dz = +theplane[0]/theplane[2];
			for (ix = scanmin; ix<=scanmax; ix++,zb++) {
				if (z<*zb) *zb = z;
				if (z<closest)	closest = z;
				if (z>furthest) furthest = z;
				z += dz;
				}
			}
		else {
			denom = theplane[0]*(xs-sz2) + theplane[1]*(fy-sz2)+theplane[2];
			for (ix = scanmin; ix<=scanmax; ix++,zb++) {
				z = theplane[3]/denom;  // ZSIGN reversed
				denom += theplane[0];
				if (z<*zb) *zb = z;
				if (z<closest)	closest = z;
				if (z>furthest) furthest = z;
				}
			}
		}
	}


/* move point p towards q fraction f of the way */
static inline void  MovePtTowardsPt(Point2& p, const Point2& q, float f) { p += f*(q-p); }
static inline void Move3DPtTowardsPt(Point3& p, const Point3& q, float f) { p += f*(q-p); }

inline Point2 ShadBufRendImp::Proj2D(Point3 p) {
	if (parallel) return(Point2(sz2+p.x*xscale, sz2+p.y*yscale));
	else return(Point2(sz2+p.x/p.z*xscale, sz2+p.y/p.z*yscale));
	}

void ShadBufRendImp::ReadyShadWireTriangle(Point3 *pv, Point2*spt, Point2 *wspt, int *vis) {
	int in,i,ip;
	float d,f;
	Point3 wpt[3];
	if (!RGC->wireMode&&(matreq&MTLREQ_WIRE_ABS)) {
		float wsz = curMtl->WireSize();
		wpt[0] = pv[0];
		wpt[1] = pv[1];
		wpt[2] = pv[2];
		for (i=0; i<3; i++) {
			if (vis[i]) {
				in = nextpt[i];
				ip = prevpt[i];
				/* distance of point ip from line (i,in) */
				d = Dist3DPtToLine(&wpt[i],&wpt[in],&wpt[ip]);
				f = wsz/(2.0f*d);
				if (f>1.0) f=1.0f;			 
				/* move pts "i" and "in" toward point ip */
				Move3DPtTowardsPt(wpt[i],wpt[ip],f);
				Move3DPtTowardsPt(wpt[in],wpt[ip],f);
				}
			}
		wspt[0] = Proj2D(wpt[0]);
		wspt[1] = Proj2D(wpt[1]);
		wspt[2] = Proj2D(wpt[2]);
		}
	else {
		float wsz = (RGC->wireMode)?RGC->wire_thick: curMtl->WireSize();

		wsz *= 0.5f;
		wspt[0] = spt[0];	 
		wspt[1] = spt[1];  
		wspt[2] = spt[2];
		for (i=0; i<3; i++) {
			if (vis[i]) {
				in = nextpt[i];
				ip = prevpt[i];
				/* distance of point ip from line (i,in) */
				d = DistPtToLine(&wspt[i],&wspt[in],&wspt[ip]);
				f = wsz/d;
				if (f>1.0f) f=1.0f;			 
				/* move pts "i" and "in" toward point ip */
				MovePtTowardsPt(wspt[i],wspt[ip],f);
				MovePtTowardsPt(wspt[in],wspt[ip],f);
				}
			}
		}
	}


static void clipSeg(SPlane& plane, Point3& beg, Point3& end, Point3& mid, Point2& bbeg, Point2& bend, Point2& bmid)
	{ 
    float	d1, d2; 
    float 	s, t;
	d1 =  DistToPlane(beg,plane);
	d2 =  DistToPlane(end,plane);
    s = d2 / (d2 - d1);
    t = 1.0f - s;
	mid = s*beg + t*end;
	bmid = s*bbeg + t*bend;
	}


static Point2 defbar0(0.0f,0.0f);
static Point2 defbar1(1.0f,0.0f);
static Point2 defbar2(0.0f,1.0f);

//-----------------------------------------------------------------------
// A general clipper
//  "planes" is an array of nPlanes clipping planes in the camera coord system.
// This allows for clipping against hither, yon, and mirror, or other combos
// in one pass
//-----------------------------------------------------------------------
void Clip(RenderInstance *inst, ClipFlags *clipFlags, SPlane *planes, int nPlanes, Matrix3& objToView, ShFaceClip& fclip, RenderGlobalContext *rgc, int projType ) {
	Point3 h[10], clpH[10];	 // points in cam coords
	Point2 b[10], clpB[10];  // barycentric coords of clipped points in orig triangle
	BOOL vis[10],clpVis[10]; // edge visibility
	WORD cf[10], orCf, andCf;
	BOOL anyIn = 0;
	int nClip;

	if (inst->IsClip())  // don't clip the clip object
		return;

	clipFlags->ClearFlags();

	Mesh *mesh = inst->mesh;
	if (mesh==NULL) 
		return;

	clipFlags->numFaces = mesh->numFaces;

	BOOL negParity = inst->TestFlag(INST_TM_NEGPARITY);

	for (int nf = 0; nf<mesh->numFaces; nf++) {
		Face &f = mesh->faces[nf];
		int ct = 3;
		orCf = 0;
	    andCf = 0xffff;
		for (int i =0; i<3; i++) {
			h[i] =  objToView*mesh->verts[f.v[i]];
	        cf[i]  = CompClipFlags(h[i],planes,nPlanes);
			orCf  |= cf[i];
	        andCf &= cf[i];
			vis[i] = (f.flags & (VIS_BIT<<i))?1:0;
			}
		if (orCf==0) {
			// No points outside of clip volume: not clipped
			anyIn = TRUE;
			continue;
			}
		
		// OK, this face is clipped all or partially:
		clipFlags->SetClip(nf);
		
	    if (andCf) {
			// all points outside single side: trivial reject
			continue;
			}

		BOOL backface = !isFacing(h[0],h[1],h[2],projType);
		if (backface||negParity) {
			ULONG matreq = inst->MtlRequirements(f.getMatID());
			BOOL dblSided =  ((matreq&MTLREQ_2SIDE)||rgc->force2Side)?1:0; 
			if (backface) {
				if (!(dblSided||negParity)) 
					continue;
				}
			else {
				if (negParity&&!dblSided) 
					continue;
				}
			}

		// initialize bary coords
		b[0] = defbar0; b[1] = defbar1; b[2] = defbar2;
	
		// at least one edge is clipped: do clipping

		// put in extra point since segment clipper doesn't fill in end point
		// make copy of first point
		h[ct] = h[0];
		cf[ct] = cf[0];
		b[ct] = b[0]; 
		ct++;
	    for (int planeCt = 0; planeCt < nPlanes; planeCt++) {
	        int plane = 1 << planeCt;
			// check if any edges are clipped by this plane
	        if ( !(orCf & plane))
		    	continue;
		    nClip = 0;

		    for (i = 1; i < ct; i++)  {
				if ((cf[i-1] & plane) == 0) {
					// first endpoint is visible, so copy it to clpHomoPts...
					clpH[nClip] = h[i-1];
					clpB[nClip] = b[i-1];
					clpVis[nClip] = vis[i-1];
					nClip++;
			    	if (cf[i] & plane) {
						//  second endpoint is invisible: we must clip this segment
						//  this is the only place we add an extra point!!!
						clipSeg(planes[planeCt], h[i-1], h[i], clpH[nClip], 
									 b[i-1], b[i], clpB[nClip]);
						clpVis[nClip] = 0;
						nClip++;
			    		}
		    		}
				// first endpoint is invisible...
				else if ((cf[i] & plane) == 0) {
					// second endpoint is visible: we must clip this segment
					clipSeg(planes[planeCt], h[i-1], h[i], clpH[nClip], 
							 b[i-1], b[i], clpB[nClip]);
					clpVis[nClip] = vis[i-1];
		    		nClip++;
					}
				// else both endpoints are invisible: there's nothing to do!
		   		} // next point

		   	// copy newly clipped points to homoPts/exPrms, so we can repeat
		   	orCf = 0;
		   	for (i = 0; i < nClip; i++) {
				h[i] = clpH[i];
				b[i] = clpB[i];
				vis[i] = clpVis[i];
				// get clip flags for next plane ready...
		        cf[i]  = CompClipFlags(clpH[i],planes,nPlanes);
				orCf |= cf[i];
		 		}
			// put in extra point for next round
			// copy last point
			h[nClip] = clpH[0];
			b[nClip] = clpB[0];
			cf[nClip] = cf[0];
		  	ct = nClip+1;
	    	} // next plane
		
		ct--;   // get rid of extra point 
		if (ct<3) 
			continue;
		fclip.ClipMe(f,nf,ct,h,b,vis);
		}
	if (!anyIn) 
		clipFlags->culled = 1;
	}


#ifdef SHOWSHADMAP
/* Display shadow map */
void ShadBufRendImp::ShowShadbuf() {
	int ypos,ix;
	float *z;
	if (buffer==NULL) return;

	BitmapInfo bi;
	bi.SetName(_T(""));
	bi.SetWidth(shadsize);
	bi.SetHeight(shadsize);
	bi.SetType(BMM_TRUE_32);
	Bitmap *testBM;
	testBM = TheManager->Create(&bi);

	PixelBuf l64(shadsize);
	if (fabs(furthest-closest)<.0001f) furthest += 1.0f;
	for(ypos=0; ypos<shadsize; ypos++) {
		BMM_Color_64 *pb=l64.Ptr();
		z = &buffer[ypos*shadsize];
		for(ix=0; ix<shadsize; ++ix) {
			if(z[ix]>=0.0f) {
				float f;
				if (furthest==closest) f = 0.5f;
				else f =(z[ix]-closest)/(furthest-closest); 
				pb[ix].r = pb[ix].g = pb[ix].b = (unsigned short)(f*65535.0f);
				}
			else {
				pb[ix].r = 250<<8;
				pb[ix].g = 120<<8;
				pb[ix].b = 0;
				}
			pb[ix].a = 0xffff;
			}
		testBM->PutPixels(0,ypos, shadsize,l64.Ptr());
		}
	testBM->Display(_T("Shadow Test"), BMM_UR);
	MessageBox(NULL, _T("Howdy"), _T(" Shadow Buffer Test"), MB_OK|MB_ICONEXCLAMATION|MB_TASKMODAL);
	testBM->DeleteThis();
	}
#endif

