/**********************************************************************
 *<
	FILE: light.cpp

	DESCRIPTION:  Omni light implementation

	CREATED BY: Dan Silva

	HISTORY: Created 2 January 1995.
			 Almost total rewrite, Don Brittain, Oct 95.
			 Added attenuation to dir light, target direct light,
			    near attenuation ranges, new atten range display, 
			    shadowing omni lights, etc-- Dan Silva, Jan 1997.

 *>	Copyright (c) 1995, All Rights Reserved.
 **********************************************************************/

#include "prim.h"
#include "target.h"
#include "hsv.h"
#include "genlight.h"
#include "light.h"
#include "render.h"
#include "max.h"
#include "shadbuf.h"
#include "shadbuf.h"
#include <bmmlib.h>

#define OMNI_MESH		0
#define DIR_MESH		1

#define NUM_HALF_ARC	5
#define NUM_ARC_PTS	    (2*NUM_HALF_ARC+1)
#define NUM_CIRC_PTS	28
#define SEG_INDEX		7


// Parameter block indices
#define PB_COLOR		0	   
#define PB_INTENSITY	1
#define PB_CONTRAST		2
#define PB_DIFFSOFT		3
#define PB_HOTSIZE		4   //PB_OMNIMAP_SIZE  2 
#define PB_FALLSIZE		5	//PB_OMNIMAP_BIAS  3
#define PB_ASPECT		6	//PB_OMNIMAP_RANGE 4 
#define PB_MAP_SIZE		7	//PB_OMNIRAY_BIAS  5
#define PB_MAP_BIAS 	8	//PB_OMNIATSTART1  6
#define PB_MAP_RANGE	9	//PB_OMNIATEND1    7
#define PB_RAY_BIAS		10	//PB_OMNIATSTART   8
#define PB_ATTENSTART1	11	//PB_OMNIATNEND    9
#define PB_ATTENEND1 	12	
#define PB_ATTENSTART	13
#define PB_ATTENEND		14
#define PB_TDIST		15


//#define PB_RADIUS		 16
//#define PB_NUMBER_LIGHTS 17

// indices for OMNI only
#define PB_OMNIMAP_SIZE    4 
#define PB_OMNIMAP_BIAS    5
#define PB_OMNIMAP_RANGE   6 
#define PB_OMNIRAY_BIAS    7
#define PB_OMNIATSTART1    8
#define PB_OMNIATEND1      9
#define PB_OMNIATSTART     10
#define PB_OMNIATEND       11

#define DEF_TDIST		240.0f // 160.0f


static int waitPostLoad = 0;

#if 0
#define FLto255i(x)	((int)((x)*255.0f + 0.5f))
#define FLto255f(x)	((float)((x)*255.0f + 0.5f))
#else
#define FLto255i(x)	((int)((x)*255.0f))
#define FLto255f(x)	((float)((x)*255.0f))
#endif


inline float MaxF(float a, float b) { return a>b?a:b; }
inline float MinF(float a, float b) { return a<b?a:b; }

inline float MaxAbs(float x, float y, float z) {
	float ax = (float)fabs(x);
	float ay = (float)fabs(y);
	float az = (float)fabs(z);
	return ax>ay? (ax?az:ax):ay>az?ay:az;
	}

// in prim.cpp  - The dll instance handle
extern HINSTANCE hInstance;
void resetLightParams();

class STSpotLightClassDesc:public ClassDesc {
	public:
	int 			IsPublic() { return 1; }
	void *			Create(BOOL loading = FALSE) { return new GeneralLight(TSPOT_LIGHT); }
	int 			BeginCreate(Interface *i);
	int 			EndCreate(Interface *i);
	const TCHAR *	ClassName() { return GetString(IDS_DB_TARGET_SPOT_CLASS); }
	SClass_ID		SuperClassID() { return LIGHT_CLASS_ID; }
	Class_ID		ClassID() { return Class_ID(SOFTSPOT_CLASS_ID); }
	const TCHAR* 	Category() { return _T("");  }
	void			ResetClassParams(BOOL fileReset) { if(fileReset) resetLightParams(); }
	// Class IO
	BOOL			NeedsToSave() { return TRUE; }
	IOResult 		Save(ISave *isave);
	IOResult 		Load(ILoad *iload);
	
};

static STSpotLightClassDesc StspotLightDesc;

ClassDesc *GetSTSpotLightDesc() { return &StspotLightDesc; }



// Class variables of GeneralLight
Mesh GeneralLight::staticMesh[2];
short GeneralLight::meshBuilt = 0;
int GeneralLight::dlgRed = 180;
int GeneralLight::dlgGreen = 180;
int GeneralLight::dlgBlue = 180;
int GeneralLight::dlgH = 0;
int GeneralLight::dlgS = 0;
int GeneralLight::dlgV = 180;
float GeneralLight::dlgIntensity = 1.0f;
float GeneralLight::dlgContrast = 0.0f;
float GeneralLight::dlgDiffsoft = 0.0f;
float GeneralLight::dlgHotsize = 43.0f;
float GeneralLight::dlgFallsize = 45.0f;
float GeneralLight::dlgTDist = DEF_TDIST;
float GeneralLight::dlgAtten1Start = 0.0f;
float GeneralLight::dlgAtten1End = 40.0f; 
float GeneralLight::dlgAttenStart = 80.0f;
float GeneralLight::dlgAttenEnd = 200.0f; //320.0f;
short GeneralLight::dlgShowCone = 0;
short GeneralLight::dlgUseAtten = 0;
short GeneralLight::dlgShowAtten = 0;
short GeneralLight::dlgShape = CIRCLE_LIGHT;
float GeneralLight::dlgAspect = 10.0f;
float GeneralLight::dlgMapBias = 4.0f;
float GeneralLight::dlgMapRange = 4.0f;
int GeneralLight::dlgMapSize = 256;
float GeneralLight::dlgRayBias = 1.0f;


//float GeneralLight::dlgRadius = 20.0f;
//int GeneralLight::dlgNumberLights = 20;

// Global shadow settings
short GeneralLight::globShadowType = 0;  // 0: ShadowMap   1: RayTrace
short GeneralLight::globAbsMapBias = 1;
float GeneralLight::globMapRange = 4.0f;
float GeneralLight::globMapBias = 4.0f;
int   GeneralLight::globMapSize = 256;
float GeneralLight::globRayBias = 0.2f;

HWND GeneralLight::hGeneralLight = NULL;
HWND GeneralLight::hSpotLight = NULL;
HWND GeneralLight::hOmni = NULL;
HWND GeneralLight::hShadow = NULL;

void resetLightParams()
{
	GeneralLight::dlgRed = 180;
	GeneralLight::dlgGreen = 180;
	GeneralLight::dlgBlue = 180;
	GeneralLight::dlgH = 0;
	GeneralLight::dlgS = 0;
	GeneralLight::dlgV = 180;
	GeneralLight::dlgIntensity = 1.0f;
	GeneralLight::dlgContrast = 0.0f;
	GeneralLight::dlgDiffsoft = 0.0f;
	GeneralLight::dlgHotsize = 43.0f;
	GeneralLight::dlgFallsize = 45.0f;
	GeneralLight::dlgTDist = DEF_TDIST;
	GeneralLight::dlgAtten1Start = 0.0f;
	GeneralLight::dlgAtten1End = 40.0f;
	GeneralLight::dlgAttenStart = 80.0f;
	GeneralLight::dlgAttenEnd = 200.0f;
	GeneralLight::dlgShowCone = 0;
	GeneralLight::dlgUseAtten = 0;
	GeneralLight::dlgShowAtten = 0;
	GeneralLight::dlgShape = CIRCLE_LIGHT;
	GeneralLight::dlgAspect = 1.0f;
	GeneralLight::dlgMapBias = 4.0f;
	GeneralLight::dlgMapRange = 4.0f;
	GeneralLight::dlgMapSize = 256;
	GeneralLight::dlgRayBias = 0.2f;

	// Global shadow settings
	GeneralLight::globShadowType = 0;  // 0: ShadowMap   1: RayTrace
	GeneralLight::globAbsMapBias = 1;
	GeneralLight::globMapRange = 4.0f;
	GeneralLight::globMapBias = 4.0f;
	GeneralLight::globMapSize = 256;
	GeneralLight::globRayBias = 0.2f;
	HWND hl = GetDlgItem(GeneralLight::hGeneralLight, IDC_LIGHT_COLOR);
	if(hl)
		InvalidateRect(hl, NULL, TRUE);
}

IObjParam *GeneralLight::iObjParams;
ISpinnerControl *GeneralLight::redSpin;
ISpinnerControl *GeneralLight::greenSpin;
ISpinnerControl *GeneralLight::blueSpin;
ISpinnerControl *GeneralLight::hSpin;
ISpinnerControl *GeneralLight::sSpin;
ISpinnerControl *GeneralLight::vSpin;
ISpinnerControl *GeneralLight::intensitySpin;
ISpinnerControl *GeneralLight::contrastSpin;
ISpinnerControl *GeneralLight::diffsoftSpin;
ISpinnerControl *GeneralLight::hotsizeSpin;
ISpinnerControl *GeneralLight::fallsizeSpin;
ISpinnerControl *GeneralLight::tDistSpin;
ISpinnerControl *GeneralLight::attenStartSpin;
ISpinnerControl *GeneralLight::attenEndSpin;
ISpinnerControl *GeneralLight::attenStart1Spin;
ISpinnerControl *GeneralLight::attenEnd1Spin;
ISpinnerControl *GeneralLight::aspectSpin;
ISpinnerControl *GeneralLight::mapBiasSpin;
ISpinnerControl *GeneralLight::mapSizeSpin;
ISpinnerControl *GeneralLight::mapRangeSpin;
ISpinnerControl *GeneralLight::rayBiasSpin;
IColorSwatch *GeneralLight::colorSwatch;
ICustButton *GeneralLight::projMapName;
ProjMapDADMgr GeneralLight::projDADMgr;


ISpinnerControl *GeneralLight::radiusSpin;
ISpinnerControl *GeneralLight::numberLightsSpin;


//------------------------------------------------------------------------------
//
// Projection Map Drag and Drop manager;
//
//------------------------------------------------------------------------------
SClass_ID ProjMapDADMgr::GetDragType(HWND hwnd, POINT p) {
	return TEXMAP_CLASS_ID;
	}

BOOL ProjMapDADMgr::OkToDrop(ReferenceTarget *dropThis, HWND hfrom, HWND hto, POINT p, SClass_ID type, BOOL isNew) {
	if (hfrom==hto) return FALSE;
	return (type==TEXMAP_CLASS_ID)?1:0;
	}

ReferenceTarget *ProjMapDADMgr::GetInstance(HWND hwnd, POINT p, SClass_ID type) {
	return gl->projMap;
	}

void ProjMapDADMgr::Drop(ReferenceTarget *dropThis, HWND hwnd, POINT p, SClass_ID type) {
	Texmap *m = (Texmap *)dropThis;
	gl->AssignProjectorMap(m, FALSE);
	}

// extern void* CreateGeneralLight() { return new GeneralLight; }

// This generic snap routine is used to snap to the 0,0,0 point of the given node.  For lights,
// this works to snap all types.

static void GenericSnap(TimeValue t, INode* inode, SnapInfo *snap, IPoint2 *p, ViewExp *vpt) 
{
	// Make sure the vertex priority is active and at least as important as the best snap so far
	if(snap->vertPriority > 0 && snap->vertPriority <= snap->priority) {
		Matrix3 tm = inode->GetObjectTM(t);	
		GraphicsWindow *gw = vpt->getGW();	
   	
		gw->setTransform(tm);

		Point2 fp = Point2((float)p->x, (float)p->y);
		IPoint3 screen3;
		Point2 screen2;
		Point3 vert(0.0f,0.0f,0.0f);

		gw->wTransPoint(&vert,&screen3);

		screen2.x = (float)screen3.x;
		screen2.y = (float)screen3.y;

		// Are we within the snap radius?
		int len = (int)Length(screen2 - fp);
		if(len <= snap->strength) {
			// Is this priority better than the best so far?
			if(snap->vertPriority < snap->priority) {
				snap->priority = snap->vertPriority;
				snap->bestWorld = vert * tm;
				snap->bestScreen = screen2;
				snap->bestDist = len;
			}
			// Closer than the best of this priority?
			else if(len < snap->bestDist) {
				snap->priority = snap->vertPriority;
				snap->bestWorld = vert * tm;
				snap->bestScreen = screen2;
				snap->bestDist = len;
			}
		}
	}
}

static void InitColorSpinner( HWND hDlg, ISpinnerControl *spin, int editID, float v) 
{
	spin->SetLimits(0.0f, 255.0f, FALSE );
	spin->SetValue(v,FALSE);
	spin->SetScale( 1.0f );
	spin->LinkToEdit( GetDlgItem(hDlg,editID), EDITTYPE_INT );
}

static void InitAngleSpinner( HWND hDlg, ISpinnerControl *spin, int editID, float v)
{
	spin->SetLimits(0.5f, 179.5f, FALSE);
	spin->SetValue(v,FALSE);
	spin->SetScale(.1f);
	spin->LinkToEdit( GetDlgItem(hDlg, editID), EDITTYPE_FLOAT );
}

static void InitRangeSpinner( HWND hDlg, ISpinnerControl *spin, int editID, float v, BOOL allowNegative)
{
//	spin->SetLimits(allowNegative?-99999.0f:0.0f, 99999.0f, FALSE);
	spin->SetLimits(0.0f, 99999.0f, FALSE);
	spin->SetValue(v,FALSE);
	spin->SetScale(1.0f);
	spin->LinkToEdit( GetDlgItem(hDlg, editID), EDITTYPE_UNIVERSE );
}

#define RELEASE_SPIN(x)   if (gl->x) { ReleaseISpinner(gl->x); gl->x = NULL;}
#define RELEASE_COLSW(x)   if (gl->x) { ReleaseIColorSwatch(gl->x); gl->x = NULL;}
#define RELEASE_BUT(x)   if (gl->x) { ReleaseICustButton(gl->x); gl->x = NULL;}


static int AttenFromID(int id) {
	switch (id) {
		case IDC_START_RANGE1_SPIN: return ATTEN1_START;
		case IDC_END_RANGE1_SPIN:   return ATTEN1_END;
		case IDC_START_RANGE_SPIN:  return ATTEN_START;
		case IDC_END_RANGE_SPIN:    return ATTEN_END;
		default: return 0;
		}
	}

ISpinnerControl *GeneralLight::AttenSpinner(int atype) {
	switch (atype) {
		case ATTEN1_START: return attenStart1Spin; break;
		case ATTEN1_END:   return attenEnd1Spin;   break;
		case ATTEN_START:  return attenStartSpin;  break;
		case ATTEN_END:    return attenEndSpin;    break;
		default: return attenStartSpin;
		}
	}

void GeneralLight::ChangeAtten(int atype) {
	float val = AttenSpinner(atype)->GetFVal();
	ISpinnerControl *isp;
	TimeValue t = iObjParams->GetTime();
	SetAtten(t, atype, val);	
	
	for (int i = atype+1; i<4; i++) {
		isp = AttenSpinner(i);
		float f = isp->GetFVal();
		if (val>f) {
			isp->SetValue(val,FALSE);
			SetAtten(t, i, val);	
			}
		}
	for (i = atype-1; i>=0; i--) {
		isp = AttenSpinner(i);
		float f = isp->GetFVal();
		if (val<f) {
			isp->SetValue(val,FALSE);
			SetAtten(t, i, val);	
			}
		}
	}

void GeneralLight::SetSpinnerBracket(ISpinnerControl *spin, int pbid) {
	spin->SetKeyBrackets(pblock->KeyFrameAtTime(pbid,iObjParams->GetTime()));
	}

BOOL CALLBACK GeneralLightParamDialogProc(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam )
{	
	int id;
	GeneralLight *gl = (GeneralLight *)GetWindowLong( hDlg, GWL_USERDATA );
	if ( !gl && message != WM_INITDIALOG ) 
		return FALSE;
	TimeValue t = gl->iObjParams->GetTime();
	float tmpHot, tmpFall;

	switch ( message ) {
	case WM_INITDIALOG:
		gl = (GeneralLight *)lParam;
	
		SetWindowLong( hDlg, GWL_USERDATA, (LONG)gl );
		SetDlgFont( hDlg, gl->iObjParams->GetAppHFont() );
		CheckDlgButton( hDlg, IDC_LIGHT_ON, gl->GetUseLight() );
		CheckDlgButton( hDlg, IDC_AFFECT_DIFFUSE, gl->affectDiffuse);
		CheckDlgButton( hDlg, IDC_SOFTEN_DIFFUSE, gl->softenDiffuse);
		CheckDlgButton( hDlg, IDC_AFFECT_SPECULAR, gl->affectSpecular);

		CheckDlgButton( hDlg, IDC_SHOW_CONE, gl->GetConeDisplay() );
		CheckDlgButton( hDlg, IDC_SHOW_RANGES, gl->GetAttenDisplay() );
		CheckDlgButton( hDlg, IDC_SHOW_RANGES1, gl->GetAttenNearDisplay() );
		CheckDlgButton( hDlg, IDC_USE_ATTEN, gl->GetUseAtten() );
		CheckDlgButton( hDlg, IDC_USE_ATTEN1, gl->GetUseAttenNear() );
		CheckRadioButton( hDlg, IDC_RECT_LIGHT, IDC_CIRCLE_LIGHT, IDC_RECT_LIGHT+gl->GetSpotShape());

		CheckRadioButton( hDlg, IDC_ATT_NONE, IDC_ATT_INVSQ, IDC_ATT_NONE+gl->GetDecayType());

		CheckDlgButton( hDlg, IDC_GLOBAL_SET, gl->GetUseGlobal() );
		CheckDlgButton( hDlg, IDC_CAST_SHADOWS, gl->GetShadow() );

		CheckRadioButton( hDlg, IDC_SHADOW_MAPS, IDC_RAY_TRACED_SHADOWS, 
			IDC_SHADOW_MAPS + gl->GetShadowType());
		CheckDlgButton( hDlg, IDC_ABS_MAP_BIAS, gl->GetAbsMapBias() );

//		EnableWindow(GetDlgItem(hDlg, IDC_LASPECT), !gl->GetSpotShape());
//		EnableWindow(GetDlgItem(hDlg, IDC_LASPECT_SPIN), !gl->GetSpotShape());
		EnableWindow(GetDlgItem(hDlg, IDC_BITMAP_FIT), !gl->GetSpotShape());
		SetDlgItemText(hDlg,IDC_EXCLUDE_DLG, 
			GetString(gl->Include()?IDS_DS_INCLUDE__: IDS_DS_EXCLUDE__));

		CheckDlgButton( hDlg, IDC_OVERSHOOT, gl->GetOvershoot() );
		CheckDlgButton( hDlg, IDC_PROJECTOR, gl->GetProjector() );

//		gl->FixHotFallConstraint();

		if (gl->type==TSPOT_LIGHT || gl->type==TDIR_LIGHT) {
			TCHAR buf[40];
			_stprintf(buf,_T("%0.3f"),gl->targDist);
			SetWindowText(GetDlgItem(hDlg,IDC_TARG_DISTANCE),buf);
			}
		return FALSE;

	case WM_DESTROY:
		RELEASE_SPIN( redSpin);
		RELEASE_SPIN( greenSpin );
		RELEASE_SPIN( blueSpin );
		RELEASE_SPIN( hSpin );
		RELEASE_SPIN( sSpin );
		RELEASE_SPIN( vSpin );
		RELEASE_SPIN( intensitySpin );
		RELEASE_SPIN( contrastSpin );
		RELEASE_SPIN( diffsoftSpin );
		RELEASE_COLSW( colorSwatch);
		RELEASE_SPIN( hotsizeSpin );
		RELEASE_SPIN( fallsizeSpin );
		RELEASE_SPIN( tDistSpin );
		RELEASE_SPIN( aspectSpin );
		RELEASE_SPIN( mapSizeSpin );
		RELEASE_SPIN( mapBiasSpin );
		RELEASE_SPIN( mapRangeSpin );
		RELEASE_SPIN( rayBiasSpin );
		RELEASE_SPIN( attenStartSpin); 
		RELEASE_SPIN( attenEndSpin); 
		RELEASE_SPIN( attenStart1Spin); 
		RELEASE_SPIN( attenEnd1Spin); 
		RELEASE_BUT( projMapName); 

		RELEASE_SPIN( radiusSpin );
		RELEASE_SPIN( numberLightsSpin );
		

		return FALSE;

	case CC_SPINNER_CHANGE:			
		{
		int id = LOWORD(wParam);
		if (!theHold.Holding()) theHold.Begin();
		BOOL redraw = TRUE;
		switch ( id ) {
		case IDC_LREDSPINNER:
		case IDC_LGREENSPINNER:
		case IDC_LBLUESPINNER:
			gl->SetRGBColor(t, Point3(gl->redSpin->GetFVal()/255.0f, 
					gl->greenSpin->GetFVal()/255.0f, gl->blueSpin->GetFVal()/255.0f ));
			gl->UpdateColBrackets(gl->iObjParams->GetTime());
			break;
		case IDC_LHSPINNER:
		case IDC_LSSPINNER:
		case IDC_LVSPINNER:
			gl->SetUpdateHSVSpin(FALSE);
			gl->SetHSVColor(t, Point3(gl->hSpin->GetFVal()/255.0, 
					gl->sSpin->GetFVal()/255.0, gl->vSpin->GetFVal()/255.0 ));
			gl->UpdateColBrackets(gl->iObjParams->GetTime());
			break;
		case IDC_LMULTSPINNER:
			gl->SetIntensity(t, gl->intensitySpin->GetFVal());
			gl->SetSpinnerBracket(gl->intensitySpin,PB_INTENSITY);
			break;
		case IDC_LCONTRASTSPIN2:
			gl->SetContrast(t, gl->contrastSpin->GetFVal());
			gl->SetSpinnerBracket(gl->contrastSpin,PB_CONTRAST);
			break;
		case IDC_DIFFSOFTSPIN:
			gl->SetDiffuseSoft(t, gl->diffsoftSpin->GetFVal());
			gl->SetSpinnerBracket(gl->diffsoftSpin,PB_DIFFSOFT);
			break;

		case IDC_LHOTSIZESPINNER:
		case IDC_LFALLOFFSPINNER:
			tmpHot = gl->hotsizeSpin->GetFVal();
			tmpFall = gl->fallsizeSpin->GetFVal();

			if((tmpHot > tmpFall - gl->iObjParams->GetLightConeConstraint())) {
				if(LOWORD(wParam) == IDC_LHOTSIZESPINNER)
					tmpFall = tmpHot + gl->iObjParams->GetLightConeConstraint();
				else
					tmpHot = tmpFall - gl->iObjParams->GetLightConeConstraint();
				gl->hotsizeSpin->SetValue(tmpHot, FALSE);
				gl->fallsizeSpin->SetValue(tmpFall, FALSE);
				gl->SetHotspot(t, tmpHot);
				gl->SetFallsize(t, tmpFall);
			}
			else {
				if(LOWORD(wParam) == IDC_LHOTSIZESPINNER)
					gl->SetHotspot(t, tmpHot);
				else
					gl->SetFallsize(t, tmpFall);
			}
			gl->SetSpinnerBracket(gl->hotsizeSpin,PB_HOTSIZE);
			gl->SetSpinnerBracket(gl->fallsizeSpin,PB_FALLSIZE);
			gl->BuildSpotMesh(max(gl->hotsizeSpin->GetFVal(), gl->fallsizeSpin->GetFVal()));
			break;
		case IDC_LTDISTSPINNER:
			gl->SetTDist(t, gl->tDistSpin->GetFVal());
			gl->SetSpinnerBracket(gl->tDistSpin,PB_TDIST);
			break;				   
		case IDC_START_RANGE_SPIN:
			gl->ChangeAtten(ATTEN_START);
			gl->SetSpinnerBracket(gl->attenStartSpin, gl->type==OMNI_LIGHT? PB_OMNIATSTART:PB_ATTENSTART);
			break;
		case IDC_END_RANGE_SPIN:
			gl->ChangeAtten(ATTEN_END);
			gl->SetSpinnerBracket(gl->attenEndSpin, gl->type==OMNI_LIGHT? PB_OMNIATEND:PB_ATTENEND);
			break;
		case IDC_START_RANGE1_SPIN:
			gl->ChangeAtten(ATTEN1_START);
			gl->SetSpinnerBracket(gl->attenStart1Spin, gl->type==OMNI_LIGHT? PB_OMNIATSTART1:PB_ATTENSTART1);
			break;
		case IDC_END_RANGE1_SPIN:
			gl->ChangeAtten(ATTEN1_END);
			gl->SetSpinnerBracket(gl->attenEnd1Spin, gl->type==OMNI_LIGHT? PB_OMNIATEND1:PB_ATTENEND1);
			break;
		case IDC_LASPECT_SPIN:
			gl->SetAspect( t, gl->aspectSpin->GetFVal());
			gl->SetSpinnerBracket(gl->aspectSpin,PB_ASPECT);
			break;
		case IDC_MAP_SIZE_SPIN:
			gl->SetMapSize( t, gl->mapSizeSpin->GetIVal());
			gl->SetSpinnerBracket(gl->mapSizeSpin, gl->type==OMNI_LIGHT? PB_OMNIMAP_SIZE:PB_MAP_SIZE);
			redraw = FALSE;
			break;
		case IDC_MAP_BIAS_SPIN:
			gl->SetMapBias( t, gl->mapBiasSpin->GetFVal());
			gl->SetSpinnerBracket(gl->mapBiasSpin, gl->type==OMNI_LIGHT? PB_OMNIMAP_BIAS:PB_MAP_BIAS);
			redraw = FALSE;
			break;
		case IDC_MAP_RANGE_SPIN:
			gl->SetMapRange( t, gl->mapRangeSpin->GetFVal());
			gl->SetSpinnerBracket(gl->mapRangeSpin, gl->type==OMNI_LIGHT? PB_OMNIMAP_RANGE:PB_MAP_RANGE);
			redraw = FALSE;
			break;
		case IDC_RT_BIAS_SPIN:
			gl->SetRayBias( t, gl->rayBiasSpin->GetFVal());
			gl->SetSpinnerBracket(gl->rayBiasSpin, gl->type==OMNI_LIGHT? PB_OMNIRAY_BIAS:PB_RAY_BIAS);
			redraw = FALSE;
			break;

/*		case IDC_RADIUS_SPIN:
			gl->SetRadius( t, gl->radiusSpin->GetFVal());
			gl->SetSpinnerBracket(gl->radiusSpin, gl->type==OMNI_LIGHT? PB_OMNIMAP_SIZE:PB_RADIUS);
			redraw = FALSE;
			break;
*/
/*		case IDC_NUMBER_LIGHTS_SPIN:
			gl->SetNumberLights( t, gl->numberLightsSpin->GetIVal());
			gl->SetSpinnerBracket(gl->numberLightsSpin, gl->type==OMNI_LIGHT? PB_OMNIMAP_SIZE:PB_NUMBER_LIGHTS);
			redraw = FALSE;
			break;
*/


		}
		if (redraw)
			gl->iObjParams->RedrawViews(gl->iObjParams->GetTime(),REDRAW_INTERACTIVE);
		}
		gl->SetUpdateHSVSpin(TRUE);
		return TRUE;

	case CC_SPINNER_BUTTONDOWN:
		theHold.Begin();
		return TRUE;

	case CC_COLOR_CHANGE:
		if(HIWORD(wParam))  {	// button up
			DWORD rgb = gl->colorSwatch->GetColor();
			gl->SetRGBColor(t, Point3(GetRValue(rgb)/255.0f, GetGValue(rgb)/255.0f, GetBValue(rgb)/255.0f));
			gl->iObjParams->RedrawViews(gl->iObjParams->GetTime(),REDRAW_END);
			gl->UpdateUI(t);
			}
		return TRUE;

	case WM_CUSTEDIT_ENTER:
	case CC_SPINNER_BUTTONUP:
		if (HIWORD(wParam) || message==WM_CUSTEDIT_ENTER) 
			theHold.Accept(GetString(IDS_DS_PARAMCHG));
		else 
			theHold.Cancel();
		gl->iObjParams->RedrawViews(gl->iObjParams->GetTime(),REDRAW_END);
		return TRUE;

	case WM_MOUSEACTIVATE:
		gl->iObjParams->RealizeParamPanel();
		return FALSE;

	case WM_LBUTTONDOWN:
	case WM_LBUTTONUP:
	case WM_RBUTTONDOWN:
	case WM_RBUTTONUP:
	case WM_MOUSEMOVE:
		gl->iObjParams->RollupMouseMessage(hDlg,message,wParam,lParam);
		return FALSE;

	case WM_COMMAND:			
		switch( id = LOWORD(wParam) ) {
		case IDC_LIGHT_ON:
			gl->SetUseLight( IsDlgButtonChecked( hDlg, IDC_LIGHT_ON ) );
			gl->iObjParams->RedrawViews(t);
			break;
		case IDC_SOFTEN_DIFFUSE:
			gl->softenDiffuse = IsDlgButtonChecked( hDlg, IDC_SOFTEN_DIFFUSE );
			gl->NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
			gl->iObjParams->RedrawViews(t);
			break;
		case IDC_AFFECT_DIFFUSE:
			gl->affectDiffuse = IsDlgButtonChecked( hDlg, IDC_AFFECT_DIFFUSE );
			gl->NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
			gl->iObjParams->RedrawViews(t);
			break;
		case IDC_AFFECT_SPECULAR:
			gl->affectSpecular =  IsDlgButtonChecked( hDlg, IDC_AFFECT_SPECULAR );
			gl->NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
			gl->iObjParams->RedrawViews(t);
			break;
		case IDC_SHOW_CONE:
			gl->SetConeDisplay( IsDlgButtonChecked( hDlg, IDC_SHOW_CONE ) );
			gl->iObjParams->RedrawViews(t);
			break;
		case IDC_USE_ATTEN1:
			gl->SetUseAttenNear( IsDlgButtonChecked( hDlg, IDC_USE_ATTEN1 ) );
			gl->iObjParams->RedrawViews(t);
			break;
		case IDC_SHOW_RANGES1:
			gl->SetAttenNearDisplay( IsDlgButtonChecked( hDlg, IDC_SHOW_RANGES1 ) );
			gl->iObjParams->RedrawViews(t);
			break;
		case IDC_USE_ATTEN:
			gl->SetUseAtten( IsDlgButtonChecked( hDlg, IDC_USE_ATTEN ) );
			gl->iObjParams->RedrawViews(t);
			break;
		case IDC_SHOW_RANGES:
			gl->SetAttenDisplay( IsDlgButtonChecked( hDlg, IDC_SHOW_RANGES ) );
			gl->iObjParams->RedrawViews(t);
			break;

		case IDC_ATT_NONE:
		case IDC_ATT_INV:
		case IDC_ATT_INVSQ:
			CheckRadioButton( hDlg, IDC_ATT_NONE, IDC_ATT_INVSQ, id);
			gl->SetDecayType(id-IDC_ATT_NONE);
			break;

		case IDC_RECT_LIGHT:
		case IDC_CIRCLE_LIGHT:
			if(wParam > 2048)
				break;		// spurious message sent by dbl-clicking the blue spinner
							// some how ends up here with wParam > 2048 -- so punt it
			CheckRadioButton(hDlg, IDC_RECT_LIGHT, IDC_CIRCLE_LIGHT, id);
			gl->SetSpotShape((id == IDC_RECT_LIGHT)?RECT_LIGHT:CIRCLE_LIGHT);
//			EnableWindow(GetDlgItem(hDlg, IDC_LASPECT), !gl->GetSpotShape());
//			EnableWindow(GetDlgItem(hDlg, IDC_LASPECT_SPIN), !gl->GetSpotShape());
			EnableWindow(GetDlgItem(hDlg, IDC_BITMAP_FIT), !gl->GetSpotShape());
			gl->UpdateUI(t);
			gl->iObjParams->RedrawViews(t);
			break;
		case IDC_SHADOW_MAPS:
		case IDC_RAY_TRACED_SHADOWS:
			CheckRadioButton( hDlg, IDC_SHADOW_MAPS, IDC_RAY_TRACED_SHADOWS, id);
			gl->SetShadowType((id==IDC_SHADOW_MAPS)?0:1);
			break;
		case IDC_GLOBAL_SET:
			gl->SetUseGlobal( IsDlgButtonChecked( hDlg, id));
			gl->ReloadShadCtrls(hDlg,t);
			break;
		case IDC_CAST_SHADOWS:
			gl->SetShadow( IsDlgButtonChecked( hDlg, id));
			break;
		case IDC_ABS_MAP_BIAS:
			gl->SetAbsMapBias( IsDlgButtonChecked( hDlg, id));
			break;
		case IDC_OVERSHOOT:
			gl->SetOvershoot( IsDlgButtonChecked( hDlg, id));
			gl->iObjParams->RedrawViews(t);
			break;
		case IDC_PROJECTOR:
			gl->SetProjector(IsDlgButtonChecked( hDlg, id));
			break;
		case IDC_PROJ_MAPNAME:
			gl->BrowseProjectorMap(hDlg);
//			gl->iObjParams->PutMtlToMtlEditor(gl->projMap);
			break;
//		case IDC_PROJ_IMAGE:
//			gl->AssignProjectorMap(hDlg);
//			break;
		case IDC_EXCLUDE_DLG:
			gl->iObjParams->DoExclusionListDialog(&gl->exclList, TRUE /*gl->IsSpot()*/ );
			SetDlgItemText(hDlg,IDC_EXCLUDE_DLG, 
				GetString(gl->Include()?IDS_DS_INCLUDE__: IDS_DS_EXCLUDE__));
			return TRUE;
		case IDC_BITMAP_FIT:
			gl->DoBitmapFit(hDlg,t);
			break;
		return FALSE;
		}
	default:
		return FALSE;
	}
}

void GeneralLight::FixHotFallConstraint() {
	if (iObjParams==NULL)
		return;
	TimeValue t = iObjParams->GetTime();
	if(IsSpot()) {
		if(GetFallsize(t) < GetHotspot(t) + iObjParams->GetLightConeConstraint())
			SetFallsize(t, GetHotspot(t) + iObjParams->GetLightConeConstraint());
		}
	}

void GeneralLight::ReloadShadCtrls(HWND hDlg, TimeValue t) {
//	EnableWindow(GetDlgItem(hDlg, IDC_SHADOW_MAPS), onOff);
//	EnableWindow(GetDlgItem(hDlg, IDC_RAY_TRACED_SHADOWS), onOff);
//	EnableWindow(GetDlgItem(hDlg, IDC_MAP_BIAS), onOff);
//	EnableWindow(GetDlgItem(hDlg, IDC_MAP_BIAS_SPIN), onOff);
//	EnableWindow(GetDlgItem(hDlg, IDC_MAP_SIZE), onOff);
//	EnableWindow(GetDlgItem(hDlg, IDC_MAP_SIZE_SPIN), onOff);
//	EnableWindow(GetDlgItem(hDlg, IDC_MAP_RANGE), onOff);
//	EnableWindow(GetDlgItem(hDlg, IDC_MAP_RANGE_SPIN), onOff);
//	EnableWindow(GetDlgItem(hDlg, IDC_RT_BIAS), onOff);
//	EnableWindow(GetDlgItem(hDlg, IDC_RT_BIAS_SPIN), onOff);
//	EnableWindow(GetDlgItem(hDlg, IDC_ABS_MAP_BIAS), onOff);
	CheckRadioButton( hDlg, IDC_SHADOW_MAPS, IDC_RAY_TRACED_SHADOWS, 
		IDC_SHADOW_MAPS + GetShadowType());
	CheckDlgButton( hDlg, IDC_ABS_MAP_BIAS, GetAbsMapBias() );
	mapSizeSpin->SetValue( GetMapSize(t),FALSE );
	mapBiasSpin->SetValue( GetMapBias(t),FALSE );
	mapRangeSpin->SetValue( GetMapRange(t),FALSE );
	rayBiasSpin->SetValue( GetRayBias(t),FALSE );

//	radiusSpin->SetValue( GetRadius(t),FALSE );
//	numberLightsSpin->SetValue( GetNumberLights(t),FALSE );


	}


void GeneralLight::BeginEditParams( IObjParam *ip, ULONG flags,Animatable *prev )
	{
	iObjParams = ip;
	TimeValue t = ip->GetTime();
	Point3 color;
	
	projDADMgr.Init(this);

	if ( !hGeneralLight ) {
		hGeneralLight = ip->AddRollupPage( 
				hInstance, 
				/*type == DIR_LIGHT ? MAKEINTRESOURCE(IDD_DIRLIGHTPARAM) :*/ MAKEINTRESOURCE(IDD_LIGHT_PARAM),
				GeneralLightParamDialogProc, 
				GetString(IDS_DB_GENERAL_PARAMS), 
				(LPARAM)this );		
		ip->RegisterDlgWnd(hGeneralLight);

		redSpin   =	GetISpinner( GetDlgItem(hGeneralLight, IDC_LREDSPINNER) );
		greenSpin = GetISpinner( GetDlgItem(hGeneralLight, IDC_LGREENSPINNER) );
		blueSpin  = GetISpinner( GetDlgItem(hGeneralLight, IDC_LBLUESPINNER) );
		hSpin 	  = GetISpinner( GetDlgItem(hGeneralLight, IDC_LHSPINNER) );
		sSpin 	  = GetISpinner( GetDlgItem(hGeneralLight, IDC_LSSPINNER) );
		vSpin 	  = GetISpinner( GetDlgItem(hGeneralLight, IDC_LVSPINNER) );
		color = GetRGBColor(t);
		InitColorSpinner(hGeneralLight, redSpin, IDC_LRED, FLto255f(color.x));
		InitColorSpinner(hGeneralLight, greenSpin, IDC_LGREEN, FLto255f(color.y));
		InitColorSpinner(hGeneralLight, blueSpin, IDC_LBLUE, FLto255f(color.z));
		colorSwatch = GetIColorSwatch(GetDlgItem(hGeneralLight, IDC_LIGHT_COLOR), 
				RGB(FLto255i(color.x), FLto255i(color.y), FLto255i(color.z)), 
				GetString(IDS_DS_LIGHTCOL));
		colorSwatch->ForceDitherMode(1);
		colorSwatch->SetModal();
		color = GetHSVColor(t);
		InitColorSpinner(hGeneralLight, hSpin, IDC_LH, FLto255f(color.x));
		InitColorSpinner(hGeneralLight, sSpin, IDC_LS, FLto255f(color.y));
		InitColorSpinner(hGeneralLight, vSpin, IDC_LV, FLto255f(color.z));
		intensitySpin = GetISpinner( GetDlgItem(hGeneralLight, IDC_LMULTSPINNER) );
		intensitySpin->SetLimits(-50.0f, 50.0f, FALSE);
		intensitySpin->SetValue(GetIntensity(t),FALSE);
		intensitySpin->SetScale(0.01f);
		intensitySpin->LinkToEdit( GetDlgItem(hGeneralLight, IDC_LMULT), EDITTYPE_FLOAT );

		contrastSpin = SetupFloatSpinner(hGeneralLight, IDC_LCONTRASTSPIN2, IDC_LCONTRAST2,0.0f,100.0f,
			0.0f,1.0f);
		diffsoftSpin = SetupFloatSpinner(hGeneralLight, IDC_DIFFSOFTSPIN, IDC_DIFFSOFT,0.0f,100.0f,
			0.0f,1.0f);

		contrastSpin->SetValue(GetContrast(t), FALSE);
		diffsoftSpin->SetValue(GetDiffuseSoft(t), FALSE);

		attenStartSpin = GetISpinner( GetDlgItem(hGeneralLight,IDC_START_RANGE_SPIN) );
		attenEndSpin = GetISpinner( GetDlgItem(hGeneralLight,IDC_END_RANGE_SPIN) );
		InitRangeSpinner(hGeneralLight, attenStartSpin, IDC_START_RANGE, GetAtten(t, ATTEN_START), IsDir());
		InitRangeSpinner(hGeneralLight, attenEndSpin, IDC_END_RANGE, GetAtten(t, ATTEN_END), IsDir());
		SetAttenDisplay( IsDlgButtonChecked(hGeneralLight, IDC_SHOW_RANGES) );
		SetUseAtten( IsDlgButtonChecked(hGeneralLight, IDC_USE_ATTEN) );

		attenStart1Spin = GetISpinner( GetDlgItem(hGeneralLight,IDC_START_RANGE1_SPIN) );
		attenEnd1Spin = GetISpinner( GetDlgItem(hGeneralLight,IDC_END_RANGE1_SPIN) );
		InitRangeSpinner(hGeneralLight, attenStart1Spin, IDC_START_RANGE1, GetAtten(t, ATTEN1_START),IsDir());
		InitRangeSpinner(hGeneralLight, attenEnd1Spin, IDC_END_RANGE1, GetAtten(t, ATTEN1_END),IsDir());
		SetAttenNearDisplay( IsDlgButtonChecked(hGeneralLight, IDC_SHOW_RANGES1) );
		SetUseAttenNear( IsDlgButtonChecked(hGeneralLight, IDC_USE_ATTEN1) );


		if(IsSpot()) {
			hSpotLight = ip->AddRollupPage( 
					hInstance, 
					(type == FSPOT_LIGHT||type==DIR_LIGHT) ? MAKEINTRESOURCE(IDD_FREE_SPOTLIGHT) : MAKEINTRESOURCE(IDD_SPOTLIGHT),
					GeneralLightParamDialogProc, 
					IsDir() ? GetString(IDS_DB_DIR_PARAMS) : GetString(IDS_DB_SPOT_PARAMS), 
					(LPARAM)this );		
			ip->RegisterDlgWnd(hSpotLight);				

			projMapName = GetICustButton(GetDlgItem(hSpotLight, IDC_PROJ_MAPNAME));	
			projMapName->SetText(projMap?projMap->GetFullName().data():GetString(IDS_DB_NONE));
			projMapName->SetDADMgr(&projDADMgr);
			
			if (type == FSPOT_LIGHT||type==DIR_LIGHT) {
				tDistSpin = GetISpinner( GetDlgItem(hSpotLight, IDC_LTDISTSPINNER) );
				tDistSpin->SetLimits(1.0f, 10000.0f, FALSE);
				tDistSpin->SetValue(GetTDist(t),FALSE);
				tDistSpin->SetScale(1.0f);
				tDistSpin->LinkToEdit( GetDlgItem(hSpotLight, IDC_LTDIST), EDITTYPE_UNIVERSE );
				}

			aspectSpin = GetISpinner( GetDlgItem(hSpotLight, IDC_LASPECT_SPIN) );
			aspectSpin->SetLimits(0.001f, 100.0f, FALSE);
			aspectSpin->SetValue(GetAspect(t),FALSE);
			aspectSpin->SetScale(0.01f);
			aspectSpin->LinkToEdit( GetDlgItem(hSpotLight, IDC_LASPECT), EDITTYPE_FLOAT );

			ShowWindow(GetDlgItem(hSpotLight, IDC_LTDIST), type == FSPOT_LIGHT||type==DIR_LIGHT ? SW_SHOW : SW_HIDE);
			ShowWindow(GetDlgItem(hSpotLight, IDC_LTDISTSPINNER), type == FSPOT_LIGHT||type==DIR_LIGHT ? SW_SHOW : SW_HIDE);
			SetConeDisplay( IsDlgButtonChecked(hSpotLight, IDC_SHOW_CONE) );

			hotsizeSpin = GetISpinner( GetDlgItem(hSpotLight,IDC_LHOTSIZESPINNER) );
			fallsizeSpin = GetISpinner( GetDlgItem(hSpotLight,IDC_LFALLOFFSPINNER) );

			if(IsDir()) {
				hotsizeSpin->SetLimits(.001f, (float)1.0e30, FALSE);
				hotsizeSpin->SetValue(GetHotspot(t),FALSE);
				hotsizeSpin->SetScale(1.0f);
				hotsizeSpin->LinkToEdit( GetDlgItem(hSpotLight, IDC_LHOTSIZE), EDITTYPE_UNIVERSE );
	
				fallsizeSpin->SetLimits(.001f, (float)1.0e30, FALSE);
				fallsizeSpin->SetValue(GetFallsize(t),FALSE);
				fallsizeSpin->SetScale(1.0f);
				fallsizeSpin->LinkToEdit( GetDlgItem(hSpotLight, IDC_LFALLOFF), EDITTYPE_UNIVERSE );
				}
			else {
				InitAngleSpinner(hSpotLight, hotsizeSpin, IDC_LHOTSIZE, GetHotspot(t) );
				InitAngleSpinner(hSpotLight, fallsizeSpin, IDC_LFALLOFF, GetFallsize(t) );
				}
			}			
		if (type == OMNI_LIGHT) {
			hOmni = ip->AddRollupPage( 
				hInstance, 
				MAKEINTRESOURCE(IDD_OMNI),
				GeneralLightParamDialogProc, 
				GetString(IDS_DS_PROJ_PARAMS), 
				(LPARAM)this );		
			ip->RegisterDlgWnd(hOmni);				
			projMapName = GetICustButton(GetDlgItem(hOmni, IDC_PROJ_MAPNAME));		
			projMapName->SetText(projMap?projMap->GetFullName().data():GetString(IDS_DB_NONE));
			projMapName->SetDADMgr(&projDADMgr);
			}
				
		//--Shadow Rollup ---------------------------------------------------

		hShadow = ip->AddRollupPage( 
				hInstance, 
				MAKEINTRESOURCE(IDD_LIGHT_SSHADOW),
				GeneralLightParamDialogProc, 
				GetString(IDS_DB_SHADOW_PARAMS), 
				(LPARAM)this );		
		ip->RegisterDlgWnd(hShadow);				

		mapSizeSpin = SetupIntSpinner(hShadow, IDC_MAP_SIZE_SPIN, IDC_MAP_SIZE,
			0, 10000, GetMapSize(t)); 
		mapBiasSpin = SetupFloatSpinner(hShadow, IDC_MAP_BIAS_SPIN, IDC_MAP_BIAS,
			0.0f, 1000.0f, GetMapBias(t), .1f); 
		mapRangeSpin = SetupFloatSpinner(hShadow, IDC_MAP_RANGE_SPIN, IDC_MAP_RANGE,
			0.0f, 20.0f, GetMapRange(t), .1f); 
		rayBiasSpin = SetupFloatSpinner(hShadow, IDC_RT_BIAS_SPIN, IDC_RT_BIAS,
			0.0f, 20.0f, GetRayBias(t), .1f); 

/*		radiusSpin = SetupFloatSpinner(hShadow, IDC_RADIUS_SPIN, IDC_RADIUS,
			0.0f, 10000.0f, GetRadius(t)); 
		numberLightsSpin = SetupIntSpinner(hShadow, IDC_NUMBER_LIGHTS_SPIN, IDC_NUMBER_LIGHTS,
			0, 100, GetNumberLights(t)); 
*/


		if(IsSpot()) {
			if(GetFallsize(t) < GetHotspot(t) + ip->GetLightConeConstraint())
				SetFallsize(t, GetHotspot(t) + ip->GetLightConeConstraint());
			}
		} 
	else {
		SetWindowLong( hGeneralLight, GWL_USERDATA, (LONG)this );
		if(IsSpot())
			SetWindowLong( hSpotLight, GWL_USERDATA, (LONG)this );
		if(type==OMNI_LIGHT)
			SetWindowLong( hOmni, GWL_USERDATA, (LONG)this );
		SetWindowLong( hShadow, GWL_USERDATA, (LONG)this );
		
		color = GetRGBColor(t);
		redSpin->SetValue(FLto255f(color.x), FALSE);
		greenSpin->SetValue(FLto255f(color.y), FALSE);
		blueSpin->SetValue(FLto255f(color.z), FALSE);
		color = GetHSVColor(t);
		hSpin->SetValue(FLto255f(color.x), FALSE);
		sSpin->SetValue(FLto255f(color.y), FALSE);
		vSpin->SetValue(FLto255f(color.z), FALSE);
		intensitySpin->SetValue(GetIntensity(t), FALSE);


		attenStartSpin->SetValue(GetAtten(t, ATTEN_START), FALSE);
		attenEndSpin->SetValue(GetAtten(t, ATTEN_END), FALSE);
		SetAttenDisplay( IsDlgButtonChecked(hGeneralLight, IDC_SHOW_RANGES) );
		SetUseAtten( IsDlgButtonChecked(hGeneralLight, IDC_USE_ATTEN) );

		attenStart1Spin->SetValue(GetAtten(t, ATTEN1_START), FALSE);
		attenEnd1Spin->SetValue(GetAtten(t, ATTEN1_END), FALSE);
		SetAttenNearDisplay( IsDlgButtonChecked(hGeneralLight, IDC_SHOW_RANGES1) );
		SetUseAttenNear( IsDlgButtonChecked(hGeneralLight, IDC_USE_ATTEN1) );

		if(IsSpot()) {
			hotsizeSpin->SetValue(GetHotspot(t), FALSE);
			fallsizeSpin->SetValue(GetFallsize(t), FALSE);
			if (type == FSPOT_LIGHT||type==DIR_LIGHT) 
				tDistSpin->SetValue(GetTDist(t), FALSE);
			SetConeDisplay( IsDlgButtonChecked(hSpotLight, IDC_SHOW_CONE) );
			}

		if (projMapName)
			projMapName->SetDADMgr(&projDADMgr);
		mapSizeSpin->SetValue( GetMapSize(t) ,FALSE);
		mapBiasSpin->SetValue( GetMapBias(t) ,FALSE);
		mapRangeSpin->SetValue( GetMapRange(t) ,FALSE);
		rayBiasSpin->SetValue( GetRayBias(t),FALSE );

//		radiusSpin->SetValue( GetRadius(t) ,FALSE);
//		numberLightsSpin->SetValue( GetNumberLights(t) ,FALSE);


		}
	}

		
void GeneralLight::EndEditParams( IObjParam *ip, ULONG flags,Animatable *next )
{
	TimeValue t = ip->GetTime();
	Point3 color;
	if (projMapName)
		projMapName->SetDADMgr(NULL);

	color = GetRGBColor(t);
	dlgRed =   FLto255i(color.x);
	dlgGreen = FLto255i(color.y);
	dlgBlue =  FLto255i(color.z);
	color = GetHSVColor(t);
	dlgH = FLto255i(color.x);
	dlgS = FLto255i(color.y);
	dlgV = FLto255i(color.z);
	dlgIntensity = GetIntensity(t);
	dlgContrast = GetContrast(t);
	dlgDiffsoft = GetDiffuseSoft(t);
	if(IsSpot()) {
		dlgHotsize = GetHotspot(t);
		dlgFallsize = GetFallsize(t);
		dlgTDist = GetTDist(t);
		dlgShowCone = IsDlgButtonChecked(hSpotLight, IDC_SHOW_CONE );
		dlgShape = GetSpotShape();
		}

	dlgAtten1Start = GetAtten(t, ATTEN1_START);
	dlgAtten1End = GetAtten(t, ATTEN1_END);
	dlgAttenStart = GetAtten(t, ATTEN_START);
	dlgAttenEnd = GetAtten(t, ATTEN_END);
	dlgShowAtten = IsDlgButtonChecked(hGeneralLight, IDC_SHOW_RANGES );

	if ( flags&END_EDIT_REMOVEUI ) {
		HWND hwgl = hGeneralLight;
		hGeneralLight = NULL; // This keeps UpdateUI from jumping in

		if ( IsSpot() ) {
			ip->UnRegisterDlgWnd(hSpotLight);		
			ip->DeleteRollupPage(hSpotLight);
			hSpotLight = NULL;
			}

		if (type==OMNI_LIGHT) {
			ip->UnRegisterDlgWnd(hOmni);		
			ip->DeleteRollupPage(hOmni);
			hOmni = NULL;
			}
		ip->UnRegisterDlgWnd(hShadow);		
		ip->DeleteRollupPage(hShadow);
		hShadow = NULL;

		ip->UnRegisterDlgWnd(hwgl);		
		ip->DeleteRollupPage(hwgl);
		} 
	else {
		SetWindowLong( hGeneralLight, GWL_USERDATA, 0 );
		if( IsSpot() ) {
			SetWindowLong( hSpotLight, GWL_USERDATA, 0 );
			}
		if (type==OMNI_LIGHT) {
			SetWindowLong( hOmni, GWL_USERDATA, 0 );
			}		
		SetWindowLong( hShadow, GWL_USERDATA, 0 );
		}
	iObjParams = NULL;
	}


void GeneralLight::EnumAuxFiles(NameEnumCallback& nameEnum, DWORD flags) {
	if ((flags&FILE_ENUM_INACTIVE)||projector)
		if (projMap) projMap->EnumAuxFiles(nameEnum,flags);
	} 

void GeneralLight::DoBitmapFit(HWND hwnd, TimeValue t) {
	if(!(IsSpot()&&shape==RECT_LIGHT)) return;
	BitmapInfo bi;
	TheManager->SelectFileInputEx(&bi, hwnd, GetString(IDS_DB_SELECT_TO_FIT));
	if (bi.Name()[0]) {
		TheManager->GetImageInfo(&bi, bi.Name());
       	int w = bi.Width();
        int h = bi.Height();
		if (h) {
			SetAspect(t,float(w)/float(h));
			iObjParams->RedrawViews(t,REDRAW_END);
			}
		}
	}

void GeneralLight::AssignProjectorMap(Texmap *m, BOOL newmat) {	
	assert(!m || IsTex(m));
	ReplaceReference(PROJMAP_REF,m);
	if (projMap) {
		projMapName->SetText(projMap->GetFullName().data());
		if (newmat) 
			projMap->RecursInitSlotType(type==OMNI_LIGHT?MAPSLOT_ENVIRON:MAPSLOT_TEXTURE);
		} 
	else {
		projMapName->SetText(GetString(IDS_DB_NONE));
		}		
	if (projMap) {
		projector = TRUE;
		CheckDlgButton(type==OMNI_LIGHT?hOmni:hSpotLight,IDC_PROJECTOR, projector);
		}
	NotifyDependents(FOREVER,0,REFMSG_SUBANIM_STRUCTURE_CHANGED);
	}

void GeneralLight::BrowseProjectorMap(HWND hWnd) {	
	BOOL newMat, cancel;
	Texmap *m = (Texmap *)iObjParams->DoMaterialBrowseDlg(
		hWnd,BROWSE_MAPSONLY|BROWSE_INCNONE,newMat,cancel);
	if (!cancel)
		AssignProjectorMap(m,newMat);
	}

void GeneralLight::SetProjMap(Texmap* pmap)
	{
	ReplaceReference(PROJMAP_REF,pmap);
	if (projMap) projector = TRUE;	
	NotifyDependents(FOREVER,0,REFMSG_SUBANIM_STRUCTURE_CHANGED);
	}

static int GetTargetPoint(TimeValue t, INode *inode, Point3& p) 
{
	Matrix3 tmat;
	if (inode->GetTargetTM(t,tmat)) {
		p = tmat.GetTrans();
		return 1;
	}
	else 
		return 0;
}


void GeneralLight::UpdateTargDistance(TimeValue t, INode* inode) {
	if (type == TSPOT_LIGHT&&hSpotLight) {
		Point3 pt,v[3];
		if (GetTargetPoint(t, inode, pt)) {
			Matrix3 tm = inode->GetObjectTM(t);
			float den = FLength(tm.GetRow(2));
			float dist = (den!=0) ? FLength(tm.GetTrans()-pt) / den : 0.0f;
			targDist = dist;
			TCHAR buf[40];
			_stprintf(buf,_T("%0.3f"),targDist);
			SetWindowText(GetDlgItem(hSpotLight,IDC_TARG_DISTANCE),buf);
			}
		}
	}

TCHAR *GeneralLight::GetObjectName()
{ 
	switch(type) {
	case OMNI_LIGHT:
		return GetString(IDS_DB_OMNI_LIGHT);
	case DIR_LIGHT:
		return GetString(IDS_DB_DIR_LIGHT);
	case TDIR_LIGHT:
		return GetString(IDS_DS_TDIR_LIGHT);
	case FSPOT_LIGHT:
		return GetString(IDS_DB_FREE_SPOT);
	case TSPOT_LIGHT:
		return GetString(IDS_DB_TARGET_SPOT);
	default:
		return NULL;
	}
}

void GeneralLight::InitNodeName(TSTR& s)
{
	switch(type) {
	case OMNI_LIGHT:
		s = GetString(IDS_DB_OMNI);
		return; 
	case DIR_LIGHT:
		s = GetString(IDS_DB_DIRECT);
		return;
	case TDIR_LIGHT:
		s = GetString(IDS_DS_TDIRECT);
		return;
	case FSPOT_LIGHT:
		s = GetString(IDS_DB_FSPOT); 
		return;
	case TSPOT_LIGHT:
		s = GetString(IDS_DB_SPOT); 
		return;
	default:
		s = GetString(IDS_DB_LIGHT); 
	}
	return;
}

#define FZ (0.0f)

#define SET_QUAD(face, v0, v1, v2, v3) \
	staticMesh[DIR_MESH].faces[face].setVerts(v0, v1, v2); \
	staticMesh[DIR_MESH].faces[face].setEdgeVisFlags(1,1,0); \
	staticMesh[DIR_MESH].faces[face+1].setVerts(v0, v2, v3); \
	staticMesh[DIR_MESH].faces[face+1].setEdgeVisFlags(0,1,1);

void GeneralLight::BuildStaticMeshes()
{
	if(!meshBuilt) {
		int nverts = 6;
		int nfaces = 8;
		// Build a leetle octahedron
		staticMesh[OMNI_MESH].setNumVerts(nverts);
		staticMesh[OMNI_MESH].setNumFaces(nfaces);
		float s = 8.0f;
		staticMesh[OMNI_MESH].setVert(0, Point3( FZ,FZ, -s));
		staticMesh[OMNI_MESH].setVert(1, Point3( s, FZ, FZ));
		staticMesh[OMNI_MESH].setVert(2, Point3( FZ, s, FZ));
		staticMesh[OMNI_MESH].setVert(3, Point3(-s, FZ, FZ));
		staticMesh[OMNI_MESH].setVert(4, Point3( FZ,-s, FZ));
		staticMesh[OMNI_MESH].setVert(5, Point3( FZ,FZ,  s));
		staticMesh[OMNI_MESH].faces[0].setVerts(0,1,4);
		staticMesh[OMNI_MESH].faces[1].setVerts(0,4,3);
		staticMesh[OMNI_MESH].faces[2].setVerts(0,3,2);
		staticMesh[OMNI_MESH].faces[3].setVerts(0,2,1);
		staticMesh[OMNI_MESH].faces[4].setVerts(5,1,2);
		staticMesh[OMNI_MESH].faces[5].setVerts(5,2,3);
		staticMesh[OMNI_MESH].faces[6].setVerts(5,3,4);
		staticMesh[OMNI_MESH].faces[7].setVerts(5,4,1);
		for (int i=0; i<nfaces; i++) {
			staticMesh[OMNI_MESH].faces[i].setSmGroup(i);
			staticMesh[OMNI_MESH].faces[i].setEdgeVisFlags(1,1,1);
			}
		staticMesh[OMNI_MESH].buildNormals();
		staticMesh[OMNI_MESH].EnableEdgeList(1);

		// Build an "arrow"
		nverts = 13;
		nfaces = 16;
		staticMesh[DIR_MESH].setNumVerts(nverts);
		staticMesh[DIR_MESH].setNumFaces(nfaces);
		s = 4.0f;
		float s4 = 16.0f;
		staticMesh[DIR_MESH].setVert( 0, Point3( -s,-s, FZ));
		staticMesh[DIR_MESH].setVert( 1, Point3(  s,-s, FZ));
		staticMesh[DIR_MESH].setVert( 2, Point3(  s, s, FZ));
		staticMesh[DIR_MESH].setVert( 3, Point3( -s, s, FZ));
		staticMesh[DIR_MESH].setVert( 4, Point3( -s,-s, -s4));
		staticMesh[DIR_MESH].setVert( 5, Point3(  s,-s, -s4));
		staticMesh[DIR_MESH].setVert( 6, Point3(  s, s, -s4));
		staticMesh[DIR_MESH].setVert( 7, Point3( -s, s, -s4));
		s *= (float)2.0;
		staticMesh[DIR_MESH].setVert( 8, Point3( -s,-s, -s4));
		staticMesh[DIR_MESH].setVert( 9, Point3(  s,-s, -s4));
		staticMesh[DIR_MESH].setVert(10, Point3(  s, s, -s4));
		staticMesh[DIR_MESH].setVert(11, Point3( -s, s, -s4));
		staticMesh[DIR_MESH].setVert(12, Point3( FZ,FZ, -s4-s));
		SET_QUAD( 0, 1, 0, 4, 5);
		SET_QUAD( 2, 0, 3, 7, 4);
		SET_QUAD( 4, 3, 2, 6, 7);
		SET_QUAD( 6, 2, 1, 5, 6);
		SET_QUAD( 8, 0, 1, 2, 3);
		SET_QUAD(10, 8, 9, 10, 11);
		staticMesh[DIR_MESH].faces[12].setVerts(8,12,9);
		staticMesh[DIR_MESH].faces[12].setEdgeVisFlags(1,1,1);
		staticMesh[DIR_MESH].faces[13].setVerts(9,12,10);
		staticMesh[DIR_MESH].faces[13].setEdgeVisFlags(1,1,1);
		staticMesh[DIR_MESH].faces[14].setVerts(10,12,11);
		staticMesh[DIR_MESH].faces[14].setEdgeVisFlags(1,1,1);
		staticMesh[DIR_MESH].faces[15].setVerts(11,12,8);
		staticMesh[DIR_MESH].faces[15].setEdgeVisFlags(1,1,1);
		for (i=0; i<nfaces; i++)
			staticMesh[DIR_MESH].faces[i].setSmGroup(i);
		staticMesh[DIR_MESH].buildNormals();
		staticMesh[DIR_MESH].EnableEdgeList(1);

		meshBuilt = 1;
	}
}


void GeneralLight::BuildSpotMesh(float coneSize)
{
	// build a cone
	if(coneSize < 0.0f)
		return;
	int nverts = 9;
	int nfaces = 8;
	spotMesh.setNumVerts(nverts);
	spotMesh.setNumFaces(nfaces);
	double radang = 3.1415926 * coneSize / 360.0;
	float h = 20.0f;					// hypotenuse
	float d = -h * (float)cos(radang);	// dist from origin to cone circle
	float r = h * (float)sin(radang);	// radius of cone circle
	float s = 0.70711f * r;  			// sin(45) * r
	spotMesh.setVert(0, Point3( FZ, FZ, FZ));
	spotMesh.setVert(1, Point3( -r, FZ, d));
	spotMesh.setVert(2, Point3( -s, -s, d));
	spotMesh.setVert(3, Point3( FZ, -r, d));
	spotMesh.setVert(4, Point3(  s, -s, d));
	spotMesh.setVert(5, Point3(  r, FZ, d));
	spotMesh.setVert(6, Point3(  s,  s, d));
	spotMesh.setVert(7, Point3( FZ,  r, d));
	spotMesh.setVert(8, Point3( -s,  s, d));
	spotMesh.faces[0].setVerts(0,1,2);
	spotMesh.faces[1].setVerts(0,2,3);
	spotMesh.faces[2].setVerts(0,3,4);
	spotMesh.faces[3].setVerts(0,4,5);
	spotMesh.faces[4].setVerts(0,5,6);
	spotMesh.faces[5].setVerts(0,6,7);
	spotMesh.faces[6].setVerts(0,7,8);
	spotMesh.faces[7].setVerts(0,8,1);
	for (int i=0; i<nfaces; i++) {
		spotMesh.faces[i].setSmGroup(i);
		spotMesh.faces[i].setEdgeVisFlags(1,1,1);
	}
	spotMesh.buildNormals();
	spotMesh.EnableEdgeList(1);
}

void GeneralLight::UpdateColBrackets(TimeValue t) {
	BOOL isKey = pblock->KeyFrameAtTime(PB_COLOR,t); 
	redSpin->SetKeyBrackets(isKey);
	greenSpin->SetKeyBrackets(isKey);
	blueSpin->SetKeyBrackets(isKey);
	colorSwatch->SetKeyBrackets(isKey);
	hSpin->SetKeyBrackets(isKey);
	sSpin->SetKeyBrackets(isKey);
	vSpin->SetKeyBrackets(isKey);
	}

void GeneralLight::UpdateUI(TimeValue t)
	{
	Point3 color;

	if ( hGeneralLight && !waitPostLoad &&
		GetWindowLong(hGeneralLight,GWL_USERDATA)==(LONG)this) {
		color = GetRGBColor(t);
		redSpin->SetValue( FLto255f(color.x), FALSE );
		greenSpin->SetValue( FLto255f(color.y), FALSE );
		blueSpin->SetValue( FLto255f(color.z), FALSE );
		colorSwatch->SetColor(RGB(FLto255i(color.x), FLto255i(color.y), FLto255i(color.z)));

		if(updateHSVSpin) {
			color = GetHSVColor(t);
			hSpin->SetValue( FLto255f(color.x), FALSE );
			sSpin->SetValue( FLto255f(color.y), FALSE );
			vSpin->SetValue( FLto255f(color.z), FALSE );
			}

		UpdateColBrackets(t);

		intensitySpin->SetValue( GetIntensity(t), FALSE);
		intensitySpin->SetKeyBrackets(pblock->KeyFrameAtTime(PB_INTENSITY,t));

		contrastSpin->SetValue( GetContrast(t), FALSE);
		contrastSpin->SetKeyBrackets(pblock->KeyFrameAtTime(PB_CONTRAST,t));

		diffsoftSpin->SetValue(GetDiffuseSoft(t), FALSE);
		diffsoftSpin->SetKeyBrackets(pblock->KeyFrameAtTime(PB_DIFFSOFT,t));

		CheckDlgButton( hGeneralLight, IDC_LIGHT_ON, GetUseLight());
		SetDlgItemText(hGeneralLight,IDC_EXCLUDE_DLG, 
			GetString(exclList.TestFlag(NT_INCLUDE)?IDS_DS_INCLUDE__: IDS_DS_EXCLUDE__));

		if(attenStartSpin) {
			attenStartSpin->SetValue( GetAtten(t, ATTEN_START), FALSE);
			attenStartSpin->SetKeyBrackets(pblock->KeyFrameAtTime(type== OMNI_LIGHT? PB_OMNIATSTART:PB_ATTENSTART,t));
			}
		if(attenEndSpin) {
			attenEndSpin->SetValue( GetAtten(t, ATTEN_END), FALSE);
			attenEndSpin->SetKeyBrackets(pblock->KeyFrameAtTime(type== OMNI_LIGHT? PB_OMNIATEND:PB_ATTENEND,t));
			}
		CheckDlgButton( hGeneralLight, IDC_SHOW_RANGES, GetAttenDisplay() );
		CheckDlgButton( hGeneralLight, IDC_USE_ATTEN, GetUseAtten() );

		if(attenStart1Spin) {
			attenStart1Spin->SetValue( GetAtten(t, ATTEN1_START), FALSE);
			attenStart1Spin->SetKeyBrackets(pblock->KeyFrameAtTime(type== OMNI_LIGHT? PB_OMNIATSTART1:PB_ATTENSTART1,t));
			}
		if(attenEnd1Spin) {
			attenEnd1Spin->SetValue( GetAtten(t, ATTEN1_END), FALSE);
			attenEnd1Spin->SetKeyBrackets(pblock->KeyFrameAtTime(type== OMNI_LIGHT? PB_OMNIATEND1:PB_ATTENEND1,t));
			}
		CheckDlgButton( hGeneralLight, IDC_SHOW_RANGES1, GetAttenNearDisplay() );
		CheckDlgButton( hGeneralLight, IDC_USE_ATTEN1, GetUseAttenNear() );

		if( IsSpot() ) {
			float hot, fall;
			hotsizeSpin->SetValue( hot = GetHotspot(t), FALSE );
			hotsizeSpin->SetKeyBrackets(pblock->KeyFrameAtTime(PB_HOTSIZE,t));
			fallsizeSpin->SetValue( fall = GetFallsize(t), FALSE );
			fallsizeSpin->SetKeyBrackets(pblock->KeyFrameAtTime(PB_FALLSIZE,t));

			if (type == FSPOT_LIGHT||type==DIR_LIGHT) {
				tDistSpin->SetValue( GetTDist(t), FALSE );
				tDistSpin->SetKeyBrackets(pblock->KeyFrameAtTime(PB_TDIST,t));
				}

			aspectSpin->SetValue( GetAspect(t), FALSE );
			aspectSpin->SetKeyBrackets(pblock->KeyFrameAtTime(PB_ASPECT,t));
			CheckDlgButton( hSpotLight, IDC_SHOW_CONE, GetConeDisplay() );

			BuildSpotMesh(max(hot, fall));
			CheckDlgButton( hSpotLight, IDC_OVERSHOOT, GetOvershoot() ); 
			CheckDlgButton( hSpotLight, IDC_PROJECTOR, GetProjector() ); 
			}
	
		CheckRadioButton( hShadow, IDC_SHADOW_MAPS, IDC_RAY_TRACED_SHADOWS, 
			IDC_SHADOW_MAPS+GetShadowType());
		CheckDlgButton( hShadow, IDC_GLOBAL_SET, GetUseGlobal() ); 
		CheckDlgButton( hShadow, IDC_CAST_SHADOWS, GetShadow() );
		CheckDlgButton( hShadow, IDC_ABS_MAP_BIAS, GetAbsMapBias() ); 
		mapSizeSpin->SetValue( GetMapSize(t), FALSE );
		mapSizeSpin->SetKeyBrackets(pblock->KeyFrameAtTime(type==OMNI_LIGHT?PB_OMNIMAP_SIZE:PB_MAP_SIZE,t));
		mapBiasSpin->SetValue( GetMapBias(t), FALSE );
		mapBiasSpin->SetKeyBrackets(pblock->KeyFrameAtTime(type==OMNI_LIGHT?PB_OMNIMAP_BIAS:PB_MAP_BIAS,t));
	    mapRangeSpin->SetValue( GetMapRange(t), FALSE );
		mapRangeSpin->SetKeyBrackets(pblock->KeyFrameAtTime(type==OMNI_LIGHT?PB_OMNIMAP_RANGE:PB_MAP_RANGE,t));
		rayBiasSpin->SetValue( GetRayBias(t), FALSE );
		rayBiasSpin->SetKeyBrackets(pblock->KeyFrameAtTime(type==OMNI_LIGHT?PB_OMNIRAY_BIAS:PB_RAY_BIAS,t));

//		radiusSpin->SetValue( GetRadius(t), FALSE );
//		radiusSpin->SetKeyBrackets(pblock->KeyFrameAtTime(type==OMNI_LIGHT?PB_OMNIMAP_SIZE:PB_RADIUS,t));

//		numberLightsSpin->SetValue( GetNumberLights(t), FALSE );
//		numberLightsSpin->SetKeyBrackets(pblock->KeyFrameAtTime(type==OMNI_LIGHT?PB_OMNIMAP_SIZE:PB_NUMBER_LIGHTS,t));

		}
	}

#define LIGHT_VERSION 8		// current version

// description for version 0
static ParamBlockDescID descV0[] = {
	{ TYPE_FLOAT, NULL, TRUE, 1 },		// red
	{ TYPE_FLOAT, NULL, TRUE, 2 },		// green
	{ TYPE_FLOAT, NULL, TRUE, 3 },		// blue
	{ TYPE_FLOAT, NULL, TRUE, 7 } };	// hotsize

// description for version 1
static ParamBlockDescID descV1[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start
	{ TYPE_FLOAT, NULL, TRUE, 6 },		// atten end
	{ TYPE_FLOAT, NULL, TRUE, 7 },		// hotsize
	{ TYPE_FLOAT, NULL, TRUE, 8 },		// fallsize
	{ TYPE_FLOAT, NULL, TRUE, 9 },		// tdist
	{ TYPE_FLOAT, NULL, TRUE, 11 },		// aspect
	{ TYPE_FLOAT, NULL, TRUE, 12 } };	// blur


// description for version 2
static ParamBlockDescID descV2[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start
	{ TYPE_FLOAT, NULL, TRUE, 6 },		// atten end
	{ TYPE_FLOAT, NULL, TRUE, 7 },		// hotsize
	{ TYPE_FLOAT, NULL, TRUE, 8 },		// fallsize
	{ TYPE_FLOAT, NULL, TRUE, 9 },		// tdist
	{ TYPE_FLOAT, NULL, TRUE, 11 },		// aspect
	{ TYPE_FLOAT, NULL, TRUE, 12 },  	// blur
	{ TYPE_INT, NULL, TRUE, 13 },  	// mapSize
	{ TYPE_FLOAT, NULL, TRUE, 14 },  	// mapBias
	{ TYPE_FLOAT, NULL, TRUE, 15 },  	// mapRange
	{ TYPE_FLOAT, NULL, TRUE, 16 }};  	// raytrace bias

// description for version 3 -- all but omni
static ParamBlockDescID descV3[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 7 },		// hotsize
	{ TYPE_FLOAT, NULL, TRUE, 8 },		// fallsize
	{ TYPE_FLOAT, NULL, TRUE, 11 },		// aspect
	{ TYPE_FLOAT, NULL, TRUE, 12 },  	// blur
	{ TYPE_INT, NULL, TRUE, 13 },  	// mapSize
	{ TYPE_FLOAT, NULL, TRUE, 14 },  	// mapBias
	{ TYPE_FLOAT, NULL, TRUE, 15 },  	// mapRange
	{ TYPE_FLOAT, NULL, TRUE, 16 },  	// raytrace bias
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start		<- atten comes last because dir light doesn't have it
	{ TYPE_FLOAT, NULL, TRUE, 6 },		// atten end
	{ TYPE_FLOAT, NULL, TRUE, 9 }};		// tdist

// description for version 3 -- only omni
static ParamBlockDescID descV3Omni[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start
	{ TYPE_FLOAT, NULL, TRUE, 6 }};		// atten end

// description for versions 4 -- all but omni

static ParamBlockDescID descV4[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 7 },		// hotsize
	{ TYPE_FLOAT, NULL, TRUE, 8 },		// fallsize
	{ TYPE_FLOAT, NULL, TRUE, 11 },		// aspect
	{ TYPE_INT, NULL, TRUE, 13 },  	// mapSize
	{ TYPE_FLOAT, NULL, TRUE, 14 },  	// mapBias
	{ TYPE_FLOAT, NULL, TRUE, 15 },  	// mapRange
	{ TYPE_FLOAT, NULL, TRUE, 16 },  	// raytrace bias
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start		<- atten comes last because dir light doesn't have it
	{ TYPE_FLOAT, NULL, TRUE, 6 },		// atten end
	{ TYPE_FLOAT, NULL, TRUE, 9 }};		// tdist

// description for version 4 -- only omni
static ParamBlockDescID descV4Omni[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start
	{ TYPE_FLOAT, NULL, TRUE, 6 }};		// atten end

// description for versions 5 - all but omni
static ParamBlockDescID descV5[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 7 },		// hotsize
	{ TYPE_FLOAT, NULL, TRUE, 8 },		// fallsize
	{ TYPE_FLOAT, NULL, TRUE, 11 },		// aspect
	{ TYPE_INT,   NULL, TRUE, 13 },  	// mapSize
	{ TYPE_FLOAT, NULL, TRUE, 14 },  	// mapBias
	{ TYPE_FLOAT, NULL, TRUE, 15 },  	// mapRange
	{ TYPE_FLOAT, NULL, TRUE, 16 },  	// raytrace bias
	{ TYPE_FLOAT, NULL, TRUE, 17 },		// atten start near
	{ TYPE_FLOAT, NULL, TRUE, 18 },		// atten end near
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start 	
	{ TYPE_FLOAT, NULL, TRUE, 6 },		// atten end
	{ TYPE_FLOAT, NULL, TRUE, 9 }};		// tdist


// description for version 5 -- only omni
static ParamBlockDescID descV5Omni[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 17 },	 // atten start near
	{ TYPE_FLOAT, NULL, TRUE, 18 },	 // atten end near
	{ TYPE_FLOAT, NULL, TRUE, 5 },	 // atten start 
	{ TYPE_FLOAT, NULL, TRUE, 6 }};	 // atten end 


// description for versions 6 - all but omni
static ParamBlockDescID descV6[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 7 },		// hotsize
	{ TYPE_FLOAT, NULL, TRUE, 8 },		// fallsize
	{ TYPE_FLOAT, NULL, TRUE, 11 },		// aspect
	{ TYPE_INT,   NULL, TRUE, 13 },  	// mapSize
	{ TYPE_FLOAT, NULL, TRUE, 14 },  	// mapBias
	{ TYPE_FLOAT, NULL, TRUE, 15 },  	// mapRange
	{ TYPE_FLOAT, NULL, TRUE, 16 },  	// raytrace bias
	{ TYPE_FLOAT, NULL, TRUE, 17 },		// atten start near
	{ TYPE_FLOAT, NULL, TRUE, 18 },		// atten end near
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start 	
	{ TYPE_FLOAT, NULL, TRUE, 6 },		// atten end
	{ TYPE_FLOAT, NULL, TRUE, 9 }};		// tdist

// description for version 6 -- only omni
static ParamBlockDescID descV6Omni[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },		// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_INT,   NULL, TRUE, 13 }, 	// mapSize
	{ TYPE_FLOAT, NULL, TRUE, 14 }, 	// mapBias
	{ TYPE_FLOAT, NULL, TRUE, 15 }, 	// mapRange
	{ TYPE_FLOAT, NULL, TRUE, 16 }, 	// raytrace bias
	{ TYPE_FLOAT, NULL, TRUE, 17 },		// atten start near
	{ TYPE_FLOAT, NULL, TRUE, 18 },		// atten end near
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start 
	{ TYPE_FLOAT, NULL, TRUE, 6 }};		// atten end 


// description for versions 7 - all but omni
static ParamBlockDescID descV7[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 19},		// contrast
	{ TYPE_FLOAT, NULL, TRUE, 7 },		// hotsize
	{ TYPE_FLOAT, NULL, TRUE, 8 },		// fallsize
	{ TYPE_FLOAT, NULL, TRUE, 11 },		// aspect
	{ TYPE_INT,   NULL, TRUE, 13 },  	// mapSize
	{ TYPE_FLOAT, NULL, TRUE, 14 },  	// mapBias
	{ TYPE_FLOAT, NULL, TRUE, 15 },  	// mapRange
	{ TYPE_FLOAT, NULL, TRUE, 16 },  	// raytrace bias
	{ TYPE_FLOAT, NULL, TRUE, 17 },		// atten start near
	{ TYPE_FLOAT, NULL, TRUE, 18 },		// atten end near
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start 	
	{ TYPE_FLOAT, NULL, TRUE, 6 },		// atten end
	{ TYPE_FLOAT, NULL, TRUE, 9 }};		// tdist

// description for version 7 -- only omni
static ParamBlockDescID descV7Omni[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },		// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 19},		// contrast
	{ TYPE_INT,   NULL, TRUE, 13 }, 	// mapSize
	{ TYPE_FLOAT, NULL, TRUE, 14 }, 	// mapBias
	{ TYPE_FLOAT, NULL, TRUE, 15 }, 	// mapRange
	{ TYPE_FLOAT, NULL, TRUE, 16 }, 	// raytrace bias
	{ TYPE_FLOAT, NULL, TRUE, 17 },		// atten start near
	{ TYPE_FLOAT, NULL, TRUE, 18 },		// atten end near
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start 
	{ TYPE_FLOAT, NULL, TRUE, 6 }};		// atten end 


// description for versions 8 - all but omni
static ParamBlockDescID descV8[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },	// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 19},		// contrast
	{ TYPE_FLOAT, NULL, TRUE, 20},		// diffsoft
	{ TYPE_FLOAT, NULL, TRUE, 7 },		// hotsize
	{ TYPE_FLOAT, NULL, TRUE, 8 },		// fallsize
	{ TYPE_FLOAT, NULL, TRUE, 11 },		// aspect
	{ TYPE_INT,   NULL, TRUE, 13 },  	// mapSize
	{ TYPE_FLOAT, NULL, TRUE, 14 },  	// mapBias
	{ TYPE_FLOAT, NULL, TRUE, 15 },  	// mapRange
	{ TYPE_FLOAT, NULL, TRUE, 16 },  	// raytrace bias
	{ TYPE_FLOAT, NULL, TRUE, 17 },		// atten start near
	{ TYPE_FLOAT, NULL, TRUE, 18 },		// atten end near

	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start 	
	{ TYPE_FLOAT, NULL, TRUE, 6 },		// atten end
	{ TYPE_FLOAT, NULL, TRUE, 9 },	// tdist
};	

// description for version 8 -- only omni
static ParamBlockDescID descV8Omni[] = {
	{ TYPE_RGBA, NULL, TRUE, 10 },		// color
	{ TYPE_FLOAT, NULL, TRUE, 4 },		// intensity
	{ TYPE_FLOAT, NULL, TRUE, 19},		// contrast
	{ TYPE_FLOAT, NULL, TRUE, 20},		// diffsoft
	{ TYPE_INT,   NULL, TRUE, 13 }, 	// mapSize
	{ TYPE_FLOAT, NULL, TRUE, 14 }, 	// mapBias
	{ TYPE_FLOAT, NULL, TRUE, 15 }, 	// mapRange
	{ TYPE_FLOAT, NULL, TRUE, 16 }, 	// raytrace bias
	{ TYPE_FLOAT, NULL, TRUE, 17 },		// atten start near
	{ TYPE_FLOAT, NULL, TRUE, 18 },		// atten end near
	{ TYPE_FLOAT, NULL, TRUE, 5 },		// atten start 
	{ TYPE_FLOAT, NULL, TRUE, 6 }};		// atten end 

// number of parameters for different light types
static int oldDim[5] =  { 3,  4,  3,  4,  4};
static int oldDim1[5] = { 4,  9,  2,  9,  4};
static int oldDim2[5] = { 4, 13,  2, 13, 13};
static int oldDim3[5] = { 4, 12, 10, 13, 12};
static int oldDim4[5] = { 4, 11,  9, 12, 11};
static int oldDim5[5] = { 6, 13, 14, 14, 13};  
static int oldDim6[5] = { 10, 13, 14, 14, 13};  
static int oldDim7[5] = { 11, 14, 15, 15, 14};  
static int newDim8[5] = { 12, 15, 16, 16, 15};  

GeneralLight::GeneralLight(int tp)
{
	type = tp;
	MakeRefByID( FOREVER, 0, CreateParameterBlock( 
		type == OMNI_LIGHT ? descV8Omni : descV8, newDim8[type], LIGHT_VERSION ) );	

#ifdef USE_DLG_COLOR
	SetRGBColor( TimeValue(0), Point3(dlgRed/255.0f, dlgGreen/255.0f, dlgBlue/255.0f));
#else
	SetRGBColor( TimeValue(0), Point3(180.0f/255.0f, 180.0f/255.0f, 180.0f/255.0f));
#endif
	SetIntensity( TimeValue(0), dlgIntensity);	
	SetContrast( TimeValue(0), dlgContrast);
	SetDiffuseSoft( TimeValue(0), dlgDiffsoft);
	coneDisplay = 0;
	shape = CIRCLE_LIGHT;
	useLight = TRUE;
	attenDisplay = 0;
	useAtten = 0;
	attenNearDisplay = 0;
	useAttenNear = 0;
	enable=FALSE;
	shadow = FALSE;
	shadowType = 0;
	useGlobalShadowParams = FALSE;
	absMapBias = TRUE;
	overshoot = FALSE;
	projector= FALSE;
	projMap = NULL;
	updateHSVSpin = TRUE;
	softenDiffuse = FALSE;
	affectDiffuse = TRUE;
	affectSpecular = TRUE;
	decayType = DECAY_NONE;
	targDist = 0.0f;

	SetAtten( TimeValue(0), ATTEN1_START, dlgAtten1Start);
	SetAtten( TimeValue(0), ATTEN1_END, dlgAtten1End);
	SetAtten( TimeValue(0), ATTEN_START, dlgAttenStart);
	SetAtten( TimeValue(0), ATTEN_END, dlgAttenEnd);
	attenDisplay = dlgShowAtten;
	useAtten = dlgUseAtten;

//	SetRadius( TimeValue(0), dlgRadius);
//	SetNumberLights( TimeValue(0), dlgNumberLights);


	SetMapSize( TimeValue(0), dlgMapSize);
	SetMapBias( TimeValue(0), dlgMapBias);
	SetMapRange( TimeValue(0), dlgMapRange);
	SetRayBias( TimeValue(0), dlgRayBias);

	if( IsSpot() ) {
		SetHotspot( TimeValue(0), dlgHotsize);
		SetFallsize( TimeValue(0), dlgFallsize);
		if(dlgTDist < 1.0f)
			SetTDist( TimeValue(0), DEF_TDIST);
		else
			SetTDist( TimeValue(0), dlgTDist);
		SetAspect( TimeValue(0), dlgAspect);
		BuildSpotMesh(dlgHotsize);
		coneDisplay = dlgShowCone;
		shape = dlgShape;
		if(type == DIR_LIGHT||type == TDIR_LIGHT) {
			BuildStaticMeshes();
			mesh = &staticMesh[DIR_MESH];
			}
		else
			mesh = &spotMesh;
		}
	else {
		BuildStaticMeshes();
		mesh = &staticMesh[type == OMNI_LIGHT ? OMNI_MESH : DIR_MESH];
		}
	}

BOOL GeneralLight::SetHotSpotControl(Control *c)
	{
	pblock->SetController(PB_HOTSIZE,c);
	return TRUE;
	}

BOOL GeneralLight::SetFalloffControl(Control *c)
	{
	pblock->SetController(PB_FALLSIZE,c);
	return TRUE;
	}

BOOL GeneralLight::SetColorControl(Control *c)
	{
	pblock->SetController(PB_COLOR,c);
	return TRUE;
	}


Control* GeneralLight::GetHotSpotControl(){
	return	pblock->GetController(PB_HOTSIZE);
	}

Control* GeneralLight::GetFalloffControl(){
	return	pblock->GetController(PB_FALLSIZE);
	}

Control* GeneralLight::GetColorControl() {
	return	pblock->GetController(PB_COLOR);
	}

void GeneralLight::SetIntensity(TimeValue t, float f)
{
	pblock->SetValue( PB_INTENSITY, t, f );
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
}

float GeneralLight::GetIntensity(TimeValue t, Interval& valid)
{
	float f;
	pblock->GetValue( PB_INTENSITY, t, f, valid );
	return f;
}



void GeneralLight::SetContrast(TimeValue t, float f)
{
	pblock->SetValue( PB_CONTRAST, t, f );
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
}

float GeneralLight::GetContrast(TimeValue t, Interval& valid)
{
	float f;
	pblock->GetValue( PB_CONTRAST, t, f, valid );
	return f;
}

void GeneralLight::SetDiffuseSoft(TimeValue t, float f)
{
	pblock->SetValue( PB_DIFFSOFT, t, f );
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
}

float GeneralLight::GetDiffuseSoft(TimeValue t, Interval& valid)
{
	float f;
	pblock->GetValue( PB_DIFFSOFT, t, f, valid );
	return f;
}


void GeneralLight::SetHotspot(TimeValue t, float f)
{
	if(!IsSpot())
		return;
	if(f < 0.5f)
		f = 0.5f;
	if(!IsDir() && (f > 179.5f))
		f = 179.5f;
	pblock->SetValue( PB_HOTSIZE, t, f );
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
}

float GeneralLight::GetHotspot(TimeValue t, Interval& valid)
{
	Interval iv;

	if(!IsSpot())
		return -1.0f;
	float f;
	pblock->GetValue( PB_HOTSIZE, t, f, valid );
	if(GetFallsize(t, iv) < f )
		return GetFallsize(t, iv);
	return f;
}

void GeneralLight::SetFallsize(TimeValue t, float f)
{
	if(!IsSpot())
		return;
	if(f < 0.5f)
		f = 0.5f;
	if(!IsDir() && (f > 179.5f))
		f = 179.5f;
	pblock->SetValue( PB_FALLSIZE, t, f );
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
}

float GeneralLight::GetFallsize(TimeValue t, Interval& valid)
{
	if(!IsSpot())
		return -1.0f;
	float f;
	pblock->GetValue( PB_FALLSIZE, t, f, valid );
	return f;
}

void GeneralLight::SetAtten(TimeValue t, int which, float f)
{
//	if(type == DIR_LIGHT)
//		return;
	pblock->SetValue( (type == OMNI_LIGHT ? PB_OMNIATSTART1 : PB_ATTENSTART1) + which, t, f );
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
}

float GeneralLight::GetAtten(TimeValue t, int which, Interval& valid)
{
//	if(type == DIR_LIGHT)
//		return -1.0f;
	float f;
	pblock->GetValue( (type == OMNI_LIGHT ? PB_OMNIATSTART1 : PB_ATTENSTART1) + which, t, f, valid );
	return f;
}

void GeneralLight::SetTDist(TimeValue t, float f)
{
	if(type != FSPOT_LIGHT&&type != DIR_LIGHT)
		return;
	pblock->SetValue( PB_TDIST, t, f );
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
}

float GeneralLight::GetTDist(TimeValue t, Interval& valid)
{
	if(!IsSpot())
		return -1.0f;
	if( type == TDIR_LIGHT || type == TSPOT_LIGHT)
		return 0.0f;
	float f;
	pblock->GetValue( PB_TDIST, t, f, valid );
	return f;
}


void GeneralLight::SetRGBColor(TimeValue t, Point3 &rgb) 
{
	pblock->SetValue( PB_COLOR, t, rgb );
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
}

Point3 GeneralLight::GetRGBColor(TimeValue t, Interval& valid) 
{	
	Point3 rgb;
	pblock->GetValue( PB_COLOR, t, rgb, valid );
	return rgb;
}

void GeneralLight::SetHSVColor(TimeValue t, Point3& hsv) 
{
	DWORD rgb = HSVtoRGB ((int)(hsv[0]*255.0f), (int)(hsv[1]*255.0f), (int)(hsv[2]*255.0f));
	Point3 rgbf;
	rgbf[0] = GetRValue(rgb) / 255.0f;
	rgbf[1] = GetGValue(rgb) / 255.0f;
	rgbf[2] = GetBValue(rgb) / 255.0f;
	SetRGBColor(t, rgbf);
}

Point3 GeneralLight::GetHSVColor(TimeValue t, Interval& valid) 
{	
	int h, s, v;
	Point3 rgbf = GetRGBColor(t, valid);
	DWORD rgb = RGB((int)(rgbf[0]*255.0f), (int)(rgbf[1]*255.0f), (int)(rgbf[2]*255.0f));
	RGBtoHSV (rgb, &h, &s, &v);
	return Point3(h/255.0f, s/255.0f, v/255.0f);
}

void GeneralLight::SetUseLight(int onOff)
{
	useLight = onOff;
	NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
}


void GeneralLight::SetConeDisplay(int s, int notify) 
{
	coneDisplay = s;
	if(notify && IsSpot())
		NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
}

void GeneralLight::SetOvershoot(int a)
{
	overshoot = a;
	if(IsSpot())
		NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
}

void GeneralLight::SetSpotShape(int s)
{
	shape = s;
	if(IsSpot())
		NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
}

void GeneralLight::SetAspect(TimeValue t, float f)
{
	if(!IsSpot())
		return;
	pblock->SetValue( PB_ASPECT, t, f );
	NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
}

float GeneralLight::GetAspect(TimeValue t, Interval& valid)
{
	if(!IsSpot())
		return -1.0f;
	float f;
	pblock->GetValue( PB_ASPECT, t, f, valid );
	return f;
}

void GeneralLight::SetMapBias(TimeValue t, float f) {
//	if(!IsSpot())	return;
	if (useGlobalShadowParams) globMapBias = f;
	else {
		pblock->SetValue( type==OMNI_LIGHT?PB_OMNIMAP_BIAS:PB_MAP_BIAS, t, f );
		//NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
		}
	}

float GeneralLight::GetMapBias(TimeValue t, Interval& valid) {
//	if(!IsSpot()) return 0.0f;
	if (useGlobalShadowParams)  return globMapBias;
	float f;
	pblock->GetValue( type==OMNI_LIGHT?PB_OMNIMAP_BIAS:PB_MAP_BIAS, t, f, valid );
	return f;
	}

void GeneralLight::SetMapRange(TimeValue t, float f) {
//	if(!IsSpot())	return;
	if (useGlobalShadowParams) globMapRange = f;
	else {
		pblock->SetValue( type==OMNI_LIGHT?PB_OMNIMAP_RANGE:PB_MAP_RANGE, t, f );
		//NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
		}
	}

float GeneralLight::GetMapRange(TimeValue t, Interval& valid){
//	if(!IsSpot()) return 0.0f;
	if (useGlobalShadowParams)  return globMapRange;
	float f;
	pblock->GetValue( type==OMNI_LIGHT?PB_OMNIMAP_RANGE:PB_MAP_RANGE, t, f, valid );
	return f;
	}

void GeneralLight::SetMapSize(TimeValue t, int f) {
//	if(!IsSpot())	return;
	if (useGlobalShadowParams)  globMapSize = f;
	else {
		pblock->SetValue( type==OMNI_LIGHT?PB_OMNIMAP_SIZE:PB_MAP_SIZE, t, f );
		//NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
		}
	}

int GeneralLight::GetMapSize(TimeValue t, Interval& valid){
//	if(!IsSpot()) return 0;
	if (useGlobalShadowParams)  return globMapSize;
	int f;
	pblock->GetValue( type==OMNI_LIGHT?PB_OMNIMAP_SIZE:PB_MAP_SIZE, t, f, valid );
	return f;
	}


/*
void GeneralLight::SetRadius(TimeValue t, float f) {
//	if(!IsSpot())	return;
//	if (useGlobalShadowParams)  globMapSize = f;
//	else {
		pblock->SetValue( PB_RADIUS, t, f );
		//NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
//		}
	}

float GeneralLight::GetRadius(TimeValue t, Interval& valid){
//	if(!IsSpot()) return 0;
//	if (useGlobalShadowParams)  return globMapSize;
	float f;
	pblock->GetValue(PB_RADIUS, t, f, valid );
	return f;
	}



void GeneralLight::SetNumberLights(TimeValue t, int f) {
//	if(!IsSpot())	return;
//	if (useGlobalShadowParams)  globMapSize = f;
//	else {
//		pblock->SetValue( PB_NUMBER_LIGHTS, t, f );
		//NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
//		}
	}

int GeneralLight::GetNumberLights(TimeValue t, Interval& valid){
//	if(!IsSpot()) return 0;
//	if (useGlobalShadowParams)  return globMapSize;
	int f;
//	pblock->GetValue(PB_NUMBER_LIGHTS, t, f, valid );
	return f;
	}
*/

void GeneralLight::SetRayBias(TimeValue t, float f) {
//	if(!IsSpot())	return;
	if (useGlobalShadowParams)  globRayBias = f;
	else {
		pblock->SetValue( type==OMNI_LIGHT?PB_OMNIRAY_BIAS:PB_RAY_BIAS, t, f );
		//NotifyDependents(FOREVER, PART_ALL, REFMSG_CHANGE);
		}
	}

float GeneralLight::GetRayBias(TimeValue t, Interval& valid){
//	if(!IsSpot()) return 0.0f;
	if (useGlobalShadowParams)  return globRayBias;
	float f;
	pblock->GetValue( type==OMNI_LIGHT?PB_OMNIRAY_BIAS:PB_RAY_BIAS, t, f, valid );
	return f;
	}



// aspect <= 0   ==>  circular cone	(NUM_CIRC_PTS returned)
// aspect  > 0	 ==>  rectangular cone (4+2 points returned)
void GeneralLight::GetConePoints(TimeValue t, float aspect, float angle, float dist, Point3 *q) 
	{
	float ta = (float)tan(0.5*DegToRad(angle));   
	if(aspect <= 0.0f) { 
		// CIRCULAR
		float rad = dist * ta;
		double a;
		if(IsDir())
			rad = angle;
		for(int i = 0; i < NUM_CIRC_PTS; i++) {
			a = (double)i * 2.0 * 3.1415926 / (double)NUM_CIRC_PTS;
			q[i] = Point3(rad*(float)sin(a), rad*(float)cos(a), -dist);
			}
		q[i] = q[0] + Point3(0.0f, 15.0f, 0.0f);
		}
	else {		 
		// RECTANGULAR
		float w = IsDir()? angle : dist * ta * (float)sqrt((double)aspect);
		float h = w / aspect;
		q[0] = Point3( w, h,-dist);				
		q[1] = Point3(-w, h,-dist);				
		q[2] = Point3(-w,-h,-dist);				
		q[3] = Point3( w,-h,-dist);
		q[4] = Point3( 0.0f, h+15.0f, -dist);
		q[5] = Point3( 0.0f, h, -dist);
		}
	}

#define HOTCONE		0
#define FALLCONE	1

void GeneralLight::DrawCone(TimeValue t, GraphicsWindow *gw, float dist) 
	{
	Point3 q[NUM_CIRC_PTS+1], u[3];
	int dirLight = IsDir();
	int i;
	BOOL dispAtten = (useAtten && (extDispFlags & EXT_DISP_ONLY_SELECTED))?TRUE:attenDisplay;
	BOOL dispAttenNear = (useAttenNear && (extDispFlags & EXT_DISP_ONLY_SELECTED))?TRUE:attenNearDisplay;

	GetConePoints(t, GetSpotShape() ? -1.0f : GetAspect(t), GetHotspot(t), dist, q);
	gw->setColor( LINE_COLOR, GetUIColor(COLOR_HOTSPOT));
	if(GetSpotShape()) {  
		// CIRCULAR
		if(GetHotspot(t) >= GetFallsize(t)) {
			// draw (far) hotspot circle
			u[0] = q[0];
			u[1] = q[NUM_CIRC_PTS];
			gw->polyline( 2, u, NULL, NULL, FALSE, NULL);
			}
		gw->polyline(NUM_CIRC_PTS, q, NULL, NULL, TRUE, NULL);
		if (dirLight) {
			// draw 4 axial hotspot lines
			for (i = 0; i < NUM_CIRC_PTS; i += SEG_INDEX) {
				u[0] =  q[i]; 	u[1] =  q[i]; u[1].z += dist;
				gw->polyline( 2, u, NULL, NULL, FALSE, NULL );	
				}
			GetConePoints(t, -1.0f, GetHotspot(t), 0.0f, q);
			// draw (near) hotspot circle
			gw->polyline(NUM_CIRC_PTS, q, NULL, NULL, TRUE, NULL);
			}
		else  {
			// draw 4 axial lines
			u[0] = Point3(0,0,0);
			for (i = 0; i < NUM_CIRC_PTS; i += SEG_INDEX) {
				u[1] =  q[i];
				gw->polyline( 2, u, NULL, NULL, FALSE, NULL );	
				}
			}
	
		GetConePoints(t, -1.0f, GetFallsize(t), dist, q);
		gw->setColor( LINE_COLOR, GetUIColor(COLOR_FALLOFF));
		if(GetHotspot(t) < GetFallsize(t)) {
			// draw (far) fallsize circle
			u[0] = q[0];	u[1] = q[NUM_CIRC_PTS];
			gw->polyline( 2, u, NULL, NULL, FALSE, NULL);
			u[0] = Point3(0,0,0);
			}
		gw->polyline(NUM_CIRC_PTS, q, NULL, NULL, TRUE, NULL);
		if (dirLight) {
			float dfar = q[0].z;
			float dnear = 0.0f;
			if (dispAtten) {
				dfar  = MinF(-GetAtten(t,ATTEN_END),dfar);
				dnear = MaxF(-GetAtten(t,ATTEN_START),dnear);
				}
			if (dispAttenNear) {
				dfar  = MinF(-GetAtten(t,ATTEN1_END),dfar);
				dnear = MaxF(-GetAtten(t,ATTEN1_START),dnear);
				}
			// draw axial fallsize lines
			for (i = 0; i < NUM_CIRC_PTS; i += SEG_INDEX) {
				u[0] =  q[i];  u[0].z = dfar;	u[1] =  q[i]; u[1].z = dnear;
				gw->polyline( 2, u, NULL, NULL, FALSE, NULL );	
				}
			GetConePoints(t, -1.0f, GetFallsize(t), 0.0f, q);
			// draw (near) fallsize circle
			gw->polyline(NUM_CIRC_PTS, q, NULL, NULL, TRUE, NULL);
			}
		else {
			float cs = (float)cos(DegToRad(GetFallsize(t)*0.5f));
			float dfar = q[0].z;
			if (dispAtten) 
				dfar  = MinF(-cs*GetAtten(t,ATTEN_END),dfar);
			if (dispAttenNear) 
				dfar  = MinF(-cs*GetAtten(t,ATTEN1_END),dfar);
			for (i = 0; i < NUM_CIRC_PTS; i += SEG_INDEX) {
				u[1] =  -q[i]*dfar/dist;	
				gw->polyline( 2, u, NULL, NULL, FALSE, NULL );	
				}
			}
		}
	else { 
		// RECTANGULAR
		if(GetHotspot(t) >= GetFallsize(t))
			gw->polyline( 2, q+4, NULL, NULL, FALSE, NULL);
		gw->polyline(4, q, NULL, NULL, TRUE, NULL);
		if (dirLight) { //DIRLIGHT
			// draw axial hotspot lines
			for (i = 0; i < 4; i += 1) {
				u[0] =  q[i]; 	u[1] =  q[i]; u[1].z += dist;
				gw->polyline( 2, u, NULL, NULL, FALSE, NULL );	
				}
			GetConePoints(t, GetAspect(t), GetHotspot(t), 0.0f, q);
			// draw (near) hotspot circle
			gw->polyline(4, q, NULL, NULL, TRUE, NULL);
			}
		else {
			u[0] = Point3(0,0,0);
			for (i = 0; i < 4; i++) {
				u[1] =  q[i];
				gw->polyline( 2, u, NULL, NULL, FALSE, NULL );	
				}
			}
		GetConePoints(t, GetAspect(t), GetFallsize(t), dist, q);
		gw->setColor( LINE_COLOR, GetUIColor(COLOR_FALLOFF));
		if(GetHotspot(t) < GetFallsize(t))
			gw->polyline( 2, q+4, NULL, NULL, FALSE, NULL);
		gw->polyline(4, q, NULL, NULL, TRUE, NULL);
		if (dirLight) {
			// draw axial fallsize lines
			for (i = 0; i < 4; i += 1) {
				u[0] =  q[i]; 	u[1] =  q[i]; u[1].z += dist;
				gw->polyline( 2, u, NULL, NULL, FALSE, NULL );	
				}
			GetConePoints(t, GetAspect(t), GetFallsize(t), 0.0f, q);
			// draw (near) fallsize circle
			gw->polyline(4, q, NULL, NULL, TRUE, NULL);
			}
		else {
			for (i = 0; i < 4; i += 1) {
				u[1] =  q[i];
				gw->polyline( 2, u, NULL, NULL, FALSE, NULL );	
				}
			}
		}
	}

int GeneralLight::DrawConeAndLine(TimeValue t, INode* inode, GraphicsWindow *gw, int drawing ) 
	{
	if(!IsSpot())
		return 0;
	Matrix3 tm = inode->GetObjectTM(t);
	gw->setTransform(tm);
	if (type == TSPOT_LIGHT) {
		Point3 pt,v[3];
		if (GetTargetPoint(t, inode, pt)) {
			float den = FLength(tm.GetRow(2));
			float dist = (den!=0) ? FLength(tm.GetTrans()-pt) / den : 0.0f;
			targDist = dist;
			if (hSpotLight) {
				TCHAR buf[40];
				_stprintf(buf,_T("%0.3f"),targDist);
				SetWindowText(GetDlgItem(hSpotLight,IDC_TARG_DISTANCE),buf);
				}
			if ((drawing != -1) && (coneDisplay || (extDispFlags & EXT_DISP_ONLY_SELECTED))) 
				DrawCone(t, gw, dist);
			if(!inode->IsFrozen())
				gw->setColor( LINE_COLOR, GetUIColor(COLOR_TARGET_LINE));
			v[0] = Point3(0,0,0);
			if(drawing == -1)	// hittesting
				v[1] = Point3(0.0f, 0.0f, -0.9f * dist);
			else
				v[1] = Point3(0.0f, 0.0f, -dist);
			gw->polyline( 2, v, NULL, NULL, FALSE, NULL );	
			}
		}
	else if (type == FSPOT_LIGHT) {
		if ((drawing != -1) && (coneDisplay || (extDispFlags & EXT_DISP_ONLY_SELECTED)))
			DrawCone(t, gw, GetTDist(t));
	}
	else if (type == DIR_LIGHT) {
		if ((drawing != -1) && (coneDisplay || (extDispFlags & EXT_DISP_ONLY_SELECTED)))
			DrawCone(t, gw, GetTDist(t));
	}
	else if (type == TDIR_LIGHT) {
		Point3 pt,v[3];
		if (GetTargetPoint(t, inode, pt)) {
			float den = FLength(tm.GetRow(2));
			float dist = (den!=0) ? FLength(tm.GetTrans()-pt) / den : 0.0f;
			targDist = dist;
			if (hSpotLight) {
				TCHAR buf[40];
				_stprintf(buf,_T("%0.3f"),targDist);
				SetWindowText(GetDlgItem(hSpotLight,IDC_TARG_DISTANCE),buf);
				}
			if ((drawing != -1) && (coneDisplay || (extDispFlags & EXT_DISP_ONLY_SELECTED)))
				DrawCone(t, gw, dist);
			v[0] = Point3(0,0,0);
			if (drawing == -1) 
				v[1] = Point3(0.0f, 0.0f, -0.9f * dist);
			else
				v[1] = Point3(0.0f, 0.0f, -dist);
			gw->polyline( 2, v, NULL, NULL, FALSE, NULL );	
		}
	}
	return gw->checkHitCode();
}

void GeneralLight::SetUseAtten(int s) 
{
	useAtten = s;
	NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
}

void GeneralLight::SetUseAttenNear(int s) 
	{
	useAttenNear = s;
	NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
	}

void GeneralLight::SetAttenDisplay(int s) 
	{
	attenDisplay = s;
	NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
	}

void GeneralLight::SetAttenNearDisplay(int s) 
	{
	attenNearDisplay = s;
	NotifyDependents(FOREVER, PART_OBJ, REFMSG_CHANGE);
	}

void GeneralLight::GetAttenPoints(TimeValue t, float rad, Point3 *q)
	{
	double a;
	float sn, cs;
	for(int i = 0; i < NUM_CIRC_PTS; i++) {
		a = (double)i * 2.0 * 3.1415926 / (double)NUM_CIRC_PTS;
		sn = rad * (float)sin(a);
		cs = rad * (float)cos(a);
		q[i+0*NUM_CIRC_PTS] = Point3(sn, cs, 0.0f);
		q[i+1*NUM_CIRC_PTS] = Point3(sn, 0.0f, cs);
		q[i+2*NUM_CIRC_PTS] = Point3(0.0f, sn, cs);
		}
	}

static void DrawX(GraphicsWindow *gw, Point3 *q, int indx) {
	Point3 u[3];
	u[0] = q[0]; u[1] = q[2*indx];
	gw->polyline(2, u,NULL, NULL, FALSE, NULL);
	u[0] = q[indx]; u[1] = q[3*indx];
	gw->polyline(2, u,NULL, NULL, FALSE, NULL);
	}

int GeneralLight::GetCirXPoints(TimeValue t, float angle, float dist, Point3 *q) {
	int i;
	float ang = DegToRad(angle)/2.0f;
	float da = ang/float(NUM_HALF_ARC);
	// first draw circle:
	float d = dist*(float)cos(ang);
	GetConePoints(t, -1.0f, angle, d, q);
	int j=NUM_CIRC_PTS;
	// then draw Arc X
	float a = -ang;
	for(i = -NUM_HALF_ARC; i<= NUM_HALF_ARC; i++, a+=da) 
		q[j++] = Point3(0.0f, dist*(float)sin(a), -dist*(float)cos(a));
	a = -ang;	
	for(i = -NUM_HALF_ARC; i<= NUM_HALF_ARC; i++, a+=da) 
		q[j++] = Point3(dist*(float)sin(a), 0.0f, -dist*(float)cos(a));
	return NUM_CIRC_PTS + 2*NUM_ARC_PTS;
	}


void GeneralLight::DrawSphereArcs(TimeValue t, GraphicsWindow *gw, float r, Point3 *q) {
	GetAttenPoints(t, r, q);
	gw->polyline(NUM_CIRC_PTS, q,				NULL, NULL, TRUE, NULL);
	gw->polyline(NUM_CIRC_PTS, q+NUM_CIRC_PTS,	NULL, NULL, TRUE, NULL);
	gw->polyline(NUM_CIRC_PTS, q+2*NUM_CIRC_PTS,NULL, NULL, TRUE, NULL);
	}

void GeneralLight::DrawCircleX(TimeValue t, GraphicsWindow *gw, float angle, float dist, Point3 *q) {
 	GetCirXPoints(t, angle,dist,q);
	gw->polyline(NUM_CIRC_PTS, q,NULL, NULL, TRUE, NULL);  // circle 
	gw->polyline(NUM_ARC_PTS, q+NUM_CIRC_PTS,NULL, NULL, FALSE, NULL); // vert arc
	gw->polyline(NUM_ARC_PTS, q+NUM_CIRC_PTS+NUM_ARC_PTS,NULL, NULL, FALSE, NULL);  // horiz arc
	}


// Draw the the arcs cut out of the sphere by the viewing cone.
int GeneralLight::GetRectXPoints(TimeValue t, float angle, float dist, Point3 *q) {
	int i;
	if(dist==0.0f) dist = .00001f;
	float ang = DegToRad(angle)/2.0f;
	float da,sn,cs,x,y,z,a;
	float aspect = GetAspect(t);
	float w = dist * (float)tan(ang) * (float)sqrt((double)aspect);
	float h = w/aspect;
	float wang = (float)atan(w/dist);
	float hang = (float)atan(h/dist);
	float aw = float(atan(w/dist)*cos(hang));  // half-angle of top and bottom arcs
	float ah = float(atan(h/dist)*cos(wang));  // half-angle of left and right arcs
	int j = 0;

	// draw horizontal and vertical center lines
	da = wang/float(NUM_HALF_ARC);
	for(i = -NUM_HALF_ARC, a = -wang; i<= NUM_HALF_ARC; i++, a+=da) 
		q[j++] = Point3(dist*(float)sin(a), 0.0f, -dist*(float)cos(a));
	da = hang/float(NUM_HALF_ARC);
	for(i = -NUM_HALF_ARC, a = -hang; i<= NUM_HALF_ARC; i++, a+=da) 
		q[j++] = Point3(0.0f, dist*(float)sin(a), -dist*(float)cos(a));


	// draw top and bottom arcs
	da = aw/float(NUM_HALF_ARC);
	sn = (float)sin(hang);
	cs = (float)cos(hang);
	for (i = -NUM_HALF_ARC, a = -aw; i<= NUM_HALF_ARC; i++, a+=da) {
		x =  dist*(float)sin(a); 
		z = -dist*(float)cos(a);
		q[j]             = Point3(x, z*sn, z*cs);  				
		q[j+NUM_ARC_PTS] = Point3(x,-z*sn, z*cs);  				
		j++;
		}
	
	j+= NUM_ARC_PTS;

	// draw left and right arcs
	da = ah/float(NUM_HALF_ARC);
	sn = (float)sin(wang);
	cs = (float)cos(wang);
	for (i = -NUM_HALF_ARC, a = -ah; i<= NUM_HALF_ARC; i++, a+=da) {
		y =  dist*(float)sin(a); 
		z = -dist*(float)cos(a);
		q[j]             = Point3( z*sn, y, z*cs);  				
		q[j+NUM_ARC_PTS] = Point3(-z*sn, y, z*cs);  				
		j++;
		}

	return 6*NUM_ARC_PTS;
	}


// Draw warped rectangle
void GeneralLight::DrawWarpRect(TimeValue t, GraphicsWindow *gw, float angle, float dist, Point3 *q) {
 	GetRectXPoints(t, angle,dist,q);
	for (int i=0; i<6; i++)
		gw->polyline(NUM_ARC_PTS, q+i*NUM_ARC_PTS,NULL, NULL, FALSE, NULL);  
	}

int GeneralLight::DrawAtten(TimeValue t, INode *inode, GraphicsWindow *gw)
	{
	BOOL dispAtten = (useAtten && (extDispFlags & EXT_DISP_ONLY_SELECTED))?TRUE:attenDisplay;
	BOOL dispAttenNear = (useAttenNear && (extDispFlags & EXT_DISP_ONLY_SELECTED))?TRUE:attenNearDisplay;

	if (dispAtten||dispAttenNear) {
		Point3 q[3*NUM_CIRC_PTS+1];
		Matrix3 tm = inode->GetObjectTM(t);
		gw->setTransform(tm);
		BOOL froze = inode->IsFrozen();
		if(!froze)
			gw->setColor( LINE_COLOR, GetUIColor(COLOR_START_RANGE));
		if (IsDir()) {
			int npts, indx;
			float asp;
			if (GetSpotShape()) { npts = NUM_CIRC_PTS; 	asp  = -1.0f; 	indx = SEG_INDEX; 	}
			else { 	npts = 4;  	asp  = GetAspect(t); 	indx = 1; 	} 
			if (dispAtten) {
				GetConePoints(t, asp, GetFallsize(t), GetAtten(t,ATTEN_START), q);
				gw->polyline(npts, q,NULL, NULL, TRUE, NULL);
				DrawX(gw,q,indx);
				if (!froze)	gw->setColor( LINE_COLOR, GetUIColor(COLOR_END_RANGE));
				GetConePoints(t, asp, GetFallsize(t), GetAtten(t,ATTEN_END), q);
				gw->polyline(npts, q,NULL, NULL, TRUE, NULL);
				DrawX(gw,q,indx);
				}
			if (dispAttenNear) {
				if(!froze)	gw->setColor( LINE_COLOR, GetUIColor(COLOR_START_RANGE1));
				GetConePoints(t, asp, GetFallsize(t), GetAtten(t,ATTEN1_START), q);
				gw->polyline(npts, q,NULL, NULL, TRUE, NULL);
				DrawX(gw,q,indx);
				if (!froze)	gw->setColor( LINE_COLOR, GetUIColor(COLOR_END_RANGE1));
				GetConePoints(t, asp, GetFallsize(t), GetAtten(t,ATTEN1_END), q);
				gw->polyline(npts, q,NULL, NULL, TRUE, NULL);
				DrawX(gw,q,indx);
				}
			}
		else {
			if (type==OMNI_LIGHT||(IsSpot()&&overshoot)){ 
				// only draw the spheres for omni's and for spotlights with overshoot on.
				if (dispAtten) {
					DrawSphereArcs(t, gw, GetAtten(t, ATTEN_START), q);
					if (!froze) gw->setColor( LINE_COLOR, GetUIColor(COLOR_END_RANGE));
					DrawSphereArcs(t, gw, GetAtten(t, ATTEN_END), q);
					}
				if (dispAttenNear) {
					if(!froze)	gw->setColor( LINE_COLOR, GetUIColor(COLOR_START_RANGE1));
					DrawSphereArcs(t, gw, GetAtten(t, ATTEN1_START), q);
					if (!froze) gw->setColor( LINE_COLOR, GetUIColor(COLOR_END_RANGE1));
					DrawSphereArcs(t, gw, GetAtten(t, ATTEN1_END), q);
					}
				//return gw->checkHitCode();
				}
			else { 
				if (GetSpotShape()) { 
					if (dispAtten) {
						DrawCircleX(t, gw, GetFallsize(t), GetAtten(t,ATTEN_START),q);
						if (!froze) gw->setColor( LINE_COLOR, GetUIColor(COLOR_END_RANGE));
						DrawCircleX(t, gw, GetFallsize(t), GetAtten(t,ATTEN_END),q);
						}
					if (dispAttenNear) {
						if (!froze)	gw->setColor( LINE_COLOR, GetUIColor(COLOR_START_RANGE1));
						DrawCircleX(t, gw, GetFallsize(t), GetAtten(t,ATTEN1_START),q);
						if (!froze) gw->setColor( LINE_COLOR, GetUIColor(COLOR_END_RANGE1));
						DrawCircleX(t, gw, GetFallsize(t), GetAtten(t,ATTEN1_END),q);
						}
					}
				else { 
					if (dispAtten) {
						DrawWarpRect(t, gw, GetFallsize(t), GetAtten(t,ATTEN_START),q);
						if (!froze) gw->setColor( LINE_COLOR, GetUIColor(COLOR_END_RANGE));
						DrawWarpRect(t, gw, GetFallsize(t), GetAtten(t,ATTEN_END),q);
						}
					if (dispAttenNear) {
						if (!froze)	gw->setColor( LINE_COLOR, GetUIColor(COLOR_START_RANGE1));
						DrawWarpRect(t, gw, GetFallsize(t), GetAtten(t,ATTEN1_START),q);
						if (!froze) gw->setColor( LINE_COLOR, GetUIColor(COLOR_END_RANGE1));
						DrawWarpRect(t, gw, GetFallsize(t), GetAtten(t,ATTEN1_END),q);
						}
					}
				}
			}
		}
	return 0;
	}

GeneralLight::~GeneralLight() 
	{
	DeleteAllRefsFromMe();
	pblock = NULL;
	}

class GeneralLightCreateCallBack: public CreateMouseCallBack 
{
	GeneralLight *ob;
public:
	int proc( ViewExp *vpt,int msg, int point, int flags, IPoint2 m, Matrix3& mat);
	void SetObj(GeneralLight *obj) { ob = obj; }
};

int GeneralLightCreateCallBack::proc(ViewExp *vpt,int msg, int point, int flags, IPoint2 m, Matrix3& mat) 
{
	Point3 p0;
	#ifdef _OSNAP
	if (msg == MOUSE_FREEMOVE)
	{
		#ifdef _3D_CREATE
			vpt->SnapPreview(m,m,NULL, SNAP_IN_3D);
		#else
			vpt->SnapPreview(m,m,NULL, SNAP_IN_PLANE);
		#endif
	}
	#endif

	if (msg==MOUSE_POINT || msg==MOUSE_MOVE) {		
		//mat.SetTrans(vpt->GetPointOnCP(m));
		#ifdef _3D_CREATE
			mat.SetTrans( vpt->SnapPoint(m,m,NULL,SNAP_IN_3D) );
		#else
			mat.SetTrans( vpt->SnapPoint(m,m,NULL,SNAP_IN_PLANE) );
		#endif

		ob->enable = TRUE;
		if (point==1) {
			if (msg==MOUSE_POINT) 
					return 0;
		}
	}
	else if (msg == MOUSE_ABORT)
		return CREATE_ABORT;

	return TRUE;
}

static GeneralLightCreateCallBack sGeneralLgtCreateCB;

CreateMouseCallBack* GeneralLight::GetCreateMouseCallBack() 
{
	sGeneralLgtCreateCB.SetObj(this);

	return &sGeneralLgtCreateCB;
}


void GeneralLight::GetMat(TimeValue t, INode* inode, ViewExp *vpt, Matrix3& tm) 
{
	tm = inode->GetObjectTM(t);
	tm.NoScale();
	float scaleFactor = vpt->NonScalingObjectSize()*vpt->GetVPWorldWidth(tm.GetTrans())/(float)360.0;
	tm.Scale(Point3(scaleFactor,scaleFactor,scaleFactor));
}

void GeneralLight::GetDeformBBox(TimeValue t, Box3& box, Matrix3 *tm, BOOL useSel )
{
	box = mesh->getBoundingBox(tm);
}


void GeneralLight::BoxCircle(TimeValue t, float r, float d, Box3& box, Point3 *q, int extraPt, Matrix3 *tm) {
	int npts;
	float asp;
	if ( GetSpotShape()) { 	npts =  NUM_CIRC_PTS+extraPt; 	asp = -1.0f; }
	else { npts = 4+extraPt;  asp = GetAspect(t); } 
 	GetConePoints(t, asp , r, d, q);
 	box.IncludePoints(q,npts,tm);
	}

void GeneralLight::BoxLight(TimeValue t, INode *inode, Box3& box, Matrix3 *tm) {
	Point3 pt;
	float d;
	int npts;
	if (GetTargetPoint(t, inode, pt)) {
		Point3 loc = inode->GetObjectTM(t).GetTrans();
		d = FLength(loc - pt) / FLength(inode->GetObjectTM(t).GetRow(2));
		box += tm? (*tm)*Point3(0.0f, 0.0f, -d): Point3(0.0f, 0.0f, -d);
		}
	else {
		d = GetTDist(t);
		if (coneDisplay) 
			box += tm? (*tm)*Point3(0.0f, 0.0f, -d): Point3(0.0f, 0.0f, -d);
		}
	if( coneDisplay || (extDispFlags & EXT_DISP_ONLY_SELECTED)) {
		Point3 q[NUM_CIRC_PTS+1];
		float rad = MaxF(GetHotspot(t), GetFallsize(t));
		if (IsDir()) 
			BoxCircle(t,rad,0.0f,box,q,1,tm);
		BoxCircle(t,rad,d,box,q,1,tm);
		}
	BOOL dispAtten, dispAttenNear;
	if (useAtten && (extDispFlags & EXT_DISP_ONLY_SELECTED))  
		dispAtten = dispAttenNear = TRUE;
	else {
		dispAtten = attenDisplay;
		dispAttenNear = attenNearDisplay;
		}
	if( dispAtten||dispAttenNear) {
		Point3 q[3*NUM_CIRC_PTS];
		if (IsDir()) {
			if (dispAtten) {
				BoxCircle(t, GetFallsize(t), GetAtten(t,ATTEN_END), box, q,0,tm);
				BoxCircle(t, GetFallsize(t), GetAtten(t,ATTEN_START), box, q,0,tm);
				}
			if (dispAttenNear) {
				BoxCircle(t, GetFallsize(t), GetAtten(t,ATTEN1_END), box, q,0,tm);
				BoxCircle(t, GetFallsize(t), GetAtten(t,ATTEN1_START), box, q,0,tm);
				}
			}
		else {
			if (type==OMNI_LIGHT||(IsSpot()&&overshoot)) { 
				float rad = MaxF(GetAtten(t, ATTEN_START), GetAtten(t, ATTEN_END));
				GetAttenPoints(t, rad, q);
				box.IncludePoints(q,3*NUM_CIRC_PTS,tm);
				}
			else {
				if (GetSpotShape()) { 
					if (dispAtten) {
						npts = GetCirXPoints(t, GetFallsize(t), GetAtten(t,ATTEN_START),q);
						box.IncludePoints(q,npts,tm);
						npts = GetCirXPoints(t, GetFallsize(t), GetAtten(t,ATTEN_END),q);
						box.IncludePoints(q,npts,tm);
						}
					if (dispAttenNear) {
						npts = GetCirXPoints(t, GetFallsize(t), GetAtten(t,ATTEN1_START),q);
						box.IncludePoints(q,npts,tm);
						npts = GetCirXPoints(t, GetFallsize(t), GetAtten(t,ATTEN1_END),q);
						box.IncludePoints(q,npts,tm);
						}
					}
				else {
					if (dispAtten) {
						npts = GetRectXPoints(t, GetFallsize(t), GetAtten(t,ATTEN_START),q);
						box.IncludePoints(q,npts,tm);
						npts = GetRectXPoints(t, GetFallsize(t), GetAtten(t,ATTEN_END),q);
						box.IncludePoints(q,npts,tm);
						}
					if (dispAttenNear) {
						npts = GetRectXPoints(t, GetFallsize(t), GetAtten(t,ATTEN1_START),q);
						box.IncludePoints(q,npts,tm);
						npts = GetRectXPoints(t, GetFallsize(t), GetAtten(t,ATTEN1_END),q);
						box.IncludePoints(q,npts,tm);
						}
					}
				}
			}
		}
	}

void GeneralLight::GetLocalBoundBox(TimeValue t, INode* inode, ViewExp* vpt, Box3& box )
{
	Point3 loc = inode->GetObjectTM(t).GetTrans();
	float scaleFactor = vpt->NonScalingObjectSize()*vpt->GetVPWorldWidth(loc) / 360.0f;
	box = mesh->getBoundingBox();
	box.Scale(scaleFactor);
	BoxLight(t, inode, box, NULL);
	}

void GeneralLight::GetWorldBoundBox(TimeValue t, INode* inode, ViewExp* vpt, Box3& box )
	{
	int nv;
	Matrix3 tm;
	GetMat(t, inode,vpt,tm);
	Point3 loc = tm.GetTrans();
	nv = mesh->getNumVerts();
	box.Init();
	if(!(extDispFlags & EXT_DISP_ZOOM_EXT)) 
		box.IncludePoints(mesh->verts,nv,&tm);
	else
		box += loc;
	tm = inode->GetObjectTM(t);
	BoxLight(t, inode, box, &tm);
	}


// From BaseObject
int GeneralLight::HitTest(TimeValue t, INode* inode, int type, int crossing, int flags, IPoint2 *p, ViewExp *vpt) 
{
	HitRegion hitRegion;
	DWORD savedLimits;
	int res;
	Matrix3 m;
	if (!enable) 
		return 0;
	GraphicsWindow *gw = vpt->getGW();	
	Material *mtl = gw->getMaterial();
	MakeHitRegion(hitRegion,type,crossing,4,p);	
	gw->setRndLimits(((savedLimits = gw->getRndLimits()) | GW_PICK) & ~GW_ILLUM);
	GetMat(t,inode,vpt,m);
	gw->setTransform(m);
	// if we get a hit on the mesh, we're done
	res = mesh->select( gw, mtl, &hitRegion, flags & HIT_ABORTONHIT);
	// if not, check the target line, and set the pair flag if it's hit
	if( !res )	{
		// this special case only works with point selection of targeted lights
		if((type != HITTYPE_POINT) || !inode->GetTarget())
			return 0;
		// don't let line be active if only looking at selected stuff and target isn't selected
		if((flags & HIT_SELONLY) && !inode->GetTarget()->Selected() )
			return 0;
		gw->clearHitCode();
		if(res = DrawConeAndLine(t, inode, gw, -1))
			inode->SetTargetNodePair(1);
	}
	gw->setRndLimits(savedLimits);
	return res;
}

void GeneralLight::Snap(TimeValue t, INode* inode, SnapInfo *snap, IPoint2 *p, ViewExp *vpt) 
{
	GenericSnap(t,inode,snap,p,vpt);
}

void GeneralLight::SetExtendedDisplay(int flags)
{
	extDispFlags = flags;
}

int GeneralLight::Display(TimeValue t, INode* inode, ViewExp *vpt, int flags) 
{
	Matrix3 m;
	if (!enable) 
		return 0;
	GraphicsWindow *gw = vpt->getGW();
	GetMat(t,inode,vpt,m);
	gw->setTransform(m);
	DWORD rlim = gw->getRndLimits();
	gw->setRndLimits(GW_WIREFRAME|GW_BACKCULL|(gw->getRndMode() & GW_Z_BUFFER));
	if (inode->Selected())
		gw->setColor( LINE_COLOR, GetSelColor());
	else if(!inode->IsFrozen())	{
		if(useLight)
			gw->setColor( LINE_COLOR, GetUIColor(COLOR_LIGHT_OBJ));
//		else
//			gw->setColor( LINE_COLOR, 0.0f, 0.0f, 0.0f);
	}
	mesh->render( gw, gw->getMaterial(),
		(flags&USE_DAMAGE_RECT) ? &vpt->GetDammageRect() : NULL, COMP_ALL);	
	DrawConeAndLine(t, inode, gw, 1);
	DrawAtten(t, inode, gw);
	gw->setRndLimits(rlim);
	return 0 ;
}


RefResult GeneralLight::EvalLightState(TimeValue t, Interval& valid, LightState* ls) 
{
	if(useLight)
		ls->color = GetRGBColor(t,valid);
	else
		ls->color = Point3(0,0,0);
	ls->on = useLight;
	ls->intens = GetIntensity(t, valid);
	ls->hotsize = GetHotspot(t, valid);
	ls->fallsize = GetFallsize(t, valid);
#if 0
	if(ls->useAtten = GetUseAtten()) {
		ls->attenStart = GetAtten(t, ATTEN_START, valid);
		ls->attenEnd = GetAtten(t, ATTEN_END, valid);
	}
	else
		ls->attenStart = ls->attenEnd = 0.0f;
#endif
	
	ls->useNearAtten = GetUseAttenNear();
	ls->nearAttenStart = GetAtten(t, ATTEN1_START, valid);
	ls->nearAttenEnd = GetAtten(t, ATTEN1_END, valid);
	ls->useAtten = GetUseAtten();
	ls->attenStart = GetAtten(t, ATTEN_START, valid);
	ls->attenEnd = GetAtten(t, ATTEN_END, valid);
	ls->shape = GetSpotShape();
	ls->aspect = GetAspect(t, valid);
	ls->overshoot = GetOvershoot();
	ls->shadow = GetShadow();
	//ls->softenDiffuse = softenDiffuse;
	//ls->decayType = decayType;
	ls->affectDiffuse = affectDiffuse;
	ls->affectSpecular = affectSpecular;

	if(type == OMNI_LIGHT)
		ls->type = OMNI_LGT;
	else if(type == DIR_LIGHT||type==TDIR_LIGHT)
		ls->type = DIRECT_LGT;
	else
		ls->type = SPOT_LGT;

	return REF_SUCCEED;
}

//
// Reference Managment:
//
					
RefResult GeneralLight::NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, PartID& partID, RefMessage message ) 
{
	switch (message) {
	case REFMSG_CHANGE:
		if(iObjParams)
			UpdateUI(iObjParams->GetTime());
		break;

	case REFMSG_GET_PARAM_DIM: {
		GetParamDim *gpd = (GetParamDim*)partID;
		if (type == OMNI_LIGHT) {
			switch (gpd->index) {
				case PB_COLOR:
					gpd->dim = stdColor255Dim;
					break;
				case PB_INTENSITY:
				case PB_CONTRAST:
				case PB_DIFFSOFT:
				case PB_OMNIMAP_SIZE:
				case PB_OMNIMAP_RANGE:
					gpd->dim = defaultDim;
					break;					
				case PB_OMNIMAP_BIAS: 
				case PB_OMNIRAY_BIAS:
				case PB_OMNIATSTART1:
				case PB_OMNIATEND1:
				case PB_OMNIATSTART:
				case PB_OMNIATEND:
					gpd->dim = stdWorldDim;
					break;					
				}
			}
		else {
			switch (gpd->index) {
				case PB_COLOR:
					gpd->dim = stdColor255Dim;
					break;
				case PB_INTENSITY:
				case PB_CONTRAST:
				case PB_DIFFSOFT:
				case PB_ASPECT:
				case PB_MAP_SIZE:
//				case PB_RADIUS:
//				case PB_NUMBER_LIGHTS:
				case PB_MAP_RANGE:
					gpd->dim = defaultDim;
					break;					
				case PB_HOTSIZE:
				case PB_FALLSIZE:
					gpd->dim = IsDir()?stdWorldDim:defaultDim;
					break;					
				case PB_MAP_BIAS: 
				case PB_RAY_BIAS:
				case PB_TDIST:
				case PB_ATTENSTART1:
				case PB_ATTENEND1:
				case PB_ATTENSTART:
				case PB_ATTENEND:
					gpd->dim = stdWorldDim;
					break;					
				}
			}
		return REF_STOP; 
		}


	case REFMSG_GET_PARAM_NAME: {
		GetParamName *gpn = (GetParamName*)partID;
		if (type == OMNI_LIGHT) {
			switch (gpn->index) {
				case PB_COLOR: gpn->name = TSTR(GetString(IDS_DB_COLOR)); break;
				case PB_INTENSITY:  gpn->name = TSTR(GetString(IDS_DB_MULTIPLIER)); break;
				case PB_CONTRAST:  gpn->name = TSTR(GetString(IDS_DS_CONTRAST)); break;
				case PB_DIFFSOFT:  gpn->name = TSTR(GetString(IDS_DS_DIFFSOFT)); break;
				case PB_OMNIMAP_SIZE: gpn->name = TSTR(GetString(IDS_RB_MAPSIZE)); break;
				case PB_OMNIMAP_BIAS: gpn->name = TSTR(GetString(IDS_RB_MAPBIAS)); break;
				case PB_OMNIMAP_RANGE: gpn->name = TSTR(GetString(IDS_RB_MAPRANGE)); break;
				case PB_OMNIRAY_BIAS: gpn->name = TSTR(GetString(IDS_RB_RAYBIAS)); break;
				case PB_OMNIATSTART1: gpn->name = TSTR(GetString(IDS_DS_ATTENSTARTNEAR)); break;
				case PB_OMNIATEND1: gpn->name = TSTR(GetString(IDS_DS_ATTENENDNEAR)); break;
				case PB_OMNIATSTART: gpn->name = TSTR(GetString(IDS_DB_ATTENSTART)); break;
				case PB_OMNIATEND: gpn->name = TSTR(GetString(IDS_DB_ATTENEND)); break;
				}
			}
		else {			
			switch (gpn->index) {
				case PB_COLOR: gpn->name =  TSTR(GetString(IDS_DB_COLOR)); break;
				case PB_INTENSITY: gpn->name = TSTR(GetString(IDS_DB_MULTIPLIER)); break;
				case PB_CONTRAST:  gpn->name = TSTR(GetString(IDS_DS_CONTRAST)); break;
				case PB_DIFFSOFT:  gpn->name = TSTR(GetString(IDS_DS_DIFFSOFT)); break;
				case PB_HOTSIZE:  gpn->name = TSTR(GetString(IDS_DB_HOTSIZE));	break;
				case PB_FALLSIZE: gpn->name = TSTR(GetString(IDS_DB_FALLSIZE));	break;
				case PB_ASPECT:	gpn->name = TSTR(GetString(IDS_DB_ASPECT_RATIO));	break;
				case PB_MAP_SIZE: gpn->name = TSTR(GetString(IDS_RB_MAPSIZE));	break;
				case PB_MAP_BIAS: gpn->name = TSTR(GetString(IDS_RB_MAPBIAS));	break;
				case PB_MAP_RANGE:	gpn->name = TSTR(GetString(IDS_RB_MAPRANGE));	break;
				case PB_RAY_BIAS: gpn->name = TSTR(GetString(IDS_RB_RAYBIAS));	break;
				case PB_ATTENSTART1: gpn->name = TSTR(GetString(IDS_DS_ATTENSTARTNEAR)); break;
				case PB_ATTENEND1:	gpn->name = TSTR(GetString(IDS_DS_ATTENENDNEAR));	break;
				case PB_ATTENSTART:	gpn->name = TSTR(GetString(IDS_DB_ATTENSTART));	break;
				case PB_ATTENEND: gpn->name = TSTR(GetString(IDS_DB_ATTENEND));	break;
				case PB_TDIST:	gpn->name = TSTR(GetString(IDS_DB_TDIST));		break;
//				case PB_RADIUS: gpn->name = TSTR(GetString(IDS_RB_RADIUS));	break;
//				case PB_NUMBER_LIGHTS: gpn->name = TSTR(GetString(IDS_RB_NUMBER_LIGHTS));	break;

				}
			}
		return REF_STOP; 
		}
	}
	return(REF_SUCCEED);
}

ObjectState GeneralLight::Eval(TimeValue time)
{
	UpdateUI(time);
	return ObjectState(this);
}

Interval GeneralLight::ObjectValidity(TimeValue time) 
	{
	Interval valid;
	valid.SetInfinite();
	if (!waitPostLoad) {
		GetRGBColor(time, valid);
		GetIntensity(time, valid);
		GetContrast(time, valid);
		GetDiffuseSoft(time, valid);
		if(IsSpot()) {
			GetHotspot(time, valid);
			GetFallsize(time, valid);
			GetTDist(time, valid);
			GetAspect(time, valid);
			}
		GetAtten(time, ATTEN_START, valid);
		GetAtten(time, ATTEN_END, valid);
		GetAtten(time, ATTEN1_START, valid);
		GetAtten(time, ATTEN1_END, valid);
		GetMapSize(time,valid);
		GetMapRange(time,valid);
		GetRayBias(time,valid);
		UpdateUI(time);
		}
	return valid;
	}

RefTargetHandle GeneralLight::Clone(RemapDir& remap) 
	{
	GeneralLight* newob = new GeneralLight(type);
	newob->enable = enable;
	newob->coneDisplay = coneDisplay;
	newob->useLight = useLight;
	newob->attenDisplay = attenDisplay;
	newob->useAtten = useAtten;
	newob->useAttenNear = useAttenNear;
	newob->attenNearDisplay = attenNearDisplay;
	newob->shape = shape;
	newob->shadow = shadow;
	newob->shadowType = shadowType;
	newob->overshoot = overshoot;
	newob->projector = projector;
	newob->absMapBias = absMapBias;
	newob->exclList = exclList;
	newob->softenDiffuse = softenDiffuse;
	newob->affectDiffuse = affectDiffuse;
	newob->affectSpecular = affectSpecular;
	newob->decayType = decayType;
	newob->ReplaceReference(0,pblock->Clone(remap));
	if (projMap) 
		newob->ReplaceReference(1,projMap->Clone(remap));
	return(newob);
	}


#define COLOR_CHUNK 0x2550		// obsolete
#define HOTSIZE_CHUNK 0x2552	// obsolete
#define CONE_DISPLAY_CHUNK	0x2560
#define RANGE_DISPLAY_CHUNK	0x2561
#define USE_ATTEN_CHUNK 0x2562
#define CROSSOVER_CHUNK 0x2563
#define SHADOW_CHUNK 0x2570
#define SHADOW_TYPE_CHUNK 0x2571
#define GLOBALSHADOW_CHUNK 0x2572
#define SPOTSHAPE_CHUNK 0x2573
#define ABS_MAP_BIAS_CHUNK 0x2574
#define OVERSHOOT_CHUNK 0x2575
#define EXCL_LIST_CHUNK 0x2576
#define PROJECTOR_CHUNK 0x2577
#define MAP_RANGE_CHUNK 0x2578
#define MAP_BIAS_CHUNK 0x2579
#define MAP_SIZE_CHUNK 0x257A
#define RAY_BIAS_CHUNK 0x257B
#define ON_OFF_CHUNK 0x2580
#define NEAR_RANGE_DISPLAY_CHUNK 0x2590
#define NEAR_USE_ATTEN_CHUNK     0x25A0
#define DONT_AFFECT_DIFF_CHUNK     0x25B0
#define DONT_AFFECT_SPEC_CHUNK     0x25B1
#define SOFTEN_DIFF_CHUNK     0x25C0
#define DECAY_TYPE_CHUNK     0x25D0

// IO
IOResult GeneralLight::Save(ISave *isave) 
{
	ULONG nb;
	
	isave->BeginChunk(ON_OFF_CHUNK);
	isave->Write(&useLight, sizeof(useLight), &nb);
	isave->EndChunk();
	isave->BeginChunk(CONE_DISPLAY_CHUNK);
	isave->Write(&coneDisplay, sizeof(coneDisplay), &nb);
	isave->EndChunk();
	isave->BeginChunk(RANGE_DISPLAY_CHUNK);
	isave->Write(&attenDisplay, sizeof(attenDisplay), &nb);
	isave->EndChunk();
	isave->BeginChunk(USE_ATTEN_CHUNK);
	isave->Write(&useAtten, sizeof(useAtten), &nb);
	isave->EndChunk();
	isave->BeginChunk(NEAR_RANGE_DISPLAY_CHUNK);
	isave->Write(&attenNearDisplay, sizeof(attenNearDisplay), &nb);
	isave->EndChunk();
	isave->BeginChunk(NEAR_USE_ATTEN_CHUNK);
	isave->Write(&useAttenNear, sizeof(useAttenNear), &nb);
	isave->EndChunk();

	isave->BeginChunk(SHADOW_CHUNK);
	isave->Write(&shadow, sizeof(shadow), &nb);
	isave->EndChunk();
	isave->BeginChunk(SHADOW_TYPE_CHUNK);
	isave->Write(&shadowType, sizeof(shadowType), &nb);
	isave->EndChunk();
	isave->BeginChunk(GLOBALSHADOW_CHUNK);
	isave->Write(&useGlobalShadowParams, sizeof(useGlobalShadowParams), &nb);
	isave->EndChunk();
	isave->BeginChunk(ABS_MAP_BIAS_CHUNK);
	isave->Write(&absMapBias, sizeof(absMapBias), &nb);
	isave->EndChunk();

	if (IsSpot()) {
		isave->BeginChunk(SPOTSHAPE_CHUNK);
		isave->Write(&shape, sizeof(shape), &nb);
		isave->EndChunk();
		if (overshoot) {
			isave->BeginChunk(OVERSHOOT_CHUNK);
			isave->EndChunk();
			}
		}

	if (projector) {
		isave->BeginChunk(PROJECTOR_CHUNK);
		isave->EndChunk();
		}
	if (!affectDiffuse) {
		isave->BeginChunk(DONT_AFFECT_DIFF_CHUNK);
		isave->EndChunk();
		}
	if (!affectSpecular) {
		isave->BeginChunk(DONT_AFFECT_SPEC_CHUNK);
		isave->EndChunk();
		}
	if (softenDiffuse) {
		isave->BeginChunk(SOFTEN_DIFF_CHUNK);
		isave->EndChunk();
		}
	if (decayType) {
		isave->BeginChunk(DECAY_TYPE_CHUNK);
		isave->Write(&decayType, sizeof(decayType), &nb);
		isave->EndChunk();
		}
	if (exclList.Count()>0) {
		isave->BeginChunk(EXCL_LIST_CHUNK);
		exclList.Save(isave);
		isave->EndChunk();
		}


	return IO_OK;
	}

class LightPostLoad : public PostLoadCallback {
public:
	GeneralLight *gl;
	Interval valid;
	LightPostLoad(GeneralLight *l) { gl = l;}
	void proc(ILoad *iload) {
		Point3 rgb;
		if (gl->pblock->GetVersion() != LIGHT_VERSION) {
			switch (gl->pblock->GetVersion()) {
			case 0:
				gl->pblock->GetValue(0, TimeValue(0), rgb[0], valid);
				gl->pblock->GetValue(1, TimeValue(0), rgb[1], valid);
				gl->pblock->GetValue(2, TimeValue(0), rgb[2], valid);
				gl->ReplaceReference(0,
					UpdateParameterBlock(
						descV0, oldDim[gl->type], gl->pblock,
						gl->type == OMNI_LIGHT ? descV8Omni : descV8, newDim8[gl->type], LIGHT_VERSION));
				gl->pblock->SetValue(0, TimeValue(0), rgb);
				gl->pblock->SetValue(1, TimeValue(0), 1.0f);
				iload->SetObsolete();
				break;

			case 1:
				gl->ReplaceReference(0,
					UpdateParameterBlock(
						descV1, oldDim1[gl->type], gl->pblock,
						gl->type == OMNI_LIGHT ? descV8Omni : descV8, newDim8[gl->type], LIGHT_VERSION));
				//gl->pblock->SetValue(PB_MAP_SIZE, TimeValue(0), gl->dlgMapSize);
				//gl->pblock->SetValue(PB_MAP_BIAS, TimeValue(0), gl->dlgMapBias);
				//gl->pblock->SetValue(PB_MAP_RANGE, TimeValue(0), gl->dlgMapRange);
				//gl->pblock->SetValue(PB_RAY_BIAS, TimeValue(0), gl->dlgRayBias);
				iload->SetObsolete();
				break;
			case 2:
				gl->ReplaceReference(0,
					UpdateParameterBlock(
						descV2, oldDim2[gl->type], gl->pblock,
						gl->type == OMNI_LIGHT ? descV8Omni : descV8, newDim8[gl->type], LIGHT_VERSION));
				iload->SetObsolete();
				break;
			case 3:
				gl->ReplaceReference(0,
					UpdateParameterBlock(
						gl->type == OMNI_LIGHT ? descV3Omni : descV3, oldDim3[gl->type], gl->pblock,
						gl->type == OMNI_LIGHT ? descV8Omni : descV8, newDim8[gl->type], LIGHT_VERSION));
				iload->SetObsolete();
				break;
			case 4:
				gl->ReplaceReference(0,
					UpdateParameterBlock(
						gl->type == OMNI_LIGHT ? descV4Omni : descV4, oldDim4[gl->type], gl->pblock,
						gl->type == OMNI_LIGHT ? descV8Omni : descV8, newDim8[gl->type], LIGHT_VERSION));
				if (gl->type == OMNI_LIGHT) {
					gl->pblock->SetValue(PB_OMNIMAP_SIZE, TimeValue(0), gl->dlgMapSize);
					gl->pblock->SetValue(PB_OMNIMAP_BIAS, TimeValue(0), gl->dlgMapBias);
					gl->pblock->SetValue(PB_OMNIMAP_RANGE, TimeValue(0), gl->dlgMapRange);
					gl->pblock->SetValue(PB_OMNIRAY_BIAS, TimeValue(0), gl->dlgRayBias);
					}
				iload->SetObsolete();
				break;
			case 5:
				gl->ReplaceReference(0,
					UpdateParameterBlock(
						gl->type == OMNI_LIGHT ? descV5Omni : descV5, oldDim5[gl->type], gl->pblock,
						gl->type == OMNI_LIGHT ? descV8Omni : descV8, newDim8[gl->type], LIGHT_VERSION));
				iload->SetObsolete();
				break;
			case 6:
				gl->ReplaceReference(0,
					UpdateParameterBlock(
						gl->type == OMNI_LIGHT ? descV6Omni : descV6, oldDim6[gl->type], gl->pblock,
						gl->type == OMNI_LIGHT ? descV8Omni : descV8, newDim8[gl->type], LIGHT_VERSION));
				iload->SetObsolete();
				break;
			case 7:
				gl->ReplaceReference(0,
					UpdateParameterBlock(
						gl->type == OMNI_LIGHT ? descV7Omni : descV7, oldDim7[gl->type], gl->pblock,
						gl->type == OMNI_LIGHT ? descV8Omni : descV8, newDim8[gl->type], LIGHT_VERSION));
				iload->SetObsolete();
				break;

			default:
				assert(0);
				break;
			}
			if (gl->pblock->GetVersion()<8) {
				if (gl->softenDiffuse) gl->SetDiffuseSoft(0,1.0f);
				}
		}
		gl->BuildSpotMesh(max(gl->GetHotspot(0), gl->GetFallsize(0)));
		waitPostLoad--;
		delete this;
	}
};

IOResult  GeneralLight::Load(ILoad *iload) 
{
	ULONG nb;
	IOResult res;
	enable = TRUE;
	while (IO_OK==(res=iload->OpenChunk())) {
		switch(iload->CurChunkID())  {
		case ON_OFF_CHUNK:
			res = iload->Read(&useLight,sizeof(useLight), &nb);
			break;
		case CONE_DISPLAY_CHUNK:
			res = iload->Read(&coneDisplay,sizeof(coneDisplay), &nb);
			break;
		case RANGE_DISPLAY_CHUNK:
			res = iload->Read(&attenDisplay,sizeof(attenDisplay), &nb);
			break;
		case NEAR_USE_ATTEN_CHUNK:
			res = iload->Read(&useAttenNear,sizeof(useAttenNear), &nb);
			break;
		case NEAR_RANGE_DISPLAY_CHUNK:
			res = iload->Read(&attenNearDisplay,sizeof(attenNearDisplay), &nb);
			break;
		case USE_ATTEN_CHUNK:
			res = iload->Read(&useAtten,sizeof(useAtten), &nb);
			break;
		case SHADOW_CHUNK:
			res = iload->Read(&shadow,sizeof(shadow), &nb);
			break;
		case SHADOW_TYPE_CHUNK:
			res = iload->Read(&shadowType,sizeof(shadowType), &nb);
			break;
		case GLOBALSHADOW_CHUNK:
			res = iload->Read(&useGlobalShadowParams,sizeof(useGlobalShadowParams), &nb);
			break;
		case SPOTSHAPE_CHUNK:
			res = iload->Read(&shape,sizeof(shape), &nb);
			break;
		case ABS_MAP_BIAS_CHUNK:
			res = iload->Read(&absMapBias,sizeof(absMapBias), &nb);
			break;
		case OVERSHOOT_CHUNK:
			overshoot = TRUE;
			break;
		case PROJECTOR_CHUNK:
			projector = TRUE;
			break;
		case DONT_AFFECT_DIFF_CHUNK:
			affectDiffuse = FALSE;
			break;
		case DONT_AFFECT_SPEC_CHUNK:
			affectSpecular = FALSE;
			break;
		case SOFTEN_DIFF_CHUNK:
			softenDiffuse = TRUE;
			break;
		case EXCL_LIST_CHUNK:
			res = exclList.Load(iload);	
			break;
		case DECAY_TYPE_CHUNK:
			res = iload->Read(&decayType,sizeof(decayType), &nb);
			break;
		}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
	}
	waitPostLoad++;
	iload->RegisterPostLoadCallback(new LightPostLoad(this));
	return IO_OK;
}


// target spot light creation stuff...

class STSpotCreationManager : public MouseCallBack, ReferenceMaker 
{
private:
	CreateMouseCallBack *createCB;	
	INode *lgtNode,*targNode;
	GeneralLight *lgtObject;
	TargetObject *targObject;
	int attachedToNode;
	IObjCreate *createInterface;
	ClassDesc *cDesc;
	Matrix3 mat;  // the nodes TM relative to the CP
	IPoint2 pt0;
	int ignoreSelectionChange;
	int lastPutCount;

	void CreateNewObject();	

	int NumRefs() { return 1; }
	RefTargetHandle GetReference(int i) { return (RefTargetHandle)lgtNode; } 
	void SetReference(int i, RefTargetHandle rtarg) { lgtNode = (INode *)rtarg; }

	// StdNotifyRefChanged calls this, which can change the partID to new value 
	// If it doesnt depend on the particular message& partID, it should return
	// REF_DONTCARE
    RefResult NotifyRefChanged(Interval changeInt, RefTargetHandle hTarget, 
	    	PartID& partID,  RefMessage message);

public:
	void Begin( IObjCreate *ioc, ClassDesc *desc );
	void End();
		
	STSpotCreationManager()	{ ignoreSelectionChange = FALSE; }
	int proc( HWND hwnd, int msg, int point, int flag, IPoint2 m );
};


#define CID_TSPOTCREATE	CID_USER + 3

class STSpotCreateMode : public CommandMode {
	STSpotCreationManager proc;
public:
	void Begin( IObjCreate *ioc, ClassDesc *desc ) { proc.Begin( ioc, desc ); }
	void End() { proc.End(); }

	int Class() { return CREATE_COMMAND; }
	int ID() { return CID_TSPOTCREATE; }
	MouseCallBack *MouseProc(int *numPoints) { *numPoints = 1000000; return &proc; }
	ChangeForegroundCallback *ChangeFGProc() { return CHANGE_FG_SELECTED; }
	BOOL ChangeFG( CommandMode *oldMode ) { return (oldMode->ChangeFGProc() != CHANGE_FG_SELECTED); }
	void EnterMode() {}
	void ExitMode() {}
	BOOL IsSticky() { return FALSE; }
};

static STSpotCreateMode theSTSpotCreateMode;

void STSpotCreationManager::Begin( IObjCreate *ioc, ClassDesc *desc )
{
	createInterface = ioc;
	cDesc           = desc;
	attachedToNode  = FALSE;
	createCB        = NULL;
	lgtNode         = NULL;
	targNode        = NULL;
	lgtObject       = NULL;
	targObject      = NULL;
	CreateNewObject();
}

void STSpotCreationManager::End()
{
	if ( lgtObject ) {
		lgtObject->EndEditParams( (IObjParam*)createInterface, END_EDIT_REMOVEUI, NULL);
		if ( !attachedToNode ) {
			// RB 4-9-96: Normally the hold isn't holding when this 
			// happens, but it can be in certain situations (like a track view paste)
			// Things get confused if it ends up with undo...
			theHold.Suspend(); 
			delete lgtObject;
			lgtObject = NULL;
			theHold.Resume();
			// RB 7/28/97: If something has been put on the undo stack since this object was created, we have to flush the undo stack.
			if (theHold.GetGlobalPutCount()!=lastPutCount) {
				GetSystemSetting(SYSSET_CLEAR_UNDO);
				}
		} 
		else if ( lgtNode ) {
			 // Get rid of the reference.
			DeleteReference(0);  // sets lgtNode = NULL
		}
	}	
}

RefResult STSpotCreationManager::NotifyRefChanged(
	Interval changeInt, 
	RefTargetHandle hTarget, 
	PartID& partID,  
	RefMessage message) 
{
	switch (message) {
		
	case REFMSG_PRENOTIFY_PASTE:
	case REFMSG_TARGET_SELECTIONCHANGE:
	 	if ( ignoreSelectionChange ) {
			break;
		}
	 	if (lgtObject && lgtNode==hTarget) {
			// this will set camNode== NULL;
			DeleteReference(0);
			goto endEdit;
		}
		// fall through

	case REFMSG_TARGET_DELETED:		
		if ( lgtObject && lgtNode==hTarget ) {
			endEdit:
			lgtObject->EndEditParams( (IObjParam*)createInterface, 0, NULL);
			lgtObject  = NULL;				
			lgtNode    = NULL;
			CreateNewObject();	
			attachedToNode = FALSE;
		}
		else if (targNode==hTarget) {
			targNode = NULL;
			targObject = NULL;
		}
		break;		
	}
	return REF_SUCCEED;
}


void STSpotCreationManager::CreateNewObject()
{
	lgtObject = (GeneralLight *)cDesc->Create();
	lastPutCount = theHold.GetGlobalPutCount();
	
	// Start the edit params process
	if ( lgtObject ) {
		lgtObject->BeginEditParams( (IObjParam*)createInterface, BEGIN_EDIT_CREATE, NULL );
	}	
}

static BOOL needToss;
			
int STSpotCreationManager::proc( 
				HWND hwnd,
				int msg,
				int point,
				int flag,
				IPoint2 m )
{	
	int res;
	TSTR targName;	
	ViewExp *vpx = createInterface->GetViewport(hwnd); 
	assert( vpx );

	switch ( msg ) {
	case MOUSE_POINT:
		switch ( point ) {
		case 0:
			pt0 = m;
			assert( lgtObject );					
			if ( createInterface->SetActiveViewport(hwnd) ) {
				return FALSE;
			}

			if (createInterface->IsCPEdgeOnInView()) { 
				res = FALSE;
				goto done;
			}

			// if lights were hidden by category, re-display them
			GetCOREInterface()->SetHideByCategoryFlags(
					GetCOREInterface()->GetHideByCategoryFlags() & ~HIDE_LIGHTS);

			if ( attachedToNode ) {
		   		// send this one on its way
		   		lgtObject->EndEditParams( (IObjParam*)createInterface, 0, NULL);
						
				// Get rid of the reference.
				if (lgtNode)
					DeleteReference(0);

				// new object
				CreateNewObject();   // creates lgtObject
			}

			needToss = theHold.GetGlobalPutCount()!=lastPutCount;

		   	theHold.Begin();	 // begin hold for undo
			mat.IdentityMatrix();

			// link it up
			lgtNode = createInterface->CreateObjectNode( lgtObject);
			attachedToNode = TRUE;
			assert( lgtNode );					
			createCB = lgtObject->GetCreateMouseCallBack();					
			createInterface->SelectNode( lgtNode );
					
			// Create target object and node
			targObject = new TargetObject;
			assert(targObject);
			targNode = createInterface->CreateObjectNode( targObject);
			assert(targNode);
			targName = lgtNode->GetName();
			targName += GetString(IDS_DB_DOT_TARGET);
			targNode->SetName(targName);
				
			// hook up camera to target using lookat controller.
			createInterface->BindToTarget(lgtNode,targNode);

			// Reference the new node so we'll get notifications.
			MakeRefByID( FOREVER, 0, lgtNode);
			
			// Position camera and target at first point then drag.
			mat.IdentityMatrix();
			//mat[3] = vpx->GetPointOnCP(m);
			#ifdef _3D_CREATE
				mat.SetTrans( vpx->SnapPoint(m,m,NULL,SNAP_IN_3D) );
			#else
				mat.SetTrans(vpx->SnapPoint(m,m,NULL,SNAP_IN_PLANE));
			#endif
			createInterface->SetNodeTMRelConstPlane(lgtNode, mat);
			createInterface->SetNodeTMRelConstPlane(targNode, mat);
			lgtObject->Enable(1);

		   	ignoreSelectionChange = TRUE;
		   	createInterface->SelectNode( targNode,0);
		   	ignoreSelectionChange = FALSE;
			res = TRUE;
			break;
					
		case 1:
			if (Length(m-pt0)<2)
				goto abort;
			//mat[3] = vpx->GetPointOnCP(m);
			#ifdef _3D_CREATE
				mat.SetTrans( vpx->SnapPoint(m,m,NULL,SNAP_IN_3D) );
			#else
				mat.SetTrans(vpx->SnapPoint(m,m,NULL,SNAP_IN_PLANE));
			#endif
			createInterface->SetNodeTMRelConstPlane(targNode, mat);
		   	ignoreSelectionChange = TRUE;
		   	createInterface->SelectNode( lgtNode);
		   	ignoreSelectionChange = FALSE;
					
		    theHold.Accept(IDS_DS_CREATE);	 

			createInterface->AddLightToScene(lgtNode); 
			createInterface->RedrawViews(createInterface->GetTime());  

			res = FALSE;	// We're done
			break;
		}			
		break;

	case MOUSE_MOVE:
		//mat[3] = vpx->GetPointOnCP(m);
		#ifdef _3D_CREATE
			mat.SetTrans( vpx->SnapPoint(m,m,NULL,SNAP_IN_3D) );
		#else
			mat.SetTrans(vpx->SnapPoint(m,m,NULL,SNAP_IN_PLANE));
		#endif
		createInterface->SetNodeTMRelConstPlane(targNode, mat);
		createInterface->RedrawViews(createInterface->GetTime());	   
		res = TRUE;
		break;

	case MOUSE_FREEMOVE:
		SetCursor(LoadCursor(hInstance, MAKEINTRESOURCE(IDC_CROSS_HAIR)));
		#ifdef _OSNAP
			//Snap Preview
			#ifdef _3D_CREATE
				vpx->SnapPreview(m,m,NULL, SNAP_IN_3D);
			#else
				vpx->SnapPreview(m,m,NULL, SNAP_IN_PLANE);
			#endif
		#endif

		break;

	case MOUSE_ABORT:
		abort:
		assert( lgtObject );
		lgtObject->EndEditParams( (IObjParam*)createInterface,0,NULL);
		// Toss the undo stack if param changes have been made
		theHold.Cancel();	 // deletes both the camera and target.
		if (needToss) 
			GetSystemSetting(SYSSET_CLEAR_UNDO);
		lgtNode = NULL;			
		targNode = NULL;	 	
		createInterface->RedrawViews(createInterface->GetTime()); 
		CreateNewObject();	
		attachedToNode = FALSE;
		res = FALSE;						
	}
	
done:
	createInterface->ReleaseViewport(vpx); 
	return res;
}

//----------------------------------------------------------------------

int STSpotLightClassDesc::BeginCreate(Interface *i)
{
	IObjCreate *iob = i->GetIObjCreate();
	
	//iob->SetMouseProc( new LACamCreationManager(iob,this), 1000000 );

	theSTSpotCreateMode.Begin( iob, this );
	iob->PushCommandMode( &theSTSpotCreateMode );
	
	return TRUE;
}


int STSpotLightClassDesc::EndCreate(Interface *i)
{
	theSTSpotCreateMode.End();
	i->RemoveMode( &theSTSpotCreateMode );

	return TRUE;
}


//----------------------------------------------------------------------
//  Light descriptors for rendering
//----------------------------------------------------------------------

#define COS_45 0.7071067f
#define COS_45_2X 1.4142136f

static float stepFactor[] = {50.0f,80.0f,140.0f};
#define MAXSTEPS 1000

class AttenRanges {
	public:
		float aStart, aEnd;	// Attenuation start and end and hot spot scaling for volume shading		
		float aNearStart, aNearEnd;	// Near Attenuation start and end and hot spot scaling for volume shading		
	};

class BaseObjLightSoft : public ObjLightDesc {
	public:		
		NameTab exclList;
		Color intensCol;   	// intens*color	
		// RB: put these into a different structure and pass them to
		// AttenIllum to avoid threading problems.
		//float aStart, aEnd;	// Attenuation start and end and hot spot scaling for volume shading		
		//float aNearStart, aNearEnd;	// Near Attenuation start and end and hot spot scaling for volume shading		

		GeneralLight* gl;

		float contrast,kA,kB,diffSoft;
		int decayType;
		BOOL diffSoften;
		BOOL lightIsZero;
		BaseObjLightSoft(INode *n);
		virtual ~BaseObjLightSoft() {}
		void DeleteThis() {delete this;}
		NameTab* GetExclList() { return &exclList; }  
//int Update(TimeValue t, const RendContext& rc, BOOL shadows);
		int Update(TimeValue t, const RendContext &rc, RenderGlobalContext *rgc, BOOL shadows, BOOL shadowGeomChanged);

		void TraverseVolume(ShadeContext& sc,const Ray &ray,int samples,float tStop,float attenStart,float attenEnd,DWORD flags,LightRayTraversal *proc);
		virtual BOOL IntersectRay(Ray &ray, float &t0, float &t1, AttenRanges &ranges) {return FALSE;}
		virtual int  IntersectRayMultiple(Ray &ray, float t0, float t1, float *t);
		virtual Color AttenuateIllum(ShadeContext& sc,Point3 p,Color &colStep,Point3 &dp,int filt, float ldp, float &distAtten, AttenRanges &ranges) {return Color(0,0,0);}		
		virtual BOOL UseAtten()=0;
		virtual BOOL IsFacingLight(Point3 &dir) {return FALSE;}
		virtual int LightType()=0;

		inline float ContrastFunc(float nl) {
			if (diffSoft!=0.0f) {
				float p = nl*nl*(2.0f-nl);	// based on Hermite interpolant 
				nl = diffSoft*p + (1.0f-diffSoft)*nl;
				}
			return (contrast==0.0f)? nl: 
				nl/(kA*nl+kB);  //  the "Bias" function described in Graphics Gems IV, pp. 401ff
			}
		inline float Decay(float att,float dist, float r0) {
			float s;
			if (decayType==DECAY_NONE || dist==0.0f) return att;
			if ((s=r0/dist)>=1.0f) return att;
			return  (decayType==DECAY_INVSQ)?  s*s*att: s*att;
			}
	};


BaseObjLightSoft::BaseObjLightSoft(INode *n) : ObjLightDesc(n) {
	ObjectState os = n->EvalWorldState(TimeValue(0));
	assert(os.obj->SuperClassID()==LIGHT_CLASS_ID);
	GeneralLight* gl = (GeneralLight*)os.obj;
	exclList = *(gl->GetExclList());
	}	

//int BaseObjLightSoft::Update(TimeValue t, const RendContext &rc, BOOL shadows) {
int BaseObjLightSoft::Update(TimeValue t, const RendContext &rc, RenderGlobalContext *rgc, BOOL shadows, BOOL shadowGeomChanged) {
	ObjLightDesc::Update(t,rc,rgc,shadows,shadowGeomChanged);
	intensCol = ls.intens*ls.color*rc.GlobalLightLevel();
	ObjectState os = inode->EvalWorldState(t);
    GeneralLight* lob = (GeneralLight *)os.obj;		
	contrast = lob->GetContrast(t);
	diffSoft = lob->GetDiffuseSoft(t)/(float) 100.f;
	decayType = lob->GetDecayType();
	float a = contrast/204.0f + 0.5f; // so "a" varies from .5 to .99
	kA = (2.0f-1.0f/a);
	kB = 1.0f-kA;
	diffSoften = lob->GetSoftenDiffuse();
	lightIsZero = intensCol==Color(0,0,0)?1:0;   	// intens*color	
	return 1;
	}

int BaseObjLightSoft::IntersectRayMultiple(Ray &ray, float t0, float t1, float *t) { 
	t[0] = t0;
	t[1] = t1;
	return 2;
	}


void BaseObjLightSoft::TraverseVolume(
		ShadeContext& sc,
		const Ray &ray,int samples,float tStop,
		float attenStart, float attenEnd,
		DWORD flags,
		LightRayTraversal *proc)
	{
	Point3 dp, p;
	float ot, t, dt, ldp, t0, t1, dtLight, distAtten;
	Color illum, colStep;	
	Ray lr;
	int filt = 0;
	AttenRanges ranges;

	if (flags&TRAVERSE_LOWFILTSHADOWS) filt=1;
	if (flags&TRAVERSE_HIFILTSHADOWS) filt=2;
	if (flags&TRAVERSE_USESAMPLESIZE) filt=3;

	lr.p   = worldToLight * ray.p;
	lr.dir = VectorTransform(worldToLight,ray.dir);
	
	if (LightType()==DIR_LIGHT&&ls.useNearAtten&&ls.useAtten) {
		float mid = (ls.nearAttenEnd+ls.attenStart)/2.0f;
		ranges.aStart    = (ls.attenStart-mid)*attenStart + mid;
		ranges.aEnd      = (ls.attenEnd-mid)*attenEnd + mid;
		ranges.aNearStart= (ls.nearAttenStart-mid)*attenStart + mid;
		ranges.aNearEnd  = (ls.nearAttenEnd-mid)*attenEnd + mid;
		}
	else {
		ranges.aStart    = ls.attenStart * attenStart;
		ranges.aEnd      = ls.attenEnd * attenEnd;
		ranges.aNearStart= ls.nearAttenStart * attenStart;
		ranges.aNearEnd  = ls.nearAttenEnd * attenEnd;
		}
		
	// See where we hit the light volume
	if (!IntersectRay(lr,t0,t1,ranges)) {
		// We missed.
		proc->Step(0.0f,tStop,Color(0,0,0),1.0f);
		return;
		}

	if (t1<0.0f || t0>tStop) {
		// We missed.
		proc->Step(0.0f,tStop,Color(0,0,0),1.0f);
		return;
		}
	if (t0<0.0f) t0 = 0.0f;
	if (t1>tStop) t1 = tStop;
	if (LightType()!=OMNI_LIGHT) {
		if (UseAtten() && (t1-t0) > (2.0f*ranges.aEnd)) {
			// We can only do this if we're not looking into a spot light.
			if (!IsFacingLight(lr.dir)) {
				t1 = 2.0f*ranges.aEnd + t0;
				}
			}
		}

	// The first (large) step
	if (t0!=0.0f) {
		if (!proc->Step(0.0f,t0,Color(0,0,0),1.0f)) {
			return;
			}
		}
	float tm[10];
	int n;	
	n = IntersectRayMultiple(lr,t0,t1,tm);

	dtLight = (t1-t0)/float(samples);
	
	for (int i=0; i<n-1; i++) {
		// Set everything up
		//dtLight = LightStepSize(sc,lr,t0,t1,minStep,filt);
		dt      = dtLight;
		if (dt<0.0001f) dt = 0.0001f;
		ot      = tm[i];
		t       = tm[i] + dt;	
		dp      = lr.dir * dt;
		ldp     = FLength(dp);
		colStep = intensCol * ldp;
		p       = lr.p + lr.dir*ot;

		// Traverse the ray inside of the light volume
		//for ( ; t<t1; t+=dt,p+=dp) {
		for (int sp=0; sp<samples && t<=tm[i+1]; sp++,t+=dt,p+=dp) {
			distAtten = 1.0f;
			illum = AttenuateIllum(sc,p,colStep,dp,filt,ldp,distAtten,ranges);

			// Call the callback.
			if (!proc->Step(ot,t,illum,distAtten)) {
				return;
				}
			ot = t;
			}

		// Do the remainder
		p      -= dp;
		ot      = t-dt;
		dt      = tm[i+1]-ot;
		dp      = lr.dir * dt;
		p      += dp;

		ldp     = FLength(dp);
		colStep = intensCol * ldp;

		// Attenuate the light at this point
		distAtten = 1.0f;
		illum = AttenuateIllum(sc,p,colStep,dp,filt,ldp,distAtten,ranges);

		// Call the callback.
		if (!proc->Step(ot,tm[i+1],illum,distAtten)) {
			return;
			}
		}

	// The last (large) step
	if (t1<tStop) {
		proc->Step(t1,tStop,Color(0,0,0),1.0f);
		}	
	}

static Point3 MapToDir(Point3 p, int k) {
	switch(k) {
		case 0: return Point3(  p.z, p.y, -p.x); // +X
		case 1: return Point3( -p.z, p.y,  p.x); // -X
		case 2: return Point3(  p.x, p.z, -p.y); // +Y 
		case 3:	return Point3(  p.x,-p.z,  p.y); // -Y
		case 4: return Point3( -p.x, p.y, -p.z); // +Z
		case 5: return p;                        // -Z
		}
 	return p;
	}

static void GetMatrixForDir(Matrix3 &origm, Matrix3 &tm, int k ) {
	tm = origm;
	switch(k) {
		case 0:	tm.PreRotateY(-HALFPI); break;	// Map 0: +X axis	
		case 1: tm.PreRotateY( HALFPI); break;	// Map 1: -X axis	
		case 2:	tm.PreRotateX( HALFPI); break;	// Map 2: +Y axis	
		case 3:	tm.PreRotateX(-HALFPI); break; 	// Map 3: -Y axis	
		case 4:	tm.PreRotateY(   PI  );	break; 	// Map 4: +Z axis	
		case 5: 						break; 	// Map 5: -Z axis	
		}
	}


static int WhichDir(Point3 &p) {
	int j = MaxComponent(p);  // the component with the maximum abs value
	return  (p[j]<0.0f) ? 2*j+1 : 2*j;
	}
static inline float Square(float v) { return v*v; }

//----------------------------------------
// Adjacent buffers:
//------------------------0	1 2	3 4	5 ----
static int  ADJPOSX[6] = {4,5,0,0,1,0};
static int  ADJNEGX[6] = {5,4,1,1,0,1};
static int  ADJPOSY[6] = {3,3,5,4,3,3};
static int  ADJNEGY[6] = {2,2,4,5,2,2};
//-----------------------------------------
#define OkIntersect(tt,axis)  if (tt>=t0 && tt<=t1) { p = ray.p + tt*ray.dir;	if ( MaxComponent(p)!=axis) s[n++]=tt; }

//--- Spot Light Shade Context, for projector lights------------------------

class SCLight: public ShadeContext {
	public:
		ShadeContext *origsc;
		TimeValue curtime;
		Point3 ltPos; // position of point in light space
		Point3 view;  // unit vector from light to point, in light space
		Point2 uv,duv;
		IPoint2 scrpos;
		float curve;
		int projType;

		BOOL 	  InMtlEditor() { return origsc->InMtlEditor(); }
		LightDesc* Light(int n) { return NULL; }
		TimeValue CurTime() { return curtime; }
		int NodeID() { return -1; }
		int FaceNumber() { return 0; }
		int ProjType() { return projType; }
		Point3 Normal() { return Point3(0,0,0); }
		Point3 GNormal() { return Point3(0,0,0); }
		Point3 ReflectVector(){ return Point3(0,0,0); }
		Point3 RefractVector(float ior){ return Point3(0,0,0); }
		Point3 CamPos() { return Point3(0,0,0); }
		Point3 V() { return view; }
		void SetView(Point3 v) { view = v; }
		Point3 P() { return ltPos; }	
		Point3 DP() { return Point3(0,0,0); }
		Point3 PObj() { return ltPos; }
		Point3 DPObj(){ return Point3(0,0,0); } 
		Box3 ObjectBox() { return Box3(Point3(-1,-1,-1),Point3(1,1,1));}   	  	
		Point3 PObjRelBox() { return view; }
		Point3 DPObjRelBox() { return Point3(0,0,0); }
		void ScreenUV(Point2& UV, Point2 &Duv) { UV = uv; Duv = duv; }
		IPoint2 ScreenCoord() { return scrpos;} 
		Point3 UVW(int chan) { return Point3(uv.x, uv.y, 0.0f); }
		Point3 DUVW(int chan) { return Point3(duv.x, duv.y, 0.0f);  }
		void DPdUVW(Point3 dP[3], int chan) {}  // dont need bump vectors
		void GetBGColor(Color &bgcol, Color& transp, BOOL fogBG=TRUE) {}   // returns Background color, bg transparency
		
		float Curve() { return curve; }

		// Transform to and from internal space
		Point3 PointTo(const Point3& p, RefFrame ito) { return p; } 
		Point3 PointFrom(const Point3& p, RefFrame ifrom) { return p; } 
		Point3 VectorTo(const Point3& p, RefFrame ito) { return p; } 
		Point3 VectorFrom(const Point3& p, RefFrame ifrom) { return p; } 
		SCLight(){ doMaps = TRUE; 	curve = 0.0f;		}
	};

#define Dx (ray.dir.x)
#define Dy (ray.dir.y)
#define Dz (ray.dir.z)
#define Px (ray.p.x)
#define Py (ray.p.y)
#define Pz (ray.p.z)

#define DIR(i) ray.dir[i]
#define POS(i)   ray.p[i]
#define BIGFLOAT 1.0e15f

 
BOOL IntersectBox(Ray &ray,float &t0,float &t1, Box3& b) {
	float q0 = -BIGFLOAT;
	float q1 =  BIGFLOAT;
	for (int i=0; i<3; i++) {
		if (DIR(i)==0.0f) {	// parallel to this pair of planes
			if ( POS(i)<b.pmin[i] || POS(i)>b.pmax[i] ) return FALSE;
			}
		else {
			float r0 = (b.pmin[i]-POS(i))/DIR(i);
			float r1 = (b.pmax[i]-POS(i))/DIR(i);
			if (r0 > r1) {float temp=r0; r0=r1; r1=temp;}
			if (r0 > q0) q0 = r0;					
			if (r1 < q1) q1 = r1;					
			if (q0>q1) return FALSE;   // missed box
			if (q1<0.0f) return FALSE; // Box is behind ray origin.
			}
		}
	t0 = q0;
	t1 = q1;
	return TRUE;
	}

BOOL IntersectCyl(Ray &ray,float &t0,float &t1, float Rsq) {
	float A,B,C;
	A = (Dx*Dx + Dy*Dy);
	B = 2.0f*(Dx*Px + Dy*Py);
	C = Px*Px + Py*Py - Rsq;
	if (fabs(A)<.00001) 
		return FALSE;
	float d = B*B-4.0f*A*C;
	if (d<0.0f) 
		return FALSE;
	float s = (float)sqrt(d);
	t0 = (-B + s)/(2.0f*A);
	t1 = (-B - s)/(2.0f*A);
	if (t0 > t1) {float temp=t0; t0=t1; t1=temp;}

	int behind = 0;
	if (Pz + t0*Dz > 0.0f) behind = 1;
	if (Pz + t1*Dz > 0.0f) behind +=2;
	switch (behind) { 
		case 0: break;
		case 1: t0 = -Pz/Dz; 	break;  // intersect with plane z==0
		case 2: t1 = -Pz/Dz; 	break;	// intersect with plane z==0
		case 3: return FALSE;
		}
	return TRUE;
	}

//#define NUMBER_LIGHTS 20
//#define RADIUS 50.0f

//--- Spot Light -------------------------------------------------------
#define MAXSUBS 500
class SSpotLight: public BaseObjLightSoft {	
	friend class SCLight;
	Point3 lightDir;  // light direction in render space
	BOOL rect;
	BOOL overshoot,projector, shadowRay;
	BOOL shadow;
	int shadsize;
	float hot_cos, fall_cos, fall_tan, fall_sin;
	float hotpct, ihotpct;	
	float zfac, xscale,yscale, fov, sz2, curve;
	float out_range,in_range, range_span;
	float decay;
	Point2 rectv0, rectv1;
	float MyBuf[4096*4096];
	ShadowType *st;

	float radius;
	int number_lights;

	Point3 SlightPos[MAXSUBS];
	Matrix3 ScamToLight[MAXSUBS];
	Matrix3 SlightToWorld[MAXSUBS];
//	ShadBufImp *SshadBuf[MAXSUBS];
	ShadowGenerator *SshadBuf[MAXSUBS];

//	ShadowBuffer *shadBuf;
//	ShadowQuadTree *rayBuf;
	Texmap* projMap;
	public:
		SSpotLight(INode *inode);
		~SSpotLight() {
			FreeShadowBuffers();
			st->DeleteThis();
			}
		void FreeShadowBuffers() {
//			if (shadBuf) {	 shadBuf->DeleteThis();	 shadBuf = NULL;   
							
//						}
			for (int numlight =0; numlight<number_lights ; numlight++)
				{
				if (SshadBuf[numlight])
					{
					SshadBuf[numlight]->DeleteThis();
					SshadBuf[numlight] = NULL;
					}
				}

//			if (rayBuf) {	rayBuf->DeleteThis();	rayBuf = NULL;	}
			}
		int Update(TimeValue t, const RendContext& rc, RenderGlobalContext *rgc, BOOL shadows, BOOL shadowGeomChanged);
		int UpdateViewDepParams(const Matrix3& worldToCam);
		BOOL ShadowIlluminate(ShadeContext &sc, Point3& normal, Color& color, Point3 &dir, float& dot_nl, float &diffuseCoef, int numlight, BOOL& inShadow);
		BOOL Illuminate(ShadeContext &sc, Point3& normal, Color& color, Point3 &dir, float& dot_nl, float &diffuseCoef);
		float SampleMap(ShadeContext& sc, Point3 pin, Point3 norm, Point3 plt, Color& lcol, int numlight);
		float ShadowSampleMap(ShadeContext& sc, Point3 pin, Point3 norm, Point3 plt, Color& lcol, int numlight, BOOL& inShadow);
		float ShadowSampleMap2(ShadeContext& sc, Point3 pin, Point3 norm, Point3 plt, Color& lcol, int numlight, BOOL& inShadow);
		float DistanceShadowSampleMap(ShadeContext& sc, Point3 pin, Point3 norm, Point3 plt, Color& lcol, int numlight);

		BOOL IntersectCone(Ray &ray,float tanAng,float &t0,float &t1);
		BOOL IntersectRectCone(Ray &ray,float &t0,float &t1);
		BOOL IntersectRay(Ray &ray,float &t0,float &t1, AttenRanges &ranges);
		Color AttenuateIllum(ShadeContext& sc,Point3 p,Color &colStep,Point3 &dp,int filt, float ldp, float &distAtten, AttenRanges &ranges);		
		AColor SampleProjMap(ShadeContext &sc, Point3 plt, Point3 dp, float x, float y);
		BOOL UseAtten() {return ls.useAtten;}
		BOOL IsFacingLight(Point3 &dir);
		int LightType() { return FSPOT_LIGHT; }
	};

SSpotLight::SSpotLight(INode *inode):BaseObjLightSoft(inode) {
	shadow = rect = overshoot =  shadowRay = FALSE; 
//	rayBuf = NULL;
//	shadBuf = NULL;
	projMap = NULL;
	st = NewDefaultShadowMapType();

	ObjectState os = inode->EvalWorldState(0);
	LightObject* lob = (LightObject *)os.obj;		
	assert(os.obj->SuperClassID()==LIGHT_CLASS_ID);
	GeneralLight* gl = (GeneralLight*)lob;

	for (int numlight =0; numlight<MAXSUBS ; numlight++)
		{
		SshadBuf[numlight] = NULL;

//		SshadBuf[numlight] = new ShadBufImp(shadType,?,this,0);
		SshadBuf[numlight] = st->CreateShadowGenerator(gl, this, 0); 
//		CreateShadowGenerator(gl, this, 0);
		}
//	if(gl->GetShadow()) 



	}


BOOL SSpotLight::IntersectCone(
		Ray &ray,float tanAng,float &t0,float &t1)
	{
#define FLOAT double
	FLOAT A, Dx2, Dy2, Dz2, den, k, root, z0, z1;

	Dx2 = Dx*Dx;
	Dy2 = Dy*Dy;
	Dz2 = Dz*Dz;
	A   = tanAng*tanAng;	

	den = 2.0f*(Dx2+Dy2-A*Dz2);
	if (den==0.0f) return FALSE;

	k = -2.0f*Dx*Px - 2.0f*Dy*Py + 2.0f*A*Dz*Pz;
	root = k*k - 2.0f*den*(Px*Px+Py*Py-A*Pz*Pz);
	if (root<0.0f) return FALSE;
	root = (float)sqrt(root);

	t0 = (float)((k-root)/den);
	t1 = (float)((k+root)/den);
	if (t0 > t1) {float temp=t0;t0=t1;t1=temp;}

	// t0 and t1 may be the iintersection with the reflected cone in the
	// positive Z direction.
	z0 = ray.p.z + ray.dir.z*t0;
	z1 = ray.p.z + ray.dir.z*t1;
	BOOL out0 = z0 > 0.0f ? TRUE : FALSE;
	BOOL out1 = z1 > 0.0f ? TRUE : FALSE;
	if (out0 && out1) {
		return FALSE;	
		}

	if (out0) {			  
		// Must be in the reflected cone, shooting out and hitting the real cone.
		
		// Do an extra check to make sure we really are inside the
		// cone and it's not just round-off.
		if (t0>0) {
			FLOAT l = (Px*Px + Py*Py)/A;
			if (Pz*Pz < l) return FALSE;
			}

		t0 = t1;
		t1 = float(1.0E30);
		}

	if (out1) {
		// Must be inside the cone, shooting out and hitting the reflected cone.
		t1 = t0;
		t0 = 0.0f;
		}
		
	return TRUE;
	}


static BOOL IntersectConeSide(
		Ray &ray,int X, int Y,
		float cs, float sn, float tn,		
		float &t)
	{
	Point2 p(ray.p[X],ray.p.z), v(cs,sn), c(ray.dir[X],ray.dir.z);
	float cv = -c.DotProd(v);
	if (cv!=0.0f) {
		float pv = p.DotProd(v);
		t = pv/cv;
		if (t>0.0f) {
			Point3 pt = ray.p + ray.dir*t;
			if (pt.z<0.0f) {
				if ((float)fabs(pt[Y]/pt.z) <= tn) {
					return TRUE;
					}
				}
			}
		}
	return FALSE;
	}

BOOL SSpotLight::IntersectRectCone(
		Ray &ray,float &t0,float &t1)
	{	
	float sn0 = rectv0.y;
	float sn1 = rectv1.y;
	float cs0 = rectv0.x;
	float cs1 = rectv1.x;
	float tn0 = (float)fabs(sn0/cs0);
	float tn1 = (float)fabs(sn1/cs1);
	int hit = 0;
	float t;
	BOOL in = FALSE;
	
	if (IntersectConeSide(ray, 0, 1, cs0, sn0, tn1, t)) {		
		t0 = t1 = t;
		hit++;
		}

	if (IntersectConeSide(ray, 1, 0, cs1, sn1, tn0, t)) {
		if (hit) {
			if (t<t0) t0 = t;
			if (t>t1) t1 = t;
		} else {
			t0 = t1 = t;
			}
		hit++;
		}

	if (IntersectConeSide(ray, 0, 1, -cs0, sn0, tn1, t)) {
		if (hit) {
			if (t<t0) t0 = t;
			if (t>t1) t1 = t;
		} else {
			t0 = t1 = t;
			}
		hit++;
		}

	if (IntersectConeSide(ray, 1, 0, cs1, -sn1, tn0, t)) {
		if (hit) {
			if (t<t0) t0 = t;
			if (t>t1) t1 = t;
		} else {
			t0 = t1 = t;
			}
		hit++;
		}
	
	float xz = float(fabs(ray.p.x)/fabs(ray.p.z));
	float yz = float(fabs(ray.p.y)/fabs(ray.p.z));
	if (ray.p.z < 0.0f && xz<tn0 && yz<tn1) {
		in = TRUE;
		}

	if (hit == 0) {
		if (in) {
			t0 = 0.0f;
			t1 = float(1.0E30);
			return TRUE;
		} else {
			return FALSE;
			}
		}

	if (hit == 1) {
		if (in) t0 = 0.0f;			
		else t1 = float(1.0E30);			
		}
	
	return TRUE;
	}

#undef Dx
#undef Dy
#undef Dz
#undef Px
#undef Py
#undef Pz


BOOL SSpotLight::IntersectRay(Ray &ray,float &t0,float &t1, AttenRanges &ranges)
	{
	if (rect) return IntersectRectCone(ray,t0,t1);
	else return IntersectCone(ray,fall_tan,t0,t1);	
	}

BOOL SSpotLight::IsFacingLight(Point3 &dir)
	{
	return dir.z>0.0f;
	}

Color SSpotLight::AttenuateIllum(
		ShadeContext& sc,
		Point3 p,Color &colStep,Point3 &dp,int filt, float ldp, float &distAtten, AttenRanges &ranges)
	{		
float mx,my,x,y;	
float atten = 1.0f, cosang;
float fatten = 0.0f;
for (int numlight =0; numlight<number_lights ; numlight++)
	{
	atten = 1.0f;
	float lv = FLength(p);
	if (lv==0.0f) 
		return colStep;	

	// Handle light distance attenuation 
	distAtten = 1.0f;
	if(ls.useNearAtten) {
	   	if(lv<ranges.aNearStart) { // beyond range
			distAtten = 0.0f;
	   		return Color(0.0f,0.0f,0.0f);
			}
		if(lv<ranges.aNearEnd)	{
			// in attenuated range
			float u = (lv - ranges.aNearStart)/(ranges.aNearEnd-ranges.aNearStart);
			atten = distAtten = u*u*(3-2*u);  // smooth cubic curve 
			}
		}
	if(ls.useAtten) {		
		if(lv>ranges.aEnd) {		// Beyond range
			distAtten = 0.0f;
			return Color(0,0,0);
			}
		if(lv>ranges.aStart)	{
			// in attenuated range
			float u = (ranges.aEnd-lv)/(ranges.aEnd-ranges.aStart);
			atten = distAtten *= u*u*(3-2*u); /* smooth cubic curve */			  
			}
		}

	atten = Decay(atten, lv, ranges.aNearEnd);
	
	cosang = (float)fabs(p.z/lv);
	if ((!rect)&&(!overshoot) && (cosang<hot_cos) && (hot_cos!=fall_cos)) {
		if (cosang<fall_cos) {
			// RB: point is outside of falloff cone
			return Color(0.0f,0.0f,0.0f);
			}
		float u = (cosang-fall_cos)/(hot_cos-fall_cos);
		atten *= u*u*(3.0f-2.0f*u);  /* smooth cubic curve */
		}
	
	// ---DS -- 4/8/96 added ||projMap

	if ((p.z<0.0f && (sc.shadow && shadow && SshadBuf[numlight])) || rect || projMap) { 
		mx =  xscale*p.x/p.z;
		my =  yscale*p.y/p.z;
		x = sz2 + mx;
		y = sz2 + my;				
				
		if (sc.shadow && shadow && SshadBuf[numlight]) {

//loop through each shadow buffer
			if (filt) {
				if (filt==1) {
						atten *= SshadBuf[numlight]->FiltSample(int(x), int(y), p.z, filt);
						
					} 
				else if (filt==2) {
					Point3 p2 = p+dp;
					if (p2.z<0.0f) {
						int x2 = int(sz2 + xscale*p2.x/p2.z);
						int y2 = int(sz2 + yscale*p2.y/p2.z);
						atten *= SshadBuf[numlight]->LineSample(
								int(x), int(y), p.z, x2, y2,p2.z);
						}
					}
				else if (filt==3) {
					// Use sample range
					atten *= SshadBuf[numlight]->Sample(sc,x,y,p.z,0.0f,0.0f);
					}
				if (atten==0.0f) return Color(0.0f,0.0f,0.0f);
			} else {		
				if (!SshadBuf[numlight]->QuickSample(int(x), int(y), p.z)) {
					return Color(0.0f,0.0f,0.0f);
					}
				}
			}		
		if (projMap) {

			AColor pc = SampleProjMap(sc,p,dp,x,y);
			return atten*colStep*Color(pc.r,pc.g,pc.b);
			}


		if (rect && !(overshoot)) {
			float u;
			float ux = (float)fabs(mx)/sz2;
			float uy = (float)fabs(my)/sz2;
			if (ux>1.0f || uy>1.0f) {
				atten = 0.0f;
				goto skipAtten;
				}
			int inflag = 0;
			if(ux>hotpct) {
				inflag = 1;	
				ux=(ux-hotpct)/ihotpct;		
				}
			if(uy>hotpct) {	
				inflag |= 2;
				uy=(uy-hotpct)/ihotpct;
				}
			switch(inflag) {
				case 1:	u = 1.0f-ux;			break;
				case 2:	u = 1.0f-uy;			break;
				case 3:	u = (1.0f-ux)*(1.0f-uy); break;
				case 0:	goto skipAtten;
				}
			atten *= u*u*(3.0f-2.0f*u);  /* smooth cubic curve */
			}
		
		}
	fatten += atten;
	}
skipAtten:
	
return colStep * fatten;
}


#define MAX(a,b) (((a)>(b))?(a):(b))

int SSpotLight::Update(TimeValue t, const RendContext &rc,RenderGlobalContext * rgc, BOOL shadows, BOOL shadowGeomChanged) {
	int res = 1;

//update st
	BaseObjLightSoft::Update(t,rc,rgc,shadows,shadowGeomChanged);

	float hs = DegToRad(ls.hotsize);
	float fs = DegToRad(ls.fallsize);
	fall_tan = (float)tan(fs/2.0f);
	hot_cos = (float)cos(hs/2.0f);
	fall_cos =(float)cos(fs/2.0f);
	fall_sin = (float)sin(fs/2.0f);
	hotpct = ls.hotsize/ls.fallsize;
	ihotpct = 1.0f - hotpct;		

	ObjectState os = inode->EvalWorldState(t);
	LightObject* lob = (LightObject *)os.obj;		
	assert(os.obj->SuperClassID()==LIGHT_CLASS_ID);
	GeneralLight* gl = (GeneralLight*)lob;

	//bank = ls.bank;
	rect = gl->GetSpotShape()==RECT_LIGHT?1:0;
	overshoot = gl->GetOvershoot();
	shadow = gl->GetShadow()&shadows;	
	shadowRay = gl->GetShadowType();
	projector =  gl->GetProjector();
	decay = gl->GetRayBias(t);

	number_lights = (int)(gl->GetRayBias(t) * (float) 20.0f);
	
	radius =  gl->GetAspect(t);
	
//	radius = 20.0f;

	fov = MAX(fs,hs);

	float aspect = rect?gl->GetAspect(t):1.0f;
	shadsize = gl->GetMapSize(t);
	if (shadsize<=0) shadsize = 10;
	sz2 = .5f*(float)shadsize;
	 
	fov = 2.0f* (float)atan(tan(fov*0.5f)*sqrt(aspect));
	zfac = -sz2 /(float)tan(0.5*(double)fov);
	xscale = zfac;								
	yscale = -zfac*aspect;
	curve =(float)fabs(1.0f/xscale); 

	//rectv0.y = float(tan(fs*0.5) * sqrt(aspect));
	//rectv1.y = rectv0.y/aspect;
	rectv0.y = fall_sin * (float)sqrt(aspect);
	rectv1.y = fall_sin / (float)sqrt(aspect);

	rectv0.x = rectv1.x = fall_cos;
	rectv0 = Normalize(rectv0);
	rectv1 = Normalize(rectv1);

	ULONG flags = gl->GetAbsMapBias()?1:0;

	st->SetMapSize(t, shadsize);
	st->SetMapBias(t, gl->GetMapBias(t));
	st->SetAbsMapBias(t, flags);



//build xy point list
	float angle_inc, r_inc;
	float turns = 6;
	float angle = 0.0f;
	float r = 0.0f;
	
	
	angle_inc = turns * 3.14f * 2.0f / number_lights;
	r_inc = radius/number_lights;
	
	for (int numlight =0; numlight<number_lights ; numlight++)
		{
		SlightPos[numlight].x  = r * (float) cos(angle);
		SlightPos[numlight].y  = r * (float) sin(angle);
		SlightPos[numlight].z  = 0.0f;
		angle += angle_inc;
		r += r_inc;
		}

	for (numlight =0; numlight<number_lights ; numlight++)
		{
		Point3 p;
		SlightToWorld[numlight] = lightToWorld;
		p = (-SlightPos[numlight]) * lightToWorld;;
		SlightToWorld[numlight].SetTrans(p);
		}


	if (projector) {
		projMap = gl->GetProjMap();
		}
	if (shadow) {	
		if (lightIsZero) {
			FreeShadowBuffers();
			}
		else {
			float clipDist = ls.useAtten? ls.attenEnd : DONT_CLIP;
			if (shadowRay) {	
/*				if (rayBuf == NULL)
					rayBuf = rc.NewShadowQuadTree();
				res = rayBuf->Update(
					*this, 
					rc, 
					lightToWorld,    	// light positio in world
					aspect,
					fov, 
					gl->GetRayBias(t),  // bias
					flags,	// flags
					clipDist
					);
*/
				}
			else {
				for (int numlight =0; numlight<number_lights ; numlight++)
					{



/*					if (SshadBuf[numlight] == NULL)
						SshadBuf[numlight] = rc.NewShadowBuffer();
					res = SshadBuf[numlight]->Update(
						*this,
						rc, 
						SlightToWorld[numlight],    // light position in world
						gl->GetMapSize(t),  // buffer size
						1.0f,
						fov, // fov
						gl->GetMapBias(t),  // bias
						gl->GetMapRange(t), //sampSize
						flags, //  flags
						clipDist
						);
*/
					float clipDist = ls.useAtten? ls.attenEnd : DONT_CLIP;
					SshadBuf[numlight]->Update(t,rc,rgc,SlightToWorld[numlight],aspect,fov,clipDist);

					}
				for (int x = 0; x < gl->GetMapSize(t); x++)
					for (int y = 0; y < gl->GetMapSize(t); y++)
						{
						MyBuf[x+y*gl->GetMapSize(t)] = 0.0f;
						for (int nl = 0; nl < number_lights; nl++)
							{
							if (SshadBuf[nl]->QuickSample( x,y, -5000.0f))
								MyBuf[x+y*gl->GetMapSize(t)] += 1.0f;
							}
						}
				for (x = 0; x < gl->GetMapSize(t); x++)
					{
					for (int y = 0; y < gl->GetMapSize(t); y++)
						{
						MyBuf[x+y*gl->GetMapSize(t)] /= (float) number_lights;
//						DebugPrint("%0.3f ",MyBuf[x+y*gl->GetMapSize(t)]  );
						}
					DebugPrint("\n");
					}

				}
			}
		}
	shadow = shadow&&(SshadBuf[0]);	
	return res;
	}


int  SSpotLight::UpdateViewDepParams(const Matrix3& worldToCam) {
	BaseObjLightSoft::UpdateViewDepParams(worldToCam);
	// get light direction in cam space
	lightDir = -FNormalize(lightToCam.GetRow(2));
//	DebugPrint("Check thread\n");

//build xy point list
	float angle_inc, r_inc;
	float turns = 6;
	float angle = 0.0f;
	float r = 0.0f;
	r_inc = 
	
	
	angle_inc = turns * 3.14f * 2.0f / number_lights;
	r_inc = radius/number_lights;
	
	for (int numlight =0; numlight<number_lights ; numlight++)
		{
		SlightPos[numlight].x  = r * (float) cos(angle);
		SlightPos[numlight].y  = r * (float) sin(angle);
		SlightPos[numlight].z  = 0.0f;
		angle += angle_inc;
		r += r_inc;
		}

	for (numlight =0; numlight<number_lights ; numlight++)
		{
		Point3 p;
		SlightToWorld[numlight] = lightToWorld;
		p = (-SlightPos[numlight]) * lightToWorld;;
		SlightToWorld[numlight].SetTrans(p);
		}


	for (numlight =0; numlight<number_lights ; numlight++)
		{
		if (SshadBuf[numlight]) 
			{
			SshadBuf[numlight]->UpdateViewDepParams(worldToCam);
			}
//build camtolight
		Matrix3 invLightTM,cameraTM;
//need to get camera
		invLightTM = Inverse(SlightToWorld[numlight]);
//need to get inve light
//camera to light CMTM * InvLight
	
		cameraTM = camToLight * lightToWorld;
		cameraTM = cameraTM * invLightTM;
		ScamToLight[numlight] = cameraTM;

	//build light pos in camera space;
		Point3 p(0.0f,0.0f,0.0f);
		SlightPos[numlight] = SlightPos[numlight] * lightToCam;



		}
	return 1;
	}


BOOL SSpotLight::ShadowIlluminate(ShadeContext &sc, Point3& normal, Color& color, Point3& dir, float& dot_nl, float& diffuseCoef, int numlight, BOOL& inShadow) 
	/* SPOTlight : compute light value */

	{

	Point3 p = sc.P();
	float atten,angle,conelimit,dist;
	float str = 1.0f;

	atten = 1.0f;

//need to adjust lightPos for each light
	Point3 L = SlightPos[numlight]-p;  // vector from light to p in camera space
	// quick test for light behind surface
	if (DotProd(normal,L)<=0.0f)	
		return 0;

//need to adjust camToLight for each light
	Point3 plt = ScamToLight[numlight] * p;
	dist = FLength(plt);
	if (dist==0.0f) 
		return 0;

	Point3 LL = -plt/dist;  

	// Handle light distance attenuation 
	if(ls.useNearAtten) {
	   	if(dist<ls.nearAttenStart)		// Beyond range 
	   		return 0;
		if(dist<ls.nearAttenEnd)	{
			// in attenuated range
			float u = (dist - ls.nearAttenStart)/(ls.nearAttenEnd-ls.nearAttenStart);
			atten = u*u*(3-2*u);  /* smooth cubic curve */
			}
		}

	if(ls.useAtten) {
	   	if(dist>ls.attenEnd)		// Beyond range 
	   		return 0;
		if(dist>ls.attenStart)	{
			/* Outside full-intensity range */
			float u = (ls.attenEnd-dist)/(ls.attenEnd-ls.attenStart);
			atten *= u*u*(3-2*u);  // smooth cubic curve 
			}
		}

	atten = Decay(atten,dist,ls.nearAttenEnd);

	conelimit = rect ? (fall_cos/COS_45_2X) : fall_cos;
	color = intensCol;

	
	
//loop through all lights here	
	L = FNormalize(L);

	if (uniformScale) {
		dot_nl = DotProd(normal,L);	
		}
	else {
		Point3	N = FNormalize(VectorTransform(ScamToLight[numlight],normal));
		/* Is light is in front of surface? */
		if ((dot_nl= DotProd(N,LL))<=0.0f) 
			return 0;
		}
	diffuseCoef = ContrastFunc(dot_nl);

	/* Is point in light cone? */
	if ((angle = LL.z) < conelimit) {
		if (!overshoot)	
			return 0; 
		goto overshoot_bypass;
		}
	if (shadow||rect||projMap) 
		{
		float satten = 0.0f;
		satten = ShadowSampleMap(sc,p,normal,plt,color,numlight,inShadow);

/*		if (inShadow)
			{
//			satten += ShadowSampleMap2(sc,p,normal,plt,color,numlight,inShadow);
//get distance in shadow
//			float decay = 0.01f;
//			decay = dlgRayBias;
//need to compute -x,-y,x,y
			float distance = DistanceShadowSampleMap(sc,p,normal,plt,color,numlight);
			if (distance > 0.0f)
				{
//				str = 1.0f-(float)exp(-decay*distance);
				str = distance/50.0f;
DebugPrint("atten %f sattne %f strength %f distance %f \n",atten, satten, str,distance);
				satten += str;

				}

			} */
		atten *= satten;
		}

	if((!rect)&&(!overshoot) && (angle < hot_cos)) {
		float u = (angle-fall_cos)/(hot_cos-fall_cos);
		atten *= u*u*(3.0f-2.0f*u);  /* smooth cubic curve */
		}
/*
	if ((inShadow) && (atten>1.0f))
		atten = 1.0f;
*/
	overshoot_bypass:
	
/*	if ((atten==0.0f) && (!inShadow))
		return 0;
*/
	if (atten==0.0f) 
		return 0;

	dir = L;  // direction in camera space
	atten = atten / number_lights;

	if (atten > 1,0)
		{
		DebugPrint("Overflow atten\n");
		}
	if (atten!=1.0f) color *= atten;

	return 1;
	}



BOOL SSpotLight::Illuminate(ShadeContext &sc, Point3& normal, Color& color, Point3 &dir, float& dot_nl, float &diffuseCoef) {
	/* SPOTlight : compute light value */


/*
for (int numlight =0; numlight<number_lights ; numlight++)
	{

//build camtolight
		Matrix3 invLightTM,cameraTM;
//need to get camera
		invLightTM = Inverse(SlightToWorld[numlight]);
//need to get inve light
//camera to light CMTM * InvLight
	
		cameraTM = camToLight * lightToWorld;
		cameraTM = cameraTM * invLightTM;
		ScamToLight[numlight] = cameraTM;

	//build light pos in camera space;
		Point3 p(0.0f,0.0f,0.0f);
		SlightPos[numlight] = SlightPos[numlight] * lightToCam;



	}
*/


int il = 0;

//get light 0;
BOOL inShadow = FALSE;
color.r = 0.0f;
color.g = 0.0f;
color.b = 0.0f;
Point3 SDir = dir;//(0.0f,0.0f,0.0f);
float Sdot_nl = dot_nl;
float SdiffuseCoef = diffuseCoef;

for (int numlight =0; numlight<number_lights ; numlight++)
//for (int numlight =0; numlight<1 ; numlight++)
	{
	Color SColor(0.0f,0.0f,0.0f);
	
	if (ShadowIlluminate(sc,normal,SColor, SDir, Sdot_nl, diffuseCoef, numlight, inShadow) )
		{
		if (numlight == 0)
			{
			dir = SDir;
			dot_nl = Sdot_nl;
//			diffuseCoef = SdiffuseCoef;
			}
		color += SColor;
		il++;
//set shadow

		}
	}


if (il == 0) return 0;
	else
	{ 
	color = color;
	return 1;
	}
}



AColor SSpotLight::SampleProjMap(ShadeContext &sc,  Point3 plt, Point3 dp, float x, float y) {
	SCLight scspot;
	float ifs = 1.0f/(float)shadsize;
	scspot.projType = 0; // perspective
	scspot.curtime = sc.CurTime();
	scspot.curve =(float)fabs(1.0f/xscale); 
	scspot.ltPos = plt;
	scspot.view = FNormalize(plt);
	scspot.uv.x = x*ifs;			
	scspot.uv.y = 1.0f-y*ifs;			

	scspot.scrpos.x = (int)(x+0.5);
	scspot.scrpos.y = (int)(y+0.5);
	scspot.filterMaps = sc.filterMaps;
	scspot.mtlNum  = sc.mtlNum;

	// antialiasing for 2D textures:
	float d = MaxAbs(dp.x,dp.y,dp.z);
	if (plt.z==0.0f) plt.z = .00001f;
	scspot.duv.x = scspot.duv.y = xscale*d*ifs/plt.z;

	// antialiasing for 3D textures: TBD
	return projMap->EvalColor(scspot);
	}

float SSpotLight::SampleMap(ShadeContext& sc, Point3 pin, Point3 norm, Point3 plt, Color& lcol, int numlight) {
	float mx,my,x,y,atten;
	Point3 pout,n;
	Color col;
	atten = 1.0f;
	if (plt.z<0.0f) {
		mx =  xscale*plt.x/plt.z;
		my =  yscale*plt.y/plt.z;
		x = sz2 + mx;
		y = sz2 + my;

		/* If it falls outside the rectangle, bail out */
		if ((float)fabs(mx)>sz2 || (float)fabs(my)>sz2)	{
			if(overshoot) return(1.0f);	/* No attenuation */
			else return(0.0f);
			}

		if (sc.shadow&&shadow) {
/*			if (rayBuf) {
				atten = rayBuf->Sample(sc,pin,lcol);
				if (atten==0.0f) return 0.0f;
				}
			else */
			if (SshadBuf[numlight])  {
				Point3 n = VectorTransform(ScamToLight[numlight],norm);
				float k = plt.z*plt.z/(zfac*DotProd(n,plt));
				atten = SshadBuf[numlight]->Sample(sc, x,y, plt.z, -n.x*k, n.y*k);
				if (atten==0.0f) return(0.0f);
				}
			}
		
	// Calculate rectangular dropoff if outside the hotspot area 
	if (rect && !(overshoot)) {
		float u;
		float ux = (float)fabs(mx)/sz2;
		float uy = (float)fabs(my)/sz2;
		int inflag = 0;
		if(ux>hotpct) {
			inflag = 1;	
			ux=(ux-hotpct)/ihotpct;		
			}
		if(uy>hotpct) {	
			inflag |= 2;
			uy=(uy-hotpct)/ihotpct;
			}
		switch(inflag) {
			case 1:	u = 1.0f-ux;			break;
			case 2:	u = 1.0f-uy;			break;
			case 3:	u = (1.0f-ux)*(1.0f-uy); break;
			case 0:	goto noAtten;
			}
		atten *= u*u*(3.0f-2.0f*u);  /* smooth cubic curve */

		}
	noAtten:		  
		if(projMap) 
			lcol *= SampleProjMap(sc,plt,sc.DP(),x,y);
		}
	return(atten);
	}


float SSpotLight::ShadowSampleMap(ShadeContext& sc, Point3 pin, Point3 norm, Point3 plt, Color& lcol, int numlight, BOOL& inShadow) {
	float mx,my,x,y,atten;
	Point3 pout,n;
	Color col;
	atten = 1.0f;
	if (plt.z<0.0f) {
		mx =  xscale*plt.x/plt.z;
		my =  yscale*plt.y/plt.z;
		x = sz2 + mx;
		y = sz2 + my;

		/* If it falls outside the rectangle, bail out */
		if ((float)fabs(mx)>sz2 || (float)fabs(my)>sz2)	{
			if(overshoot) return(1.0f);	/* No attenuation */
			else return(0.0f);
			}

		if (sc.shadow&&shadow) {
/*			if (rayBuf) {
				atten = rayBuf->Sample(sc,pin,lcol);
				if (atten==0.0f) return 0.0f;
				}
			else */
			if (SshadBuf[numlight])  {
				Point3 n = VectorTransform(ScamToLight[numlight],norm);
				float k = plt.z*plt.z/(zfac*DotProd(n,plt));
				atten = SshadBuf[numlight]->Sample(sc, x,y, plt.z, -n.x*k, n.y*k);

				
//				atten = MyBuf[(int)x + (int)y*shadsize];
				if (atten != 1.0f)

					{
					 inShadow = TRUE;
					lcol.r = MyBuf[(int)x + (int)y*shadsize];
					lcol.g = 0.0f;
					lcol.b = 0.0f;
					}
					else inShadow = FALSE;
				if (atten==0.0f) return(0.0f);
				}
			}
		
	// Calculate rectangular dropoff if outside the hotspot area 
	if (rect && !(overshoot)) {
		float u;
		float ux = (float)fabs(mx)/sz2;
		float uy = (float)fabs(my)/sz2;
		int inflag = 0;
		if(ux>hotpct) {
			inflag = 1;	
			ux=(ux-hotpct)/ihotpct;		
			}
		if(uy>hotpct) {	
			inflag |= 2;
			uy=(uy-hotpct)/ihotpct;
			}
		switch(inflag) {
			case 1:	u = 1.0f-ux;			break;
			case 2:	u = 1.0f-uy;			break;
			case 3:	u = (1.0f-ux)*(1.0f-uy); break;
			case 0:	goto noAtten;
			}
		atten *= u*u*(3.0f-2.0f*u);  /* smooth cubic curve */

		}
	noAtten:		  
		if(projMap) 
			lcol *= SampleProjMap(sc,plt,sc.DP(),x,y);
		}
	return(atten);
	}

float SSpotLight::ShadowSampleMap2(ShadeContext& sc, Point3 pin, Point3 norm, Point3 plt, Color& lcol, int numlight, BOOL& inShadow) {
	float mx,my,x,y,atten;
	Point3 pout,n;
	Color col;
	atten = 1.0f;
//	if (plt.z<0.0f) {
	Point3 tplt = plt;
	int ct =0;
	float angle = 1.0f*3.14f/180.0f;
	if (plt.z<0.0f)

	while (ct < 4) {
		plt = tplt;
		if (ct==0)
			{
//			plt.x += decay;
//rotate pl.x and z by 10 degrees
			plt.x = (plt.x * (float)cos(-angle)) - (plt.z *(float)sin(-angle));
			plt.z = (plt.x * (float)sin(-angle)) + (plt.z *(float)cos(-angle));
			}
		else if (ct==1)
			{
//			plt.x += -decay;
			plt.x = (plt.x * (float)cos(angle)) - (plt.z *(float)sin(angle));
			plt.z = (plt.x * (float)sin(angle)) + (plt.z *(float)cos(angle));
			}
		else if (ct==2)
			{
//			plt.y += decay;
			plt.y = (plt.y * (float)cos(-angle)) - (plt.z *(float)sin(-angle));
			plt.z = (plt.y * (float)sin(-angle)) + (plt.z *(float)cos(-angle));
			}
		else if (ct==3)
			{
//			plt.y += -decay;
			plt.y = (plt.y * (float)cos(angle)) - (plt.z *(float)sin(angle));
			plt.z = (plt.y * (float)sin(angle)) + (plt.z *(float)cos(angle));
			}
		ct++;
		mx =  xscale*plt.x/plt.z;
		my =  yscale*plt.y/plt.z;
		x = sz2 + mx;
		y = sz2 + my;

		/* If it falls outside the rectangle, bail out */
		if ((float)fabs(mx)>sz2 || (float)fabs(my)>sz2)	{
			if(overshoot) return(1.0f);	/* No attenuation */
			else return(0.0f);
			}

		if (sc.shadow&&shadow) {
/*			if (rayBuf) {
				atten = rayBuf->Sample(sc,pin,lcol);
				if (atten==0.0f) return 0.0f;
				}
			else */
			if (SshadBuf[numlight])  {
				Point3 n = VectorTransform(ScamToLight[numlight],norm);
				float k = plt.z*plt.z/(zfac*DotProd(n,plt));
				atten += SshadBuf[numlight]->Sample( sc,x,y, plt.z, -n.x*k, n.y*k);
				if (atten != 1.0f) inShadow = TRUE;
					else inShadow = FALSE;
//				if (atten==0.0f) return(0.0f);
				}
			}
		}
	// Calculate rectangular dropoff if outside the hotspot area 
	if (rect && !(overshoot)) {
		float u;
		float ux = (float)fabs(mx)/sz2;
		float uy = (float)fabs(my)/sz2;
		int inflag = 0;
		if(ux>hotpct) {
			inflag = 1;	
			ux=(ux-hotpct)/ihotpct;		
			}
		if(uy>hotpct) {	
			inflag |= 2;
			uy=(uy-hotpct)/ihotpct;
			}
		switch(inflag) {
			case 1:	u = 1.0f-ux;			break;
			case 2:	u = 1.0f-uy;			break;
			case 3:	u = (1.0f-ux)*(1.0f-uy); break;
			case 0:	goto noAtten;
			}
		atten *= u*u*(3.0f-2.0f*u);  /* smooth cubic curve */

		}
	noAtten:		  
		if(projMap) 
			lcol *= SampleProjMap(sc,plt,sc.DP(),x,y);
//		}	
	return(atten);
	}


float SSpotLight::DistanceShadowSampleMap(ShadeContext& sc, Point3 pin, Point3 norm, Point3 plt, Color& lcol, int numlight) {
	
	float mx,my,x,y,atten;
	Point3 pout,n;
	Color col;
	atten = 1.0f;
	float distance = 0.0f;

	Point3 DirectionToLight;
	BOOL inShadow = TRUE;

	DirectionToLight = Normalize(-plt);
	float DistanceToLight = Length(plt);

//	while ()
//float LineSample(int x1, int y1, float z1,int x2, int y2, float z2)
	
	while ((plt.z<0.0f) && (inShadow)) {
		mx =  xscale*plt.x/plt.z;
		my =  yscale*plt.y/plt.z;
		x = sz2 + mx;
		y = sz2 + my;

		if (sc.shadow&&shadow) 
			{
/*			if (rayBuf) 
				{
				atten = rayBuf->Sample(sc,pin,lcol);
				if (atten==0.0f) return 0.0f;
				}
			else */
			if (SshadBuf[numlight])  
				{
				Point3 n = VectorTransform(ScamToLight[numlight],norm);
				float k = plt.z*plt.z/(zfac*DotProd(n,plt));
//				if (SshadBuf[numlight]->QuickSample( (int)x,(int)y, plt.z)) inShadow = TRUE;
//					else inShadow = FALSE;
				atten = SshadBuf[numlight]->Sample( sc,x,y, plt.z, -n.x*k, n.y*k);
				if (atten != 1.0f) inShadow = TRUE;
					else inShadow = FALSE;

//				if (atten==0.0f) return distance;
				}
			}

		if (inShadow)
			{
			plt += DirectionToLight;
			distance += 1.0f;
			}
		}

	return(distance);
	}


ObjLightDesc *GeneralLight::CreateLightDesc(INode *n)
	{
	switch (type) {
/*
		case OMNI_LIGHT: 	return new OmniLight(n);
		case DIR_LIGHT:		
		case TDIR_LIGHT:   	return new DirLight(n);
		case FSPOT_LIGHT:	*/
		case TSPOT_LIGHT:	return new SSpotLight(n);
		default:			return NULL;
		}	
	}


//---------------------------------------------------------------------------
// Class IO: The spotlight does the saving and loading of the shadow parameters
// for all the light classes.
IOResult STSpotLightClassDesc::Save(ISave *isave){
	ULONG nb;
	isave->BeginChunk(SHADOW_TYPE_CHUNK);
	isave->Write(&GeneralLight::globShadowType, sizeof(GeneralLight::globShadowType), &nb);
	isave->EndChunk();
	isave->BeginChunk(ABS_MAP_BIAS_CHUNK);
	isave->Write(&GeneralLight::globAbsMapBias, sizeof(GeneralLight::globAbsMapBias), &nb);
	isave->EndChunk();
	isave->BeginChunk(MAP_BIAS_CHUNK);
	isave->Write(&GeneralLight::globMapBias, sizeof(GeneralLight::globMapBias), &nb);
	isave->EndChunk();
	isave->BeginChunk(MAP_RANGE_CHUNK);
	isave->Write(&GeneralLight::globMapRange, sizeof(GeneralLight::globMapRange), &nb);
	isave->EndChunk();
	isave->BeginChunk(MAP_SIZE_CHUNK);
	isave->Write(&GeneralLight::globMapSize, sizeof(GeneralLight::globMapSize), &nb);
	isave->EndChunk();
	isave->BeginChunk(RAY_BIAS_CHUNK);
	isave->Write(&GeneralLight::globRayBias, sizeof(GeneralLight::globRayBias), &nb);
	isave->EndChunk();
	return IO_OK;
	}

IOResult STSpotLightClassDesc::Load(ILoad *iload){
	ULONG nb;
	IOResult res;
	while (IO_OK==(res=iload->OpenChunk())) {
		switch(iload->CurChunkID())  {
			case SHADOW_TYPE_CHUNK:
				res = iload->Read(&GeneralLight::globShadowType, sizeof(GeneralLight::globShadowType), &nb);
				break;
			case ABS_MAP_BIAS_CHUNK:
				res = iload->Read(&GeneralLight::globAbsMapBias, sizeof(GeneralLight::globAbsMapBias), &nb);
				break;
			case MAP_BIAS_CHUNK:
				res = iload->Read(&GeneralLight::globMapBias, sizeof(GeneralLight::globMapBias), &nb);
				break;
			case MAP_RANGE_CHUNK:
				res = iload->Read(&GeneralLight::globMapRange, sizeof(GeneralLight::globMapRange), &nb);
				break;
			case MAP_SIZE_CHUNK:
				res = iload->Read(&GeneralLight::globMapSize, sizeof(GeneralLight::globMapSize), &nb);
				break;
			case RAY_BIAS_CHUNK:
				res = iload->Read(&GeneralLight::globRayBias, sizeof(GeneralLight::globRayBias), &nb);
				break;

			}
		iload->CloseChunk();
		if (res!=IO_OK) 
			return res;
		}
	return IO_OK;
	}
